Package: oeli
Type: Package
Title: Some Utilities for Developing Data Science Software
Version: 0.7.6
Authors@R: person(
    given = "Lennart", 
    family = "Oelschl\u00e4ger", 
    email = "oelschlaeger.lennart@gmail.com", 
    role = c("aut", "cre")
  )
Description: A collection of general-purpose helper functions that I (and maybe 
    others) find useful when developing data science software. Includes tools 
    for simulation, data transformation, input validation, and more.
License: GPL (>= 3)
Encoding: UTF-8
Config/testthat/edition: 3
RoxygenNote: 7.3.3
Depends: R (>= 4.1.0)
Imports: benchmarkme, checkmate, cli, cubature, dplyr, future,
        future.apply, ggplot2, glue, hexSticker, progressr, R6, Rcpp,
        SimMultiCorrData, stats, tibble, tools, utils
LinkingTo: mvtnorm, Rcpp (>= 1.1.1), RcppArmadillo, testthat
Suggests: knitr, rmarkdown, testthat (>= 3.0.0), xml2
URL: https://github.com/loelschlaeger/oeli,
        http://loelschlaeger.de/oeli/
BugReports: https://github.com/loelschlaeger/oeli/issues
LazyData: true
NeedsCompilation: yes
Packaged: 2026-04-22 22:04:13 UTC; loelschlaeger
Author: Lennart Oelschläger [aut, cre]
Maintainer: Lennart Oelschläger <oelschlaeger.lennart@gmail.com>
Repository: CRAN
Date/Publication: 2026-04-23 05:10:02 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-04-25 13:55:42 UTC; unix
Archs: oeli.so.dSYM

# Generated by roxygen2: do not edit by hand

export(Dictionary)
export(M)
export(Simulator)
export(Storage)
export(assert_correlation_matrix)
export(assert_covariance_matrix)
export(assert_list_of_lists)
export(assert_missing)
export(assert_numeric_vector)
export(assert_one_hot_matrix)
export(assert_probability_vector)
export(assert_transition_probability_matrix)
export(check_correlation_matrix)
export(check_covariance_matrix)
export(check_list_of_lists)
export(check_missing)
export(check_numeric_vector)
export(check_one_hot_matrix)
export(check_probability_vector)
export(check_transition_probability_matrix)
export(chol_to_cov)
export(chunk_vector)
export(correlated_regressors)
export(cov_to_chol)
export(ddirichlet)
export(ddirichlet_cpp)
export(delete_columns_data.frame)
export(delta)
export(diff_cov)
export(dmixnorm)
export(dmixnorm_cpp)
export(dmvnorm)
export(dmvnorm_cpp)
export(do.call_timed)
export(dtnorm)
export(dtnorm_cpp)
export(dttnorm)
export(dttnorm_cpp)
export(dwishart)
export(dwishart_cpp)
export(equidistant_vectors)
export(find_namespace_calls)
export(function_arguments)
export(function_body)
export(function_defaults)
export(gaussian_tv)
export(group_data.frame)
export(identical_structure)
export(input_check_response)
export(insert_matrix_column)
export(insert_vector_entry)
export(map_indices)
export(match_arg)
export(match_numerics)
export(matrix_diagonal_indices)
export(matrix_indices)
export(merge_lists)
export(occurrence_info)
export(package_logo)
export(permutations)
export(pmixnorm)
export(pmixnorm_cpp)
export(pmvnorm)
export(pmvnorm_cpp)
export(print_data.frame)
export(print_matrix)
export(quiet)
export(rdirichlet)
export(rdirichlet_cpp)
export(rmixnorm)
export(rmixnorm_cpp)
export(rmvnorm)
export(rmvnorm_cpp)
export(round_data.frame)
export(rtnorm)
export(rtnorm_cpp)
export(rttnorm)
export(rttnorm_cpp)
export(rwishart)
export(rwishart_cpp)
export(sample_correlation_matrix)
export(sample_covariance_matrix)
export(sample_transition_probability_matrix)
export(simulate_markov_chain)
export(split_vector_at)
export(stationary_distribution)
export(subsets)
export(system_information)
export(test_correlation_matrix)
export(test_covariance_matrix)
export(test_list_of_lists)
export(test_missing)
export(test_numeric_vector)
export(test_one_hot_matrix)
export(test_probability_vector)
export(test_transition_probability_matrix)
export(timed)
export(try_silent)
export(undiff_cov)
export(unexpected_error)
export(unique_chol)
export(user_confirm)
export(variable_name)
export(vector_occurrence)
importFrom(R6,R6Class)
importFrom(Rcpp,compileAttributes)
importFrom(SimMultiCorrData,calc_theory)
importFrom(SimMultiCorrData,rcorrvar)
importFrom(SimMultiCorrData,valid_corr)
importFrom(benchmarkme,get_cpu)
importFrom(benchmarkme,get_ram)
importFrom(checkmate,assert)
importFrom(checkmate,assert_atomic_vector)
importFrom(checkmate,assert_character)
importFrom(checkmate,assert_choice)
importFrom(checkmate,assert_count)
importFrom(checkmate,assert_data_frame)
importFrom(checkmate,assert_flag)
importFrom(checkmate,assert_function)
importFrom(checkmate,assert_int)
importFrom(checkmate,assert_integerish)
importFrom(checkmate,assert_list)
importFrom(checkmate,assert_matrix)
importFrom(checkmate,assert_names)
importFrom(checkmate,assert_number)
importFrom(checkmate,assert_numeric)
importFrom(checkmate,assert_string)
importFrom(checkmate,assert_subset)
importFrom(checkmate,assert_vector)
importFrom(checkmate,check_atomic_vector)
importFrom(checkmate,check_character)
importFrom(checkmate,check_choice)
importFrom(checkmate,check_class)
importFrom(checkmate,check_count)
importFrom(checkmate,check_data_frame)
importFrom(checkmate,check_directory_exists)
importFrom(checkmate,check_flag)
importFrom(checkmate,check_function)
importFrom(checkmate,check_int)
importFrom(checkmate,check_integerish)
importFrom(checkmate,check_list)
importFrom(checkmate,check_matrix)
importFrom(checkmate,check_names)
importFrom(checkmate,check_number)
importFrom(checkmate,check_numeric)
importFrom(checkmate,check_path_for_output)
importFrom(checkmate,check_string)
importFrom(checkmate,check_subset)
importFrom(checkmate,makeAssertion)
importFrom(checkmate,makeAssertionFunction)
importFrom(checkmate,makeTestFunction)
importFrom(checkmate,test_atomic_vector)
importFrom(checkmate,test_class)
importFrom(checkmate,test_int)
importFrom(checkmate,test_list)
importFrom(checkmate,test_matrix)
importFrom(checkmate,test_string)
importFrom(checkmate,vname)
importFrom(cli,cli_abort)
importFrom(cli,cli_inform)
importFrom(cli,cli_text)
importFrom(cli,cli_warn)
importFrom(cli,style_hyperlink)
importFrom(cubature,hcubature)
importFrom(dplyr,arrange)
importFrom(dplyr,bind_rows)
importFrom(dplyr,filter)
importFrom(future,multisession)
importFrom(future,plan)
importFrom(future.apply,future_lapply)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,theme_void)
importFrom(glue,glue)
importFrom(hexSticker,sticker)
importFrom(progressr,handler_progress)
importFrom(progressr,handlers)
importFrom(progressr,progressor)
importFrom(stats,cor)
importFrom(stats,dist)
importFrom(stats,dnorm)
importFrom(stats,rnorm)
importFrom(stats,runif)
importFrom(tibble,as_tibble)
importFrom(tibble,tibble)
importFrom(tools,package_native_routine_registration_skeleton)
importFrom(utils,combn)
importFrom(utils,head)
importFrom(utils,tail)
useDynLib(oeli, .registration=TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# Utilities for developing R code <a href="https://loelschlaeger.de/oeli/"><img src="man/figures/logo.png" align="right" height="138" /></a>

<!-- badges: start -->

[![CRAN
status](https://www.r-pkg.org/badges/version/oeli)](https://CRAN.R-project.org/package=oeli)
[![CRAN
downloads](https://cranlogs.r-pkg.org/badges/last-month/oeli)](https://CRAN.R-project.org/package=oeli)
[![R-CMD-check](https://github.com/loelschlaeger/oeli/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/loelschlaeger/oeli/actions/workflows/R-CMD-check.yaml)
[![Codecov test
coverage](https://codecov.io/gh/loelschlaeger/oeli/graph/badge.svg)](https://app.codecov.io/gh/loelschlaeger/oeli)
<!-- badges: end -->

This [R](https://CRAN.R-project.org) package provides helper functions I
found useful when developing R code - perhaps you will too! The released
package version can be installed via:

``` r
install.packages("oeli")
```

The following shows some demos. Click the headings for references on all
available helpers in each category.

### [Data](https://loelschlaeger.de/oeli/reference/index.html#data)

The `hermann` data contains historical information on editions of the
[Hermannslauf](https://de.wikipedia.org/wiki/Hermannslauf), including
the date, temperature, and winning times for men and women:

``` r
hermann
#> # A tibble: 54 × 8
#>    edition  year date        temp winner_men        seconds_men winner_women  
#>      <dbl> <dbl> <date>     <dbl> <chr>                   <dbl> <chr>         
#>  1       1  1972 1972-04-16    NA Helmut Bode              6686 Lydia Günnewig
#>  2       2  1973 1973-04-29    14 Helmut Bode              6795 Irmhild Holste
#>  3       3  1974 1974-04-28    14 Achim Stober             6942 Liane Winter  
#>  4       4  1975 1975-04-27    11 Klaus-Dieter Holz        6720 Christine Ross
#>  5       5  1976 1976-04-25    11 Heribert Bulk            6453 Liane Winter  
#>  6       6  1977 1977-04-24     9 Jim Hodey                6503 Liane Winter  
#>  7       7  1978 1978-04-30    16 Michael Heine            6856 Liane Winter  
#>  8       8  1979 1979-04-29     7 Billy Cain               6587 Liane Winter  
#>  9       9  1980 1980-04-27     9 Dieter Lippe             6706 Liane Winter  
#> 10      10  1981 1981-04-26    16 Helmut Schmidt           6735 Rotraud Zinner
#> # ℹ 44 more rows
#> # ℹ 1 more variable: seconds_women <dbl>
```

### [Distributions](https://loelschlaeger.de/oeli/reference/index.html#distribution)

The package has density and sampling functions for some distributions
not included in base R, like the Dirichlet:

``` r
ddirichlet(x = c(0.2, 0.3, 0.5), concentration = 1:3)
#> [1] 4.5
rdirichlet(concentration = 1:3)
#> [1] 0.01795087 0.41315984 0.56888929
```

Or the mixture of Gaussian distributions:

``` r
x <- c(0, 0)
mean <- matrix(c(1, 1, -1, -1), ncol = 2) # means in columns
Sigma <- matrix(c(diag(2), 0.1 * diag(2)), ncol = 2) # vectorized covariances in columns
proportions <- c(0.7, 0.3)
dmixnorm(x = x, mean = mean, Sigma = Sigma, proportions = proportions)
#> [1] 0.04100656
pmixnorm(x = x, mean = mean, Sigma = Sigma, proportions = proportions)
#> [1] 0.3171506
rmixnorm(n = 1000, mean = mean, Sigma = Sigma, proportions = proportions) |>
  as.data.frame() |> 
  ggplot2::ggplot() + ggplot2::geom_point(ggplot2::aes(x = V1, y = V2))
```

<img src="man/figures/README-mixnorm-1.png" alt="" width="50%" style="display: block; margin: auto;" />

### [Function helpers](https://loelschlaeger.de/oeli/reference/index.html#functional)

Retrieving default arguments of a `function`:

``` r
f <- function(a, b = 1, c = "", ...) { }
function_defaults(f)
#> $b
#> [1] 1
#> 
#> $c
#> [1] ""
```

### [Indexing helpers](https://loelschlaeger.de/oeli/reference/index.html#indexing)

Create all possible permutations of vector elements:

``` r
permutations(LETTERS[1:3])
#> [[1]]
#> [1] "A" "B" "C"
#> 
#> [[2]]
#> [1] "A" "C" "B"
#> 
#> [[3]]
#> [1] "B" "A" "C"
#> 
#> [[4]]
#> [1] "B" "C" "A"
#> 
#> [[5]]
#> [1] "C" "A" "B"
#> 
#> [[6]]
#> [1] "C" "B" "A"
```

### [Package helpers](https://loelschlaeger.de/oeli/reference/index.html#packaging)

Quickly have a basic logo for your new package:

``` r
logo <- package_logo("my_package", brackets = TRUE)
print(logo)
```

<img src="man/figures/README-package_logo-1.png" alt="" width="50%" style="display: block; margin: auto;" />

How to print a `matrix` without filling up the entire console?

``` r
x <- matrix(rnorm(10000), ncol = 100, nrow = 100)
print_matrix(x, rowdots = 4, coldots = 4, digits = 2, label = "what a big matrix")
#> what a big matrix : 100 x 100 matrix of doubles 
#>         [,1]  [,2] [,3] ... [,100]
#> [1,]    -0.3 -0.74 -0.1 ...   1.01
#> [2,]    1.39 -2.06 1.29 ...   -0.5
#> [3,]   -0.45 -1.57 0.43 ...   1.61
#> ...      ...   ...  ... ...    ...
#> [100,]  1.12  0.77 -1.6 ...  -0.08
```

And what about a `data.frame`?

``` r
x <- data.frame(x = rnorm(1000), y = LETTERS[1:10])
print_data.frame(x, rows = 7, digits = 0)
#>      x  y
#> 1     0 A
#> 2     1 B
#> 3     0 C
#> 4    -1 D
#> <993 rows hidden>
#>          
#> 998   0 H
#> 999   0 I
#> 1000  2 J
```

### [Simulation helpers](https://loelschlaeger.de/oeli/reference/index.html#simulation)

Let’s simulate correlated regressor values from different marginal
distributions:

``` r
labels <- c("P", "C", "N1", "N2", "U")
n <- 100
marginals <- list(
  "P" = list(type = "poisson", lambda = 2),
  "C" = list(type = "categorical", p = c(0.3, 0.2, 0.5)),
  "N1" = list(type = "normal", mean = -1, sd = 2),
  "U" = list(type = "uniform", min = -2, max = -1)
)
correlation <- matrix(
  c(1, -0.3, -0.1, 0, 0.5,
    -0.3, 1, 0.3, -0.5, -0.7,
    -0.1, 0.3, 1, -0.3, -0.3,
    0, -0.5, -0.3, 1, 0.1,
    0.5, -0.7, -0.3, 0.1, 1),
  nrow = 5, ncol = 5
)
data <- correlated_regressors(
  labels = labels, n = n, marginals = marginals, correlation = correlation
)
head(data)
#>   P C        N1         N2         U
#> 1 3 3  0.854606 -1.1509971 -1.575507
#> 2 3 3 -3.801998  0.5696389 -1.784321
#> 3 4 3 -1.024611 -1.0414976 -1.333803
#> 4 0 3 -0.490758 -0.9806894 -1.840349
#> 5 1 3 -1.681134  0.7511786 -1.939042
#> 6 2 1 -2.814986  0.7984411 -1.367265
cor(data)
#>              P          C          N1          N2          U
#> P   1.00000000 -0.2287181 -0.08793083 -0.02476611  0.4877251
#> C  -0.22871807  1.0000000  0.28573358 -0.52539377 -0.7628656
#> N1 -0.08793083  0.2857336  1.00000000 -0.30000000 -0.2694518
#> N2 -0.02476611 -0.5253938 -0.30000000  1.00000000  0.1138354
#> U   0.48772506 -0.7628656 -0.26945184  0.11383544  1.0000000
```

### [Transformation helpers](https://loelschlaeger.de/oeli/reference/index.html#transformation)

The `group_data.frame()` function groups a given `data.frame` based on
the values in a specified column:

``` r
df <- data.frame("label" = c("A", "B"), "number" = 1:10)
group_data.frame(df = df, by = "label")
#> $A
#>   label number
#> 1     A      1
#> 3     A      3
#> 5     A      5
#> 7     A      7
#> 9     A      9
#> 
#> $B
#>    label number
#> 2      B      2
#> 4      B      4
#> 6      B      6
#> 8      B      8
#> 10     B     10
```

### [Validation helpers](https://loelschlaeger.de/oeli/reference/index.html#validation)

Is my matrix a proper transition probability matrix?

``` r
matrix <- diag(4)
matrix[1, 2] <- 1
check_transition_probability_matrix(matrix)
#> [1] "Must have row sums equal to 1"
```

#' Insert entry in vector
#'
#' @description
#' This function inserts a value into a vector.
#'
#' @param v \[`atomic()`\]\cr
#' A \code{vector}.
#'
#' @param x \[`atomic(1)`\]\cr
#' The entry to be added.
#'
#' @param p \[`integer())`\]\cr
#' The position(s) where to add the value, one or more of:
#' - \code{p = 0} appends the value left
#' - \code{p = length(v)} appends the value right
#' - \code{p = n} inserts the value between the \code{n}-th and
#'   \code{(n + 1)}-th entry of \code{v}.
#'
#' @return
#' A \code{vector}.
#'
#' @export
#' @keywords transformation
#' @family vector helpers
#'
#' @examples
#' v <- 1:3
#' x <- 0
#' insert_vector_entry(v, x, 0)
#' insert_vector_entry(v, x, 1)
#' insert_vector_entry(v, x, 2)
#' insert_vector_entry(v, x, 3)
#'
#' ### also multiple positions
#' insert_vector_entry(v, x, 0:3)
#'
#' ### also trivial case
#' insert_vector_entry(integer(), integer(), integer())

insert_vector_entry <- function(v, x, p) {
  checkmate::assert_atomic_vector(v)
  n <- length(v)
  checkmate::assert_atomic_vector(x, max.len = 1)
  checkmate::assert_integerish(
    p, lower = 0, upper = n, unique = TRUE, max.len = n + 1
  )
  if (length(p) == 0) {
    return(v)
  }
  c(v, rep(x, length(p)))[order(c(seq_along(v), p + 0.5))]
}

test_that("insert vector entry works", {
  v <- 1:3
  x <- 0
  expect_equal(
    insert_vector_entry(v, x, 0),
    c(0, 1, 2, 3)
  )
  expect_equal(
    insert_vector_entry(v, x, 1),
    c(1, 0, 2, 3)
  )
  expect_equal(
    insert_vector_entry(v, x, 2),
    c(1, 2, 0, 3)
  )
  expect_equal(
    insert_vector_entry(v, x, 3),
    c(1, 2, 3, 0)
  )
  expect_equal(
    insert_vector_entry(v, x, 0:3),
    c(0, 1, 0, 2, 0, 3, 0)
  )
  expect_equal(
    insert_vector_entry(integer(), integer(), integer()),
    integer()
  )
})

#' Handling of an unexpected error
#'
#' @description
#' This function reacts to an unexpected error by throwing an error and linking
#' to an issue site with the request to submit an issue.
#'
#' @param msg \[`character(1)`\]\cr
#' An error message.
#'
#' @param issue_link \[`character(1)`\]\cr
#' The URL to an issues site.
#'
#' @return
#' No return value, but it throws an error.
#'
#' @keywords packaging
#' @family package helpers
#' @export

unexpected_error <- function(
    msg = "Ups, an unexpected error occured.",
    issue_link = "https://github.com/loelschlaeger/oeli/issues") {
  checkmate::assert_string(msg, min.chars = 1)
  checkmate::assert_string(
    issue_link,
    pattern = "^https://github.com/[[:alpha:]]*/[[:alpha:]]*/issues$"
  )
  cli::cli_abort(
    c(
      msg,
      "i" = paste(
        "Please submit an issue here:",
        cli::style_hyperlink(issue_link, issue_link)
      )
    ),
    call = NULL
  )
}

test_that("handling of an unknown error works", {
  expect_error(
    unexpected_error(issue_link = "bad_link"),
    "Assertion on 'issue_link' failed"
  )
  expect_error(
    unexpected_error(
      msg = "upps",
      issue_link = "https://github.com/loelschlaeger/oeli/issues"
    ),
    "upps"
  )
})

#' Sample correlation matrix
#'
#' @description
#' This function samples a correlation matrix by sampling a covariance matrix
#' from an inverse Wishart distribution and transforming it to a correlation
#' matrix.
#'
#' @param dim \[`integer(1)`\]\cr
#' The dimension.
#'
#' @param df \[`integer(1)`\]\cr
#' The degrees of freedom of the inverse Wishart distribution greater or equal
#' \code{dim}.
#'
#' @param scale \[`matrix()`\]\cr
#' The scale covariance matrix of the inverse Wishart distribution of dimension
#' \code{dim}.
#'
#' @return
#' A correlation \code{matrix}.
#'
#' @keywords simulation
#' @family matrix helpers
#' @export
#'
#' @examples
#' sample_correlation_matrix(dim = 3)

sample_correlation_matrix <- function(
    dim, df = dim, scale = diag(dim)
  ) {
  cov <- sample_covariance_matrix(
    dim = dim, df = df, scale = scale, diag = FALSE
  )
  D <- diag(1 / sqrt(diag(cov)))
  cor <- D %*% cov %*% D
  diag(cor) <- 1
  cor[cor > 1] <- 1
  cor[cor < -1] <- -1
  assert_correlation_matrix(cor)
  return(cor)
}

#' General system level information
#'
#' @description
#' This function returns a \code{list} of general system level information.
#'
#' @return
#' A \code{list} with elements:
#' \itemize{
#'   \item \code{maschine}, the model name of the device
#'   \item \code{cores}, the number of cores
#'   \item \code{ram}, the size of the RAM
#'   \item \code{os}, the operating system
#'   \item \code{rversion}, the R version used
#' }
#'
#' @keywords validation
#' @family package helpers
#' @export
#'
#' @examples
#' system_information()

system_information <- function() {

  ### get information
  cpu <- try(benchmarkme::get_cpu(), silent = TRUE)
  if (inherits(cpu, "try-error") || !checkmate::test_list(cpu)) {
    cpu <- list()
  }
  machine <- cpu$model_name
  if (!checkmate::test_string(machine, na.ok = FALSE, null.ok = FALSE)) {
    machine <- NA_character_
  }
  cores <- cpu$no_of_cores
  if (!checkmate::test_int(cores, lower = 0, na.ok = FALSE, null.ok = FALSE)) {
    cores <- NA_integer_
  }
  ram <- try(benchmarkme::get_ram(), silent = TRUE)
  if (!checkmate::test_class(ram, "ram")) {
    ram <- NA
  }
  os <- .Platform$OS.type
  if (!checkmate::test_string(os, na.ok = FALSE, null.ok = FALSE)) {
    os <- NA_character_
  }
  rversion <- try(getRversion(), silent = TRUE)
  if (!inherits(rversion, "R_system_version")) {
    rversion <- NA
  }

  ### build and return output
  list(
    "machine" = machine,
    "cores" = cores,
    "ram" = ram,
    "os" = os,
    "rversion" = rversion
  )
}

test_that("system information can be obtained", {
  x <- system_information()
  checkmate::expect_list(x)
  checkmate::expect_names(
    names(x), identical.to = c("machine", "cores", "ram", "os", "rversion")
  )
})

#' Cholesky root of covariance matrix
#'
#' @description
#' These functions compute the Cholesky root elements of a covariance matrix
#' and, conversely, build a covariance matrix from its Cholesky root elements.
#'
#' @param cov \[`matrix()`\]\cr
#' A covariance matrix.
#'
#' It can also be the zero matrix, in which case the Cholesky root is defined as
#' the zero matrix.
#'
#' @param chol \[`numeric()`\]\cr
#' Cholesky root elements.
#'
#' @param unique \[`logical(1)`\]\cr
#' Ensure that the Cholesky decomposition is unique by restricting the diagonal
#' elements to be positive?
#'
#' @return
#' For \code{\link{cov_to_chol}} a \code{numeric} \code{vector} of Cholesky root
#' elements.
#'
#' For \code{\link{chol_to_cov}} a covariance \code{matrix}.
#'
#' @keywords transformation
#' @family matrix helpers
#' @export
#'
#' @examples
#' cov <- sample_covariance_matrix(4)
#' chol <- cov_to_chol(cov)
#' all.equal(cov, chol_to_cov(chol))

cov_to_chol <- function(cov, unique = TRUE) {
  if (is.matrix(cov) && nrow(cov) == 0 && ncol(cov) == 0) {
    return(numeric())
  }
  input_check_response(
    check_covariance_matrix(cov),
    "cov"
  )
  input_check_response(
    checkmate::check_flag(unique),
    "unique"
  )
  if (all(as.vector(cov) == 0)) {
    cov_chol <- cov
  } else {
    cov_chol <- t(chol(cov))
  }
  chol <- cov_chol[lower.tri(cov_chol, diag = TRUE)]
  if (unique) {
    return(unique_chol(chol))
  } else {
    return(chol)
  }
}

#' @rdname cov_to_chol
#' @export

chol_to_cov <- function(chol) {
  input_check_response(
    check_numeric_vector(chol, any.missing = FALSE, finite = TRUE),
    "chol"
  )
  if (length(chol) == 0) {
    return(matrix(0, 0, 0))
  }
  dim <- -0.5 + sqrt(0.25 + 2 * length(chol))
  input_check_response(
    checkmate::check_count(dim, positive = TRUE),
    prefix = "Value {.code -0.5 + sqrt(0.25 + 2 * length(chol))} is bad:"
  )
  cov <- matrix(0, dim, dim)
  cov[lower.tri(cov, diag = TRUE)] <- chol
  cov %*% t(cov)
}

#' @rdname cov_to_chol
#' @export

unique_chol <- function(chol) {
  input_check_response(
    check_numeric_vector(chol, any.missing = FALSE, finite = TRUE),
    "chol"
  )
  if (length(chol) == 0) {
    return(numeric(0))
  }
  dim <- -0.5 + sqrt(0.25 + 2 * length(chol))
  input_check_response(
    checkmate::check_count(dim, positive = TRUE),
    prefix = "Length of {.var chol} is bad:"
  )
  root <- matrix(0, dim, dim)
  root[lower.tri(root, diag = TRUE)] <- chol
  diagonal <- diag(root)
  for (i in seq_along(diagonal)) {
    if (diagonal[i] < 0) root[, i] <- -root[, i]
  }
  root[lower.tri(root, diag = TRUE)]
}

test_that("Cholesky decomposition works", {
  cov <- sample_covariance_matrix(5)
  chol <- cov_to_chol(cov)
  expect_true(all.equal(cov, chol_to_cov(chol)))
  chol <- cov_to_chol(cov)
  expect_true(all.equal(cov, chol_to_cov(chol)))
  expect_equal(unique_chol(numeric()), numeric())
})

test_that("Edge cases in Cholesky decomposition work", {
  expect_equal(cov_to_chol(matrix(0, 0, 0)), numeric())
  expect_equal(cov_to_chol(matrix(0, 2, 2)), c(0, 0, 0))
  expect_equal(chol_to_cov(numeric()), matrix(0, 0, 0))
})

#' Storage R6 Object
#'
#' @description
#' Provides a simple indexing interface for list elements based on R6.
#' Basically, it allows to store items in a list and to regain them based on
#' identifiers defined by the user.
#'
#' @section Setting identifiers:
#' An identifier is a \code{character}, typically a binary property. Identifiers
#' can be negated by placing an exclamation mark (\code{"!"}) in front of them.
#' Identifiers that have been assigned to other elements previously do not need
#' to be specified again for new elements; instead, a default value can be used.
#' This default value can be defined either globally for all cases (via the
#' \code{$missing_identifier} field) or separately for each specific case (via
#' the method argument).
#'
#' @section User confirmation:
#' If desired, the user can be asked for confirmation when adding, extracting,
#' or removing elements using identifiers. This behavior can be set globally
#' through the \code{$confirm} field or customized separately for each specific
#' case via the method argument.
#'
#' @param identifier \[`character()`\]\cr
#' Pne or more identifiers (the identifier \code{"all"} is reserved to select
#' all elements).
#'
#' @param ids \[`integer()`\]\cr
#' One or more ids.
#'
#' @param logical \[`character(1)`\]\cr
#' In the case that multiple identifiers are selected, how should they be
#' combined? Options are:
#' - \code{"and"} (the default): the identifiers are combined with logical and
#'   (all identifiers must be `TRUE`)
#' - \code{"or"}: the identifiers are combined with logical or (at least one
#'   identifier must be `TRUE`)
#'
#' @param confirm \[`logical(1)`\]\cr
#' Prompted for confirmation?
#'
#' @param missing_identifier \[`logical(1)` | NA\]\cr
#' The value for not specified identifiers.
#'
#' @return
#' The output depends on the method:
#' - \code{$new()} returns a \code{Storage} object.
#' - \code{$add()}, \code{$remove()}, and \code{$print()} invisibly return the
#'   \code{Storage} object (to allow for method chaining)
#' - \code{$get()} returns the requested element(s)
#' - \code{$number()} returns an \code{integer}
#' - \code{$indices()} return an \code{integer} \code{vector}
#'
#' @keywords indexing
#' @family package helpers
#' @export
#'
#' @examples
#' ### 1. Create a `Storage` object:
#' my_storage <- Storage$new()
#'
#' # 2. Add elements along with identifiers:
#' my_storage$
#'   add(42, c("number", "rational"))$
#'   add(pi, c("number", "!rational"))$
#'   add("fear of black cats", c("text", "!rational"))$
#'   add("wearing a seat belt", c("text", "rational"))$
#'   add(mean, "function")
#'
#' # 3. What elements are stored?
#' print(my_storage)
#'
#' # 4. Extract elements based on identifiers:
#' my_storage$get("rational")
#' my_storage$get("!rational")
#' my_storage$get(c("text", "!rational"))
#' my_storage$get("all") # get all elements
#' my_storage$get(c("text", "!text"))
#' my_storage$get(c("text", "!text"), logical = "or")
#'
#' # 5. Extract elements based on ids:
#' my_storage$get(ids = 4:5)
#' my_storage$get(ids = 4:5, id_names = TRUE) # add the ids as names

Storage <- R6::R6Class(
  classname = "Storage",
  lock_class = TRUE,
  cloneable = FALSE,
  public = list(

    #' @description
    #' Initializing a \code{Storage} object.
    initialize = function() {
      ### elements are saved in a list
      private$elements <- list()

      ### ids are saved in a data.frame with identifiers in columns
      private$ids <- data.frame()
    },

    #' @description
    #' Adding an element.
    #' @param x \[`any`\]\cr
    #' An object to be saved.
    add = function(x, identifier, confirm = interactive() & self$confirm,
                   missing_identifier = self$missing_identifier) {
      ### input checks
      if (missing(x)) {
        stop("please specify the object 'x' to be added", call. = FALSE)
      }
      if (missing(identifier)) {
        stop("please specify at least one entry for 'identifier'", call. = FALSE)
      }
      private$check_input(
        identifier = identifier, confirm = confirm,
        missing_identifier = missing_identifier
      )
      if (any(c("all", "!all") %in% identifier)) {
        stop("the identifier \"all\" is reserved", call. = FALSE)
      }

      ### inform user about action and request confirmation
      if (confirm) {
        private$user_confirm(
          action = paste0("add a <", class(x), ">"),
          identifier = identifier, missing_identifier = missing_identifier,
          complete = TRUE
        )
      }

      ### add element
      private$add_element(x, identifier, missing_identifier)
      invisible(self)
    },

    #' @description
    #' Getting elements.
    #' @param id_names \[`logical(1)`\]\cr
    #' Name the elements according to their ids?
    get = function(
      identifier = character(), ids = integer(), logical = "and",
      confirm = interactive() & self$confirm,
      missing_identifier = self$missing_identifier,
      id_names = FALSE
    ) {
      ### input checks
      private$check_input(
        identifier = identifier, confirm = confirm, ids = ids,
        missing_identifier = missing_identifier, logical = logical
      )
      if (length(identifier) == 0 && length(ids) == 0) {
        if (!self$hide_warnings) {
          warning(
            "please specify either 'identifier' or 'ids'",
            call. = FALSE, immediate. = FALSE
          )
        }
        return(list())
      }

      ### inform user about action and request confirmation
      if (confirm) {
        private$user_confirm(
          action = "extract", identifier = identifier, ids = ids,
          missing_identifier = missing_identifier, complete = FALSE,
          logical = logical
        )
      }

      ### get elements
      private$get_element(
        identifier = identifier, ids = ids, id_names = id_names,
        logical = logical
      )
    },

    #' @description
    #' removing elements
    #' @param shift_ids \[`logical(1)`\]\cr
    #' Shift ids when in-between elements are removed?
    remove = function(
      identifier = character(), ids = integer(), logical = "and",
      confirm = interactive() & self$confirm,
      missing_identifier = self$missing_identifier, shift_ids = TRUE
    ) {
      ### input checks
      private$check_input(
        identifier = identifier, confirm = confirm, ids = ids,
        missing_identifier = missing_identifier, logical = logical
      )
      checkmate::assert_flag(shift_ids)
      if (length(identifier) == 0 && length(ids) == 0) {
        if (!self$hide_warnings) {
          warning(
            "please specify either 'identifier' or 'ids'",
            call. = FALSE, immediate. = FALSE
          )
        }
        return(list())
      }

      ### inform user about action and request confirmation
      if (confirm) {
        private$user_confirm(
          action = "remove", identifier = identifier, ids = ids,
          missing_identifier = missing_identifier, complete = FALSE,
          logical = logical
        )
      }

      ### remove elements
      if (length(identifier) > 0 && length(ids) > 0) {
        ids <- private$merge_ids(
          ids,
          private$get_ids(identifier = identifier, logical = logical),
          logical = logical
        )
      } else if (length(identifier) > 0) {
        ids <- private$get_ids(identifier = identifier, logical = logical)
      }
      if (shift_ids) {
        private$elements[ids] <- NULL
        private$ids <- private$ids[-ids, ]
      } else {
        private$elements[ids] <- list(NULL)
        private$ids[ids, ] <- NA
      }
      invisible(self)
    },

    #' @description
    #' Computing the number of identified elements.
    number = function(
      identifier = "all", missing_identifier = self$missing_identifier,
      logical = "and", confirm = FALSE
    ) {
      ### input checks
      private$check_input(
        identifier = identifier, missing_identifier = missing_identifier,
        confirm = confirm, logical = logical
      )

      ### inform user about action and request confirmation
      if (confirm) {
        private$user_confirm(
          action = "count", identifier = identifier,
          missing_identifier = missing_identifier, complete = FALSE,
          logical = logical
        )
      }

      ### perform action
      length(self$indices(identifier, confirm = confirm, logical = logical))
    },

    #' @description
    #' Returning indices based on defined identifiers.
    indices = function(
      identifier = "all", logical = "and",
      confirm = interactive() & self$confirm)
    {
      ### input checks
      private$check_input(
        identifier = identifier, logical = logical, confirm = confirm
      )

      ### inform user about action and request confirmation
      if (confirm) {
        private$user_confirm(
          action = "get indices of",
          identifier = identifier, missing_identifier = missing_identifier,
          complete = FALSE, logical = logical
        )
      }

      ### perform action
      private$get_ids(identifier = identifier, logical = logical)
    },

    #' @description
    #' Printing details of the saved elements.
    #' @param ...
    #' Currently not used.
    print = function(...) {
      nelements <- length(private$elements)
      if (nelements > 0) {
        cat("number of elements:", self$number(), "\n")
        cat("identifiers used:", self$identifier, "\n")
      } else {
        cat("no elements saved yet")
      }
      invisible(self)
    }
  ),
  active = list(

    #' @field identifier \[`character()`\]\cr
    #' The identifiers used.
    identifier = function(value) {
      if (missing(value)) {
        colnames(private$ids)
      } else {
        stop("read only")
      }
    },

    #' @field confirm \[`logical(1)`\]\cr
    #' The default value for confirmations.
    confirm = function(value) {
      if (missing(value)) {
        private$confirm_default
      } else {
        checkmate::assert_flag(value)
        private$confirm_default <- value
      }
    },

    #' @field missing_identifier \[`logical(1)`\]\cr
    #' The default value for not specified identifiers.
    missing_identifier = function(value) {
      if (missing(value)) {
        private$missing_default
      } else {
        checkmate::assert_flag(value, na.ok = TRUE)
        private$missing_default <- value
      }
    },

    #' @field hide_warnings \[`logical(1)`\]\cr
    #' Hide warnings (for example if unknown identifiers are selected)?
    hide_warnings = function(value) {
      if (missing(value)) {
        private$.hide_warnings
      } else {
        checkmate::assert_flag(value, na.ok = TRUE)
        private$.hide_warnings <- value
      }
    }
  ),
  private = list(
    elements = NULL,
    ids = NULL,
    confirm_default = FALSE,
    missing_default = NA,
    .hide_warnings = FALSE,
    check_input = function(identifier = NULL, confirm = NULL, ids = NULL, missing_identifier = NULL,
                           logical = NULL) {
      ### check 'identifier' input
      if (!is.null(identifier)) {
        checkmate::assert_character(identifier, any.missing = FALSE)
      }

      ### check 'confirm' input
      checkmate::assert_flag(confirm)

      ### check 'ids' input
      if (!is.null(ids)) {
        checkmate::assert_integerish(ids, lower = 1, any.missing = FALSE)
      } else if (length(identifier) == 0) {
        stop("need at least one identifier", call. = FALSE)
      }

      ### check 'logical' input
      if (!is.null(logical)) {
        checkmate::assert_choice(logical, choices = c("and", "or"))
      }

      ### check 'missing_identifier' input
      if (!is.null(missing_identifier)) {
        checkmate::assert_flag(missing_identifier, na.ok = TRUE)
      }
    },
    check_identifier_known = function(identifier, logical) {
      checkmate::assert_character(identifier, any.missing = FALSE, min.len = 1)
      checkmate::assert_choice(logical, c("and", "or"))
      identifier_translated <- names(private$translate_identifier(identifier))
      unknown <- which(!identifier_translated %in% c("all", self$identifier))
      if (length(unknown) > 0) {
        if (!self$hide_warnings) {
          warning(
            paste0(
              "I do not know the identifier(s) '",
              paste(identifier[unknown], collapse = "', '"), "'."
            ),
            call. = FALSE, immediate. = FALSE
          )
        }
        if (logical == "and") {
          if (!self$hide_warnings) {
            warning(
              "Because 'logical = \"and\", no element is selected.",
              call. = FALSE, immediate. = FALSE
            )
          }
          return(integer())
        } else if (logical == "or") {
          if (!self$hide_warnings) {
            warning(
              "Because 'logical = \"or\", I will ignore them.",
              call. = FALSE, immediate. = FALSE
            )
          }
          identifier <- identifier[-unknown]
        } else {
          unknown_error()
        }
      }
      return(identifier)
    },
    user_confirm = function(action = character(), identifier = character(), ids = integer(),
                            missing_identifier = self$missing_identifier, complete = TRUE,
                            logical = "and") {
      checkmate::assert_flag(complete)
      cat("You are about to", action)
      if (any(identifier == "all")) {
        cat(" all elements.\n")
      } else {
        cat(" element(s)")
        if (length(ids) > 0) {
          if (length(ids) == 1) {
            cat(" with id\n")
          } else {
            cat(" with ids\n")
          }
          print(ids)
          if (length(identifier) > 0) {
            cat("and ")
          }
        } else {
          cat(" ")
        }
        if (length(identifier) > 0) {
          if (length(identifier) == 1) {
            cat("with identifier:\n")
          } else {
            if (logical == "and") {
              cat("with all these identifiers:\n")
            } else if (logical == "or") {
              cat("with at least one of these identifiers:\n")
            } else {
              unknown_error()
            }
          }
          identifier_bool <- private$translate_identifier(identifier)
          if (complete) {
            identifier_bool <- private$complete_identifier_bool(
              identifier_bool = identifier_bool,
              missing_identifier = missing_identifier
            )
          }
          print(identifier_bool)
        }
      }
      confirmation <- user_confirm("Is that alright?", default = TRUE)
      if (!confirmation) {
        stop("Okay, then no.", call. = FALSE)
        return(invisible(self))
      }
    },
    add_identifier = function(identifier, missing_identifier) {
      checkmate::assert_character(identifier, any.missing = FALSE, min.len = 1)
      stopifnot(length(intersect(identifier, self$identifier)) == 0)
      if (nrow(private$ids) == 0) {
        private$ids[1, ] <- NA
        private$ids[, identifier] <- NA
        private$ids <- private$ids[0, ]
      } else {
        private$ids[, identifier] <- missing_identifier
      }
    },
    add_element = function(x, identifier, missing_identifier) {
      identifier_bool <- private$translate_identifier(identifier)
      new_identifier <- setdiff(names(identifier_bool), self$identifier)
      if (length(new_identifier) > 0) {
        private$add_identifier(new_identifier, missing_identifier)
      }
      id <- private$next_id()
      private$elements[[id]] <- x
      private$ids[id, ] <- self$missing_identifier
      private$ids[id, names(identifier_bool)] <- identifier_bool
    },
    next_id = function() {
      length(private$elements) + 1
    },
    get_element = function(identifier, ids, id_names, logical) {
      checkmate::assert_flag(id_names)
      if (length(identifier) > 0 && length(ids) > 0) {
        ids <- private$merge_ids(
          ids,
          private$get_ids(identifier = identifier, logical = logical),
          logical = logical
        )
      } else if (length(identifier) > 0) {
        ids <- private$get_ids(identifier = identifier, logical = logical)
      }
      structure(
        private$elements[ids],
        names = if (id_names) ids
      )
    },
    translate_identifier = function(identifier) {
      identifier_translated <- logical(0)
      for (value in identifier) {
        if (startsWith(value, "!")) {
          identifier_translated[substring(value, first = 2)] <- FALSE
        } else {
          identifier_translated[value] <- TRUE
        }
      }
      return(identifier_translated)
    },
    complete_identifier_bool = function(identifier_bool, missing_identifier) {
      identifier_bool[setdiff(self$identifier, names(identifier_bool))] <-
        missing_identifier
      return(identifier_bool)
    },
    get_ids = function(identifier, logical) {
      ### check for unknown identifier
      identifier <- private$check_identifier_known(
        identifier = identifier, logical = logical
      )

      ### check for identifier and inverse identifier
      for (i in identifier) {
        i_inverse <- ifelse(startsWith(i, "!"), substring(i, first = 2), paste0("!", i))
        if (i %in% identifier && i_inverse %in% identifier) {
          if (logical == "and") {
            identifier <- integer()
            break
          } else {
            identifier <- identifier[-which(identifier %in% c(i, i_inverse))]
            if (length(identifier) == 0) {
              identifier <- "all"
              break
            }
          }
        }
      }

      ### no identifiers selected
      if (length(identifier) == 0) {
        return(integer())
      }

      ### all identifiers selected
      if ("all" %in% identifier) {
        if (length(identifier) == 1 || logical == "or") {
          return(seq_along(private$elements))
        } else {
          identifier <- identifier[-which(identifier == "all")]
        }
      }

      ### extract ids
      identifier_bool <- private$translate_identifier(identifier)
      comparison <- if (logical == "and") {
        identical
      } else {
        function(x, y) any(x == y)
      }
      ids <- apply(
        private$ids[names(identifier_bool)],
        1,
        comparison,
        identifier_bool
      )
      sort(unique(which(ids)))
    },
    merge_ids = function(..., logical) {
      ids <- list(...)
      if (logical == "and") {
        ids <- Reduce(intersect, ids)
      } else if (logical == "or") {
        ids <- unique(unlist(ids))
      } else {
        unknown_error()
      }
      checkmate::assert_integerish(ids, lower = 1, any.missing = FALSE)
      if (length(ids) == 0) {
        return(integer())
      } else {
        sort(ids)
      }
    }
  )
)

test_that("basic Storage example works", {
  my_Storage <- Storage$new()
  checkmate::expect_r6(my_Storage, "Storage")
  my_Storage$
    add(42, c("number", "rational"))$
    add(pi, c("number", "!rational"))$
    add("fear of black cats", c("text", "!rational"))$
    add("wearing a seatbelt", c("text", "rational"))$
    add(mean, "function")
  expect_equal(
    my_Storage$get("rational"),
    list(42, "wearing a seatbelt")
  )
  expect_length(
    my_Storage$get("!rational"),
    2
  )
  expect_equal(
    my_Storage$get(c("text", "!rational")),
    list("fear of black cats")
  )
  expect_length(
    my_Storage$get(ids = 4:5, id_names = TRUE),
    2
  )
  expect_equal(
    my_Storage$get(c("text", "!text")),
    list()
  )
  expect_equal(
    my_Storage$indices(c("text", "!text"), logical = "or"),
    1:5
  )
  expect_equal(
    my_Storage$indices(c("all", "!rational"), logical = "or"),
    1:5
  )
  expect_equal(
    my_Storage$indices(identifier = c("all", "!rational"), logical = "and"),
    2:3
  )
  expect_equal(
    my_Storage$remove(ids = c(2, 4:5), shift_ids = TRUE)$get("all"),
    list(42, "fear of black cats")
  )
})

test_that("silencing R code works", {
  f <- function() {
    warning("warning")
    message("message")
    cat("cat")
    print("print")
  }
  quiet({
    expect_warning(
      quiet(f(), TRUE, FALSE, FALSE),
      "warning"
    )
    quiet(f(), TRUE, FALSE, TRUE)
    expect_warning(
      quiet(f(), TRUE, TRUE, FALSE),
      "warning"
    )
    quiet(f(), TRUE, TRUE, TRUE)
    expect_warning(
      quiet(f(), FALSE, FALSE, FALSE),
      "warning"
    )
    quiet(f(), FALSE, FALSE, TRUE)
    expect_warning(
      quiet(f(), FALSE, TRUE, FALSE),
      "warning"
    )
    quiet(f(), FALSE, TRUE, TRUE)
  })
})

test_that("dictionary with optimizers works", {
  optimizer_dictionary <- Dictionary$new(
    key_name = "label",
    alias_name = "method",
    value_names = c(
      "algorithm",
      "arg_objective",
      "arg_initial",
      "out_value",
      "out_parameter",
      "direction",
      "arguments",
      "out_elements"
    ),
    value_assert = alist(
      "algorithm" = checkmate::assert_function(),
      "arg_objective" = checkmate::assert_string(),
      "arg_initial" = checkmate::assert_string(),
      "out_value" = checkmate::assert_string(),
      "out_parameter" = checkmate::assert_string(),
      "direction" = checkmate::assert_choice(choices = c("min", "max")),
      "arguments" = checkmate::assert_list(names = "unique"),
      "out_elements" = checkmate::assert_character(any.missing = FALSE, unique = TRUE)
    ),
    allow_overwrite = FALSE,
    keys_reserved = c("custom"),
    alias_choices = c("constrained", "unconstrained")
  )
  expect_snapshot(
    print(optimizer_dictionary)
  )
  expect_s3_class(
    optimizer_dictionary,
    "Dictionary"
  )
  expect_error(
    optimizer_dictionary$add(
      "label" = "custom",
      "method" = c("unconstrained"),
      "algorithm" = stats::nlm,
      "arg_objective" = "f",
      "arg_initial" = "p",
      "out_value" = "minimum",
      "out_parameter" = "min",
      "direction" = "min",
      "arguments" = as.list(formals(stats::nlm)),
      "out_elements" = c("minimum", "estimate", "gradient", "hessian", "code", "iterations")
    ),
    "The key 'custom' must not be used."
  )
  expect_silent(
    optimizer_dictionary$add(
      "label" = "stats::nlm",
      "method" = c("unconstrained"),
      "algorithm" = stats::nlm,
      "arg_objective" = "f",
      "arg_initial" = "p",
      "out_value" = "minimum",
      "out_parameter" = "min",
      "direction" = "min",
      "arguments" = as.list(formals(stats::nlm)),
      "out_elements" = c("minimum", "estimate", "gradient", "hessian", "code", "iterations")
    )$add(
      "label" = "stats::optim",
      "algorithm" = stats::optim,
      "arg_objective" = "fn",
      "arg_initial" = "par",
      "out_value" = "minimum",
      "out_parameter" = "min",
      "direction" = "min",
      "arguments" = as.list(formals(stats::optim)),
      "out_elements" = c("par", "value", "counts", "convergence", "message", "hessian")
    )
  )
  expect_snapshot(
    print(optimizer_dictionary)
  )
  expect_equal(
    optimizer_dictionary$get("stats::nlm", "arg_initial"),
    "p"
  )
  expect_silent(
    optimizer_dictionary$remove("stats::optim")
  )
  expect_equal(
    optimizer_dictionary$keys,
    "stats::nlm"
  )
  expect_equal(
    optimizer_dictionary$alias,
    list(unconstrained = "stats::nlm")
  )
})

#' Dirichlet distribution
#'
#' @description
#' The function `ddirichlet()` computes the density of a Dirichlet distribution.
#'
#' The function `rdirichlet()` samples from a Dirichlet distribution.
#'
#' The functions with suffix `_cpp` perform no input checks, hence are faster.
#'
#' @param x \[`numeric()`\]\cr
#' A probability vector.
#'
#' @param concentration \[`numeric()`\]\cr
#' A concentration vector of the same length as `x`.
#'
#' @param log \[`logical(1)`\]\cr
#' Return the logarithm of the density value?
#'
#' @param n \[`integer(1)`\]\cr
#' The number of samples.
#'
#' @return
#' For `ddirichlet()`: The density value.
#'
#' For `rdirichlet()`: If \code{n = 1} a \code{vector} of length \code{p}, else
#' a \code{matrix} of dimension \code{n} times \code{p} with samples as rows.
#'
#' @keywords distribution
#' @family simulation helpers
#' @export
#'
#' @examples
#' x <- c(0.5, 0.3, 0.2)
#' concentration <- 1:3
#'
#' # compute density
#' ddirichlet(x = x, concentration = concentration)
#' ddirichlet(x = x, concentration = concentration, log = TRUE)
#'
#' # sample
#' rdirichlet(concentration = 1:3)
#' rdirichlet(n = 4, concentration = 1:2)

ddirichlet <- function(x, concentration, log = FALSE) {
  checkmate::assert_numeric(x, lower = 0, upper = 1, any.missing = FALSE)
  stopifnot("'x' must sum up to 1" = sum(x) == 1)
  checkmate::assert_numeric(concentration, lower = 0, any.missing = FALSE, len = length(x))
  checkmate::assert_flag(log)
  ddirichlet_cpp(x, concentration, log)
}

#' @rdname ddirichlet
#' @export

rdirichlet <- function(n = 1, concentration) {
  checkmate::assert_int(n, lower = 1)
  checkmate::assert_vector(concentration, strict = TRUE)
  checkmate::assert_numeric(concentration, any.missing = FALSE, lower = 0)
  dim <- length(concentration)
  out <- replicate(n = n, rdirichlet_cpp(concentration), simplify = TRUE)
  if (n == 1) {
    drop(out)
  } else if (dim == 1) {
    as.matrix(out)
  } else {
    t(out)
  }
}

test_that("Dirichlet density can be computed", {
  x <- c(0.5, 0.3, 0.2)
  concentration <- 1:3
  factor <- 1000
  expect_equal(
    round(ddirichlet(x = x, concentration = concentration, log = FALSE) * factor) / factor,
    0.72
  )
  expect_equal(
    round(ddirichlet(x = x, concentration = concentration, log = TRUE) * factor) / factor,
    -0.329
  )
  expect_error(
    ddirichlet(x = c(0.5, 0.3, 0.3), concentration = concentration),
    "'x' must sum up to 1"
  )
})

test_that("Dirichlet can be drawn", {
  concentration <- 1:3
  expect_length(rdirichlet(concentration = concentration), 3)
  checkmate::expect_matrix(
    rdirichlet(n = 5, concentration = concentration),
    ncols = 3, nrows = 5
  )
  expect_error(
    rdirichlet(concentration = diag(3)),
    "Assertion on 'concentration' failed: Must be of type 'vector', not 'matrix'."
  )
  checkmate::expect_matrix(
    rdirichlet(n = 5, concentration = 1),
    ncols = 1, nrows = 5
  )
})

#' Check covariance matrix
#'
#' @description
#' These functions check whether the input fulfills the properties of a
#' covariance matrix.
#'
#' @inheritParams checkmate::check_matrix
#'
#' @param dim \[`integer(1)`\]\cr
#' The matrix dimension.
#'
#' @param tolerance \[`numeric(1)`\]\cr
#' A non-negative tolerance value.
#'
#' @return
#' Same as documented in \code{\link[checkmate]{check_matrix}}.
#'
#' @keywords validation
#' @family matrix helpers
#' @export
#'
#' @examples
#' M <- matrix(c(1, 2, 3, 2, 1, 2, 3, 2, 1), nrow = 3)
#' check_covariance_matrix(M)
#' test_covariance_matrix(M)
#' \dontrun{
#' assert_covariance_matrix(M)
#' }

check_covariance_matrix <- function(
    x, dim = NULL, tolerance = sqrt(.Machine$double.eps)
  ) {
  input_check_response(
    check = checkmate::check_number(tolerance, lower = 0),
    var_name = "tolerance"
  )
  res <- checkmate::check_matrix(x, mode = "numeric")
  if (!isTRUE(res)) {
    return(res)
  }
  if (nrow(x) != ncol(x)) {
    return("Must be square")
  }
  if (any(is.na(x))) {
    return("Must not have NA values")
  }
  if (any(!is.finite(x))) {
    return("Must not have infinite values")
  }
  if (any(abs(x - t(x)) > tolerance)) {
    return("Must be symmetric")
  }
  if (any(eigen(x)$value < -tolerance)) {
    return("Must have positive eigenvalues only")
  }
  if (!is.null(dim)) {
    input_check_response(
      check = checkmate::check_count(dim, positive = TRUE),
      var_name = "dim"
    )
    if (nrow(x) != dim) {
      return(paste("Must be of dimension", dim))
    }
  }
  return(TRUE)
}

#' @rdname check_covariance_matrix
#' @inheritParams checkmate::assert_matrix
#' @export

assert_covariance_matrix <- checkmate::makeAssertionFunction(
  check_covariance_matrix
)

#' @rdname check_covariance_matrix
#' @inheritParams checkmate::test_matrix
#' @export

test_covariance_matrix <- checkmate::makeTestFunction(
  check_covariance_matrix
)

test_that("covariance matrix can be sampled", {
  expect_true(
    check_covariance_matrix(sample_covariance_matrix(3))
  )
  expect_true(
    check_covariance_matrix(sample_covariance_matrix(3, diag = TRUE))
  )
})

#' Deleting \code{data.frame} columns
#'
#' @description
#' This function deletes columns of a \code{data.frame} by name.
#'
#' @param df \[`data.frame`\]\cr
#' A \code{data.frame}.
#'
#' @param column_names \[`character()`\]\cr
#' The name(s) of column(s) of \code{df} to delete.
#'
#' @return
#' The input \code{df} without the columns defined by \code{column_names}.
#'
#' @keywords transformation
#' @family data.frame helpers
#' @export
#'
#' @examples
#' df <- data.frame("label" = c("A", "B"), "number" = 1:10)
#' delete_columns_data.frame(df = df, column_names = "label")
#' delete_columns_data.frame(df = df, column_names = "number")
#' delete_columns_data.frame(df = df, column_names = c("label", "number"))

delete_columns_data.frame <- function(df, column_names) {
  input_check_response(
    check = checkmate::check_data_frame(df),
    var_name = "df"
  )
  input_check_response(
    check = checkmate::check_names(column_names, subset.of = names(df)),
    var_name = "column_names"
  )
  df[column_names] <- NULL
  return(df)
}

test_that("can delete data.frame columns", {
  df <- data.frame("label" = c("A", "B"), "number" = 1:10)
  expect_equal(
    delete_columns_data.frame(df = df, column_names = "label"),
    structure(list(number = 1:10), row.names = c(NA, -10L), class = "data.frame")
  )
  expect_equal(
    delete_columns_data.frame(df = df, column_names = "number"),
    structure(
      list(label = c("A", "B", "A", "B", "A", "B", "A", "B", "A", "B")),
      row.names = c(NA, -10L), class = "data.frame"
    )
  )
  expect_equal(
    delete_columns_data.frame(df = df, column_names = c("label", "number")),
    structure(list(), names = character(0), row.names = c(NA, -10L), class = "data.frame")
  )
})

#' Argument matching
#'
#' @description
#' This function matches function arguments and is a modified version of
#' \code{\link[base]{match.arg}}.
#'
#' @param arg \[`character()`\]\cr
#' The function argument.
#'
#' @param choices \[`character()`\]\cr
#' Allowed values for \code{arg}.
#'
#' @param several.ok \[`logical(1)`\]\cr
#' Is \code{arg} allowed to have more than one element?
#'
#' @param none.ok \[`logical(1)`\]\cr
#' Is \code{arg} allowed to have zero elements?
#'
#' @return
#' The un-abbreviated version of the exact or unique partial match if there is
#' one. Otherwise, an error is signaled if \code{several.ok} is \code{FALSE}
#' or \code{none.ok} is \code{FALSE}.
#' When \code{several.ok} is \code{TRUE} and (at least) one element of
#' \code{arg} has a match, all un-abbreviated versions of matches are returned.
#' When \code{none.ok} is \code{TRUE} and \code{arg} has zero elements,
#' \code{character(0)} is returned.
#'
#' @export
#' @keywords packaging
#' @family package helpers

match_arg <- function(arg, choices, several.ok = FALSE, none.ok = FALSE) {
  checkmate::assert_character(arg)
  checkmate::assert_character(choices)
  checkmate::assert_flag(several.ok)
  checkmate::assert_flag(none.ok)
  arg_name <- deparse(substitute(arg))
  if (!several.ok && length(arg) > 1L) {
    cli::cli_abort(
      "{.var {arg_name}} must be of length 1.",
      call = NULL
    )
  }
  if (length(arg) == 0L) {
    if (none.ok) {
      return(character(0))
    } else {
      cli::cli_abort(
        "{.var {arg_name}} must be of length greater or equal 1.",
        call = NULL
      )
    }
  }
  i <- pmatch(arg, choices, nomatch = 0, duplicates.ok = TRUE)
  if (all(i == 0L)) {
    cli::cli_abort(
      "{.var {arg_name}} {ifelse(none.ok, 'can', 'must')} be one
      {ifelse(several.ok, 'or more', '')} of {.val {choices}}.",
      call = NULL
    )
  }
  i <- i[i > 0L]
  choices[i]
}

test_that("Argument matching works", {
  expect_equal(
    match_arg("A", "A"),
    "A"
  )
  expect_equal(
    match_arg("A", LETTERS),
    "A"
  )
  expect_error(
    match_arg(c("A", "B"), LETTERS),
    "must be of length 1"
  )
  expect_equal(
    match_arg("lo", c("loooong", "else")),
    "loooong"
  )
  expect_equal(
    match_arg(c("A", "B"), LETTERS, several.ok = TRUE),
    c("A", "B")
  )
  expect_error(
    match_arg(character(), LETTERS),
    "greater or equal 1"
  )
  expect_equal(
    match_arg(character(), LETTERS, none.ok = TRUE),
    character()
  )
  expect_error(
    match_arg("bad", LETTERS),
    "must be one of"
  )
})

#' Map indices
#'
#' @description
#' This function maps indices from an input vector to corresponding sequences of
#' grouped indices. Each element from the input specifies a group to be mapped
#' from the sequence, determined by the grouping size `n`.
#'
#' @param indices \[`integer()`\]\cr
#' An index vector, where each element specifies a group to be mapped from the
#' sequence.
#'
#' @param n \[`integer`\]\cr
#' The size of each group of consecutive indices.
#'
#' @return
#' An `integer` `vector`, containing the mapped indices according to the
#' specified group size.
#'
#' @details
#' This function is useful when working with indices arranged in fixed-size
#' groups, where each group can be referenced by a single index. For example, if
#' indices are structured in chunks of 3, calling this function with `n = 3`
#' will map the corresponding groups of 3 consecutive indices for the given
#' input indices, see the examples.
#'
#' @export
#' @keywords indexing
#' @family vector helpers
#'
#' @examples
#' # Example: Map indices based on groups of 3
#' map_indices(c(1, 3, 5), 3)

map_indices <- function(indices, n) {
  input_check_response(
    check = checkmate::check_integerish(indices, lower = 1, any.missing = FALSE),
    var_name = "indices"
  )
  input_check_response(
    check = checkmate::check_count(n),
    var_name = "n"
  )
  if (length(indices) == 0 || n == 0) return(integer(0))
  unlist(lapply(indices, function(i) ((i - 1) * n + 1):(i * n)))
}

test_that("mapping indices works", {
  expect_equal(
    map_indices(c(1, 3, 5), 0),
    integer()
  )
  expect_equal(
    map_indices(numeric(), 3),
    integer()
  )
  expect_equal(
    map_indices(c(1, 3, 5), 1),
    c(1L, 3L, 5L)
  )
  expect_equal(
    map_indices(c(1, 3, 5), 3),
    c(1L, 2L, 3L, 7L, 8L, 9L, 13L, 14L, 15L)
  )
})

#' Check probability vector
#'
#' @description
#' These functions check whether the input fulfills the properties of a
#' probability matrix.
#'
#' @param tolerance \[`numeric(1)`\]\cr
#' A non-negative tolerance value.
#'
#' @inheritParams checkmate::check_numeric
#'
#' @return
#' Same as documented in \code{\link[checkmate]{check_numeric}}.
#'
#' @keywords validation
#' @family vector helpers
#' @export
#'
#' @examples
#' p <- c(0.2, 0.3, 0.6)
#' check_probability_vector(p)
#' test_probability_vector(p)
#' \dontrun{
#' assert_probability_vector(p)
#' }

check_probability_vector <- function(
    x, len = NULL, tolerance = sqrt(.Machine$double.eps)
) {
  input_check_response(
    check = checkmate::check_number(tolerance, lower = 0),
    var_name = "tolerance"
  )
  res <- check_numeric_vector(
    x, any.missing = FALSE, len = len, lower = 0, upper = 1
  )
  if (!isTRUE(res)) {
    return(res)
  }
  if (abs(sum(x) - 1) > tolerance) {
    return("Must add up to 1")
  }
  return(TRUE)
}

#' @rdname check_probability_vector
#' @inheritParams checkmate::assert_atomic_vector
#' @inheritParams checkmate::assert_numeric
#' @export

assert_probability_vector <- checkmate::makeAssertionFunction(
  check_probability_vector
)

#' @rdname check_probability_vector
#' @inheritParams checkmate::assert_atomic_vector
#' @inheritParams checkmate::assert_numeric
#' @export

test_probability_vector <- checkmate::makeTestFunction(
  check_probability_vector
)

#' Check transition probability matrix
#'
#' @description
#' These functions check whether the input is a transition probability matrix.
#'
#' @inheritParams check_covariance_matrix
#'
#' @return
#' Same as documented in \code{\link[checkmate]{check_matrix}}.
#'
#' @keywords validation
#' @family matrix helpers
#' @export
#'
#' @examples
#' T <- matrix(c(0.8,  0.2,  0.1, 0.1,  0.7,  0.4, 0.1,  0.1,  0.6), nrow = 3)
#' check_transition_probability_matrix(T)
#' test_transition_probability_matrix(T)
#' \dontrun{
#' assert_transition_probability_matrix(T)
#' }

check_transition_probability_matrix <- function(
    x, dim = NULL, tolerance = sqrt(.Machine$double.eps)
  ) {
  input_check_response(
    check = checkmate::check_number(tolerance, lower = 0),
    var_name = "tolerance"
  )
  res <- checkmate::check_matrix(x, mode = "numeric")
  if (!isTRUE(res)) {
    return(res)
  }
  if (nrow(x) != ncol(x)) {
    return("Must be square")
  }
  if (any(is.na(x))) {
    return("Must not have NA values")
  }
  if (any(!is.finite(x))) {
    return("Must not have infinite values")
  }
  if (any(x < 0 | x > 1)) {
    return("Must have values between 0 and 1")
  }
  if (any(abs(rowSums(x) - 1) > tolerance)) {
    return("Must have row sums equal to 1")
  }
  if (!is.null(dim)) {
    checkmate::assert_count(dim, positive = TRUE)
    if (nrow(x) != dim) {
      return(paste("Must be of dimension", dim))
    }
  }
  return(TRUE)
}

#' @rdname check_transition_probability_matrix
#' @inheritParams checkmate::assert_matrix
#' @export

assert_transition_probability_matrix <- checkmate::makeAssertionFunction(
  check_transition_probability_matrix
)

#' @rdname check_transition_probability_matrix
#' @inheritParams checkmate::test_matrix
#' @export

test_transition_probability_matrix <- checkmate::makeTestFunction(
  check_transition_probability_matrix
)

#' Determine variable name
#'
#' @description
#' This function tries to determine the name of a variable passed to a
#' \code{function}.
#'
#' @param variable \[`any`\]\cr
#' Any object.
#'
#' @param fallback \[`character(1)`\]\cr
#' A fallback name if for some reason the actual variable name
#' (which must be a single \code{character}) cannot be determined.
#'
#' @return
#' A \code{character}, the variable name.
#'
#' @keywords functional
#' @family function helpers
#' @export
#'
#' @examples
#' variable_name(a)
#' f <- function(x) variable_name(x)
#' f(x = a)

variable_name <- function(variable, fallback = "unnamed") {
  argument_name <- deparse(
    eval.parent(substitute(substitute(variable))),
    width.cutoff = 500L
  )
  if (!checkmate::test_string(argument_name)) {
    checkmate::assert_string(fallback)
    argument_name <- fallback
  }
  return(argument_name)
}

test_that("determination of variable name works", {
  expect_equal(
    variable_name(a),
    "a"
  )
  f <- function(x) variable_name(x)
  expect_equal(
    f(a),
    "a"
  )
  expect_equal(
    variable_name(function(x) {
      mean(x)
    }),
    "unnamed"
  )
})

#' Simulate correlated regressor values
#'
#' @description
#' This function simulates regressor values from various marginal distributions
#' with custom correlations.
#'
#' @param labels \[`character()`\]\cr
#' Unique labels for the regressors.
#'
#' @param n \[`integer(1)`\]\cr
#' The number of values per regressor.
#'
#' @param marginals \[`list()`\]\cr
#' Optionally marginal distributions for regressors. If not specified,
#' standard normal marginal distributions are used.
#'
#' Each list entry must be named according to a regressor label, and the
#' following distributions are currently supported:
#' \describe{
#'   \item{discrete distributions}{
#'     \itemize{
#'       \item Poisson: `list(type = "poisson", lambda = ...)`
#'       \item categorical: `list(type = "categorical", p = c(...))`
#'     }
#'   }
#'   \item{continuous distributions}{
#'     \itemize{
#'       \item normal: `list(type = "normal", mean = ..., sd = ...)`
#'       \item uniform: `list(type = "uniform", min = ..., max = ...)`
#'     }
#'   }
#' }
#'
#' @param correlation \[`matrix()`\]\cr
#' A correlation matrix of dimension \code{length(labels)}, where the
#' \code{(p, q)}-th entry defines the correlation between regressor
#' \code{labels[p]} and \code{labels[q]}.
#'
#' @param verbose \[`logical(1)`\]\cr
#' Print information about the simulated regressors?
#'
#' @return
#' A \code{data.frame} with \code{n} rows and \code{length(labels)} columns.
#'
#' @keywords simulation
#' @family simulation helpers
#' @export
#'
#' @references
#' This function heavily depends on the `{SimMultiCorrData}` package.
#'
#' @examples
#' labels <- c("P", "C", "N1", "N2", "U")
#' n <- 100
#' marginals <- list(
#'   "P" = list(type = "poisson", lambda = 2),
#'   "C" = list(type = "categorical", p = c(0.3, 0.2, 0.5)),
#'   "N1" = list(type = "normal", mean = -1, sd = 2),
#'   "U" = list(type = "uniform", min = -2, max = -1)
#' )
#' correlation <- matrix(
#'   c(1, -0.3, -0.1, 0, 0.5,
#'     -0.3, 1, 0.3, -0.5, -0.7,
#'     -0.1, 0.3, 1, -0.3, -0.3,
#'     0, -0.5, -0.3, 1, 0.1,
#'     0.5, -0.7, -0.3, 0.1, 1),
#'   nrow = 5, ncol = 5
#' )
#' data <- correlated_regressors(
#'   labels = labels, n = n, marginals = marginals, correlation = correlation
#' )
#' head(data)

correlated_regressors <- function(
    labels,
    n = 100,
    marginals = list(),
    correlation = diag(length(labels)),
    verbose = FALSE
  ) {

  ### input checks
  input_check_response(
    checkmate::check_character(
      labels, min.chars = 1, any.missing = FALSE, min.len = 1, unique = TRUE
    ),
    "labels"
  )
  input_check_response(
    checkmate::check_count(n, positive = TRUE),
    "n"
  )
  P <- length(labels)
  input_check_response(
    checkmate::check_list(
      marginals, types = list(), any.missing = FALSE, max.len = P,
      names = "strict"
    ),
    "marginals"
  )
  input_check_response(
    checkmate::check_subset(
      names(marginals), choices = labels, empty.ok = TRUE
    ),
    "marginals",
    prefix = "Element names of input {var_name} are bad:"
  )
  input_check_response(
    check_correlation_matrix(
      correlation, dim = P
    ),
    "correlation"
  )
  correlation_req <- correlation
  input_check_response(
    checkmate::check_flag(verbose),
    "verbose"
  )

  ### draw data
  if (length(marginals) == 0) {

    ### no marginals specified, so draw from multivariate normal
    data <- rmvnorm(n = n, rep(0, P), correlation)
    if (n == 1) {
      data <- matrix(data, nrow = 1)
    }
    colnames(data) <- labels
    marginals_info <- rep("Normal(mean = 0, sd = 1)", P)

  } else {

    ### interpret marginal distributions
    marginals_class <- character(P)
    marginals_info <- character(P)
    M <- matrix(NA, nrow = 6, ncol = 0)
    lam <- numeric()
    marginal <- list()
    size <- numeric()
    prob <- numeric()

    for (p in seq_len(P)) {

      if (labels[p] %in% names(marginals)) {

        ### check structure
        input_check_response(
          checkmate::check_list(
            marginals[labels[p]], names = "strict", any.missing = FALSE
          ),
          "marginals",
          prefix = paste0("Element {.val ", labels[p], "} in input {.var {var_name}} is bad:")
        )
        type <- marginals[[labels[p]]][["type"]]
        input_check_response(
          checkmate::check_choice(
            type,
            c("poisson", "categorical", "normal", "uniform")
          ),
          "marginals",
          prefix = paste0("Element {.val type} of element {.val ", labels[p], "} in input {.var {var_name}} is bad:")
        )
        if (type == "poisson") {

          lam <- c(lam, marginals[[labels[p]]][["lambda"]])
          input_check_response(
            checkmate::check_number(lam, lower = 0, finite = TRUE),
            "marginals",
            prefix = paste0("Element {.val lambda} of element {.val ", labels[p], "} in input {.var {var_name}} is bad:")
          )
          marginals_class[p] <- "pois"
          marginals_info[p] <- paste0("Poisson(lambda = ", lam, ")")

        } else if (type == "categorical") {

          cat_prob_p <- marginals[[labels[p]]][["p"]]
          input_check_response(
            check_probability_vector(cat_prob_p),
            "marginals",
            prefix = paste0("Element {.val p} of element {.val ", labels[p], "} in input {.var {var_name}} is bad:")
          )
          marginal <- c(marginal, list(cumsum(cat_prob_p)[-length(cat_prob_p)]))
          marginals_class[p] <- "cat"
          marginals_info[p] <- paste0(
            "Categorical(p = ", paste(cat_prob_p, collapse = ",") , ")"
          )

        } else if (type == "normal") {

          normal_mean_p <- marginals[[labels[p]]][["mean"]]
          input_check_response(
            checkmate::check_number(normal_mean_p, finite = TRUE),
            "marginals",
            prefix = paste0("Element {.val mean} of element {.val ", labels[p], "} in input {.var {var_name}} is bad:")
          )
          normal_sd_p <- marginals[[labels[p]]][["sd"]]
          input_check_response(
            checkmate::check_number(normal_sd_p, lower = 0, finite = TRUE),
            "marginals",
            prefix = paste0("Element {.val sd} of element {.val ", labels[p], "} in input {.var {var_name}} is bad:")
          )
          M <- cbind(M, SimMultiCorrData::calc_theory(
            Dist = "Gaussian",
            params = c(normal_mean_p, normal_sd_p)
          ))
          marginals_class[p] <- "cont"
          marginals_info[p] <- paste0(
            "Normal(mean = ", normal_mean_p, ", sd = ", normal_sd_p, ")"
          )

        } else if (type == "uniform") {

          uniform_min_p <- marginals[[labels[p]]][["min"]]
          input_check_response(
            checkmate::check_number(uniform_min_p, finite = TRUE),
            "marginals",
            prefix = paste0("Element {.val min} of element {.val ", labels[p], "} in input {.var {var_name}} is bad:")
          )
          uniform_max_p <- marginals[[labels[p]]][["max"]]
          input_check_response(
            checkmate::check_number(uniform_max_p, lower = uniform_min_p, finite = TRUE),
            "marginals",
            prefix = paste0("Element {.val max} of element {.val ", labels[p], "} in input {.var {var_name}} is bad:")
          )
          M <- cbind(M, SimMultiCorrData::calc_theory(
            Dist = "Uniform",
            params = c(uniform_min_p, uniform_max_p)
          ))
          marginals_class[p] <- "cont"
          marginals_info[p] <- paste0(
            "Uniform(min = ", uniform_min_p, ", max = ", uniform_max_p, ")"
          )

        } else {
          unexpected_error()
        }


      } else {

        M <- cbind(M, SimMultiCorrData::calc_theory(
          Dist = "Gaussian", params = c(0, 1)
        ))
        marginals_class[p] <- "cont"
        marginals_info[p] <- "Normal(mean = 0, sd = 1)"

      }

    }

    ### adapt order of correlation
    marginals_class <- factor(marginals_class, levels = c("cat", "cont", "pois"), ordered = TRUE)
    marginals_order <- order(marginals_class)
    ordering_operator <- diag(P)[marginals_order, ]
    correlation <- ordering_operator %*% correlation %*% t(ordering_operator)

    ### make sure 'correlation' is within upper and lower correlation limits
    valid <- try_silent(quiet(SimMultiCorrData::valid_corr(
      k_cat = sum(marginals_class == "cat"),
      k_cont = sum(marginals_class == "cont"),
      k_pois = sum(marginals_class == "pois"),
      k_nb = sum(marginals_class == "nb"),
      method = "Polynomial",
      means =  M[1, ],
      vars =  (M[2, ])^2,
      skews = M[3, ],
      skurts = M[4, ],
      fifths = M[5, ],
      sixths = M[6, ],
      marginal = marginal,
      lam = lam,
      size = size,
      prob = prob,
      rho = correlation,
      seed = NULL
    )))

    ### check for failure
    if (inherits(valid, "fail")) {
      cli::cli_abort(
        "Unable to fit {.var correlation} with {.var marginals}, please check specification.",
        call = NULL
      )
    }

    ### report info about correlation limits
    for (i in 1:(P - 1)) {
      for (j in (i + 1):P) {
        if (correlation[i, j] < valid$L_rho[i, j]) {
          cli::cli_warn(
            "{.code correlation[{i},{j}] = {correlation[i,j]}} not possible
            (too small), increased to {valid$L_rho[i,j]}"
          )
          correlation[i, j] <- correlation[j, i] <- valid$L_rho[i, j]
        }
        if (correlation[i, j] > valid$U_rho[i, j]) {
          cli::cli_warn(
            "{.code correlation[{i},{j}] = {correlation[i,j]}} not possible
            (too large), decreased to {valid$U_rho[i,j]}"
          )
          correlation[i, j] <- correlation[j, i] <- valid$U_rho[i, j]
        }
      }
    }

    ### simulate data from marginal distributions
    sim_out <- quiet(SimMultiCorrData::rcorrvar(
      n = n,
      k_cat = sum(marginals_class == "cat"),
      k_cont = sum(marginals_class == "cont"),
      k_pois = sum(marginals_class == "pois"),
      k_nb = sum(marginals_class == "nb"),
      method = "Polynomial",
      means =  M[1, ],
      vars =  (M[2, ])^2,
      skews = M[3, ],
      skurts = M[4, ],
      fifths = M[5, ],
      sixths = M[6, ],
      marginal = marginal,
      lam = lam,
      size = size,
      prob = prob,
      rho = correlation,
      seed = NULL
    ))

    #### obtain regressors
    data <- cbind(
      sim_out$ordinal_variables,
      sim_out$continuous_variables,
      sim_out$Poisson_variables
    )

    ### adapt order of regressors and add column names
    data <- data[, order(marginals_order)]
    colnames(data) <- labels
  }

  ### print status and return data
  if (isTRUE(verbose)) {
    cat("Marginal distributions for regressors:\n")
    for (p in seq_len(P)) {
      cat("-", labels[p], "~", marginals_info[p], "\n")
    }
    if (n > 1) {
      cat("\nDeviation of empirical correlations from requested:\n")
      print(round(summary(as.numeric(stats::cor(data) - correlation_req)), 4))
    }
  }
  as.data.frame(data)
}

test_that("simulating regressors works", {

  ### multivariate normal regressors
  labels <- c("N1", "N2", "N3")
  n <- 100
  data <- correlated_regressors(labels, n, verbose = FALSE)
  checkmate::expect_data_frame(data, col.names = "strict", nrows = n)

  ### custom marginals
  set.seed(1)
  labels <- c("P", "C", "N1", "N2", "U")
  n <- 100
  marginals <- list(
    "P" = list(type = "poisson", lambda = 1),
    "C" = list(type = "categorical", p = c(0.2, 0.3, 0.5)),
    "N1" = list(type = "normal", mean = -1, sd = 2),
    "U" = list(type = "uniform", min = -2, max = -1)
  )
  correlation <- matrix(
    c(1, 0.44, -0.67, 0.23, 0.17,
      0.44, 1, -0.59, 0.17, -0.21,
      -0.67, -0.59, 1, -0.18, -0.13,
      0.23, 0.17, -0.18, 1, -0.56,
      0.17, -0.21, -0.13, -0.56, 1),
    nrow = 5, ncol = 5
  )
  data <- correlated_regressors(
    labels = labels, n = n, marginals = marginals, correlation = correlation,
    verbose = FALSE
  )
  checkmate::expect_data_frame(data, col.names = "strict", nrows = n)
})

#' Check missing formal argument
#'
#' @description
#' These functions check whether a value was specified as an argument to a
#' function.
#'
#' @param x \[`any`\]\cr
#' A formal argument.
#'
#' @return
#' Depending on the function prefix:
#' - If the check is successful, `assert_missing()` returns `x` invisibly,
#' whereas `check_missing()` and `test_missing()` return `TRUE`.
#' - If the check is not successful, `assert_missing()` throws an error message,
#' `test_missing()` returns `FALSE`, and `check_missing()` returns a string with
#' the error message.
#'
#' @keywords validation
#' @family package helpers
#' @export
#'
#' @examples
#' f <- function(x) {
#'   check_missing(x)
#' }
#' f()
#'
#' g <- function(x) {
#'   test_missing(x)
#' }
#' g()
#'
#' h <- function(x) {
#'   assert_missing(x)
#' }
#' \dontrun{
#' h()
#' }

check_missing <- function(x) {
  x_missing <- missing(x)
  if (isTRUE(x_missing)) {
    return("Argument needs a value")
  }
  return(TRUE)
}

#' @rdname check_missing
#' @export

assert_missing <- function(x) {
  checkmate::makeAssertion(
    x,
    check_missing(x),
    checkmate::vname(x),
    NULL
  )
}

#' @rdname check_missing
#' @export

test_missing <- checkmate::makeTestFunction(
  check_missing
)

test_that("check for missing formal argument works", {

  f <- function(x, type) {
    switch(
      type,
      "check" = check_missing(x),
      "test" = test_missing(x),
      "assert" = assert_missing(x)
    )
  }

  expect_equal(f(type = "check"), "Argument needs a value")
  expect_true(f(1, type = "check"))

  expect_false(f(type = "test"))
  expect_true(f(1, type = "test"))

  expect_error(f(type = "assert"), "Assertion on 'x' failed: Argument needs a value.")
  expect_silent(f(1, type = "assert"))
})

#' Truncated normal distribution
#'
#' @description
#' The function `dtnorm()` computes the density of a truncated normal
#' distribution.
#'
#' The function `rtnorm()` samples from a truncated normal distribution.
#'
#' The function `dttnorm()` and `rttnorm()` compute the density and sample from
#' a two-sided truncated normal distribution, respectively.
#'
#' The functions with suffix `_cpp` perform no input checks, hence are faster.
#'
#' @param x \[`numeric(1)`\]\cr
#' A quantile.
#'
#' @param mean \[`numeric(1)`\]\cr
#' The mean.
#'
#' @param sd \[`numeric(1)`\]\cr
#' The non-negative standard deviation.
#'
#' @param log \[`logical(1)`\]\cr
#' Return the logarithm of the density value?
#'
#' @param point,lower,upper \[`numeric(1)`\]\cr
#' The truncation point.
#'
#' @param above \[`logical(1)`\]\cr
#' Truncate from above? Else, from below.
#'
#' @return
#' For `dtnorm()` and `dttnorm()`: The density value.
#'
#' For `rtnorm()` and `rttnorm()`: The random draw
#'
#' @keywords distribution
#' @family simulation helpers
#' @export
#'
#' @examples
#' x <- c(0, 0)
#' mean <- c(0, 0)
#' Sigma <- diag(2)
#'
#' # compute density
#' dmvnorm(x = x, mean = mean, Sigma = Sigma)
#' dmvnorm(x = x, mean = mean, Sigma = Sigma, log = TRUE)
#'
#' # sample
#' rmvnorm(n = 3, mean = mean, Sigma = Sigma)
#' rmvnorm(mean = mean, Sigma = Sigma, log = TRUE)

dtnorm <- function(x, mean, sd, point, above, log = FALSE) {
  checkmate::assert_number(x)
  checkmate::assert_number(mean)
  checkmate::assert_number(sd)
  checkmate::assert_number(point)
  checkmate::assert_flag(above)
  checkmate::assert_flag(log)
  dtnorm_cpp(x, mean, sd, point, above, log)
}

#' @rdname dtnorm
#' @export

dttnorm <- function(x, mean, sd, lower, upper, log = FALSE) {
  checkmate::assert_number(x)
  checkmate::assert_number(mean)
  checkmate::assert_number(sd)
  checkmate::assert_number(lower)
  checkmate::assert_number(upper, lower = lower)
  checkmate::assert_flag(log)
  dttnorm_cpp(x, mean, sd, lower, upper, log)
}

#' @rdname dtnorm
#' @export

rtnorm <- function(mean, sd, point, above, log = FALSE) {
  checkmate::assert_number(mean)
  checkmate::assert_number(sd)
  checkmate::assert_number(point)
  checkmate::assert_flag(above)
  checkmate::assert_flag(log)
  rtnorm_cpp(mean, sd, point, above, log)
}

#' @rdname dtnorm
#' @export

rttnorm <- function(mean, sd, lower, upper, log = FALSE) {
  checkmate::assert_number(mean)
  checkmate::assert_number(sd)
  checkmate::assert_number(lower)
  checkmate::assert_number(upper, lower = lower)
  checkmate::assert_flag(log)
  rttnorm_cpp(mean, sd, lower, upper, log)
}

test_that("univariate truncated normal density can be computed", {
  x <- 1.0
  mean <- 0.0
  sd <- 1.0
  point <- 0.0
  factor <- 1000

  density_b <- dtnorm(x, mean, sd, point, FALSE, FALSE)
  density_b_rounded <- round(density_b * factor) / factor
  expect_equal(density_b_rounded, 0.484)

  density_b_log <- dtnorm(x, mean, sd, point, FALSE, TRUE)
  density_b_log_rounded <- round(density_b_log * factor) / factor
  expect_equal(density_b_log_rounded, -0.726)

  density_a <- dtnorm(x, mean, sd, point, TRUE, FALSE)
  density_a_rounded <- round(density_a * factor) / factor
  expect_equal(density_a_rounded, 0)

  density_a_log <- dtnorm(x, mean, sd, point, TRUE, TRUE)
  density_a_log_rounded <- round(density_a_log * factor) / factor
  expect_equal(density_a_log_rounded, 0)
})

test_that("univariate truncated normal can be drawn", {
  mean <- 0.0
  sd <- 1.0
  point <- 0.0

  draw_tb <- rtnorm(mean, sd, point, FALSE, FALSE)
  expect_true(draw_tb > 0.0)

  draw_ta <- rtnorm(mean, sd, point, TRUE, FALSE)
  expect_true(draw_ta < 0.0)
})

test_that("univariate doubly truncated normal density can be computed", {
  x_b <- -2.0
  x_m <- 0.0
  x_a <- 2.0
  mean <- 0.0
  sd <- 1.0
  lower <- -1.0
  upper <- 1.0
  factor <- 1000

  density_b <- dttnorm(x_b, mean, sd, lower, upper, FALSE)
  density_b_rounded <- round(density_b * factor) / factor
  expect_equal(density_b_rounded, 0)

  density_b_log <- dttnorm(x_b, mean, sd, lower, upper, TRUE)
  density_b_log_rounded <- round(density_b_log * factor) / factor
  expect_equal(density_b_log_rounded, 0)

  density_m <- dttnorm(x_m, mean, sd, lower, upper, FALSE)
  density_m_rounded <- round(density_m * factor) / factor
  expect_equal(density_m_rounded, 0.584)

  density_m_log <- dttnorm(x_m, mean, sd, lower, upper, TRUE)
  density_m_log_rounded <- round(density_m_log * factor) / factor
  expect_equal(density_m_log_rounded, -0.537)

  density_a <- dttnorm(x_a, mean, sd, lower, upper, FALSE)
  density_a_rounded <- round(density_a * factor) / factor
  expect_equal(density_a_rounded, 0)

  density_a_log <- dttnorm(x_a, mean, sd, lower, upper, TRUE)
  density_a_log_rounded <- round(density_a_log * factor) / factor
  expect_equal(density_a_log_rounded, 0)
})

test_that("univariate doubly truncated normal can be drawn", {
  mean <- 0.0
  sd <- 1.0
  lower <- -1.0
  upper <- 1.0

  draw <- rttnorm(mean, sd, lower, upper)
  expect_true(draw > -1.0)
  expect_true(draw < 1.0)
})

#' Print (abbreviated) \code{matrix}
#'
#' @description
#' This function prints a (possibly abbreviated) \code{matrix}.
#'
#' @references
#' This function is a modified version of `pprint()` from the `{ramify}`
#' package.
#'
#' @param x \[`atomic()` | `matrix`\]\cr
#' The object to be printed.
#'
#' @param rowdots \[`integer(1)`\]\cr
#' The row number which is replaced by \code{...}.
#'
#' @param coldots \[`integer(1)`\]\cr
#' The column number which is replaced by \code{...}.
#'
#' @param digits \[`integer(1)`\]\cr
#' The number of printed decimal places if input \code{x} is \code{numeric}.
#'
#' @param label \[`character(1)`\]\cr
#' A label for \code{x}. Only printed if \code{simplify = FALSE}.
#'
#' @param simplify \[`logical(1)`\]\cr
#' Simplify the output?
#'
#' @param details \[`logical(1)`\]\cr
#' Print the type and dimension of \code{x}?
#'
#' @return
#' Invisibly returns \code{x}.
#'
#' @export
#' @keywords packaging
#' @family package helpers
#'
#' @examples
#' print_matrix(x = 1, label = "single numeric")
#' print_matrix(x = LETTERS[1:26], label = "letters")
#' print_matrix(x = 1:3, coldots = 2)
#' print_matrix(x = matrix(rnorm(99), ncol = 1), label = "single column matrix")
#' print_matrix(x = matrix(1:100, nrow = 1), label = "single row matrix")
#' print_matrix(x = matrix(LETTERS[1:24], ncol = 6), label = "big matrix")
#' print_matrix(x = diag(5), coldots = 2, rowdots = 2, simplify = TRUE)

print_matrix <- function(
    x, rowdots = 4, coldots = 4, digits = 2, label = NULL, simplify = FALSE,
    details = !simplify
  ) {
  checkmate::assert_int(rowdots, lower = 1)
  checkmate::assert_int(coldots, lower = 1)
  checkmate::assert_int(digits)
  checkmate::assert_flag(details)
  checkmate::assert_flag(simplify)
  if (!is.null(label)) {
    label <- as.character(label)
    stopifnot(length(label) == 1)
  }
  add_dots <- function(x, pos) {
    if (length(x) > pos) c(x[seq_len(pos - 1)], "...", x[length(x)]) else x
  }
  if (is.numeric(x)) x <- round(x, digits)
  if (!is.null(label)) cat(label, ": ")
  if (length(x) == 1) {
    cat(x)
  } else if (!is.matrix(x)) {
    if (details) {
      cat(typeof(x), "vector of length", length(x), "\n")
    }
    cat(noquote(add_dots(x, coldots)))
  } else {
    row_labs <- if (is.null(rownames(x))) {
      paste0("[", seq_len(nrow(x)), ",]")
    } else {
      rownames(x)
    }
    col_labs <- if (is.null(colnames(x))) {
      paste0("[,", seq_len(ncol(x)), "]")
    } else {
      colnames(x)
    }
    coldots <- max(1, min(ncol(x) - 1, coldots))
    rowdots <- max(1, min(nrow(x) - 1, rowdots))
    if (coldots == 1 && rowdots == 1) {
      if (nrow(x) <= 2 && ncol(x) <= 2) {
        res <- x
      } else {
        res <- if (nrow(x) == 1) {
          matrix(c("...", x[1, ncol(x)]), 1, 2)
        } else if (ncol(x) == 1) {
          matrix(c("...", x[nrow(x), 1]), 2, 1)
        } else {
          matrix(c("...", "...", "...", x[nrow(x), ncol(x)]), 2, 2)
        }
        row_labs <- add_dots(row_labs, 1)
        col_labs <- add_dots(col_labs, 1)
      }
    } else {
      x2 <- if (nrow(x) == 1) {
        cbind(x[1, 1:coldots, drop = FALSE], x[1, ncol(x), drop = FALSE])
      } else if (ncol(x) == 1) {
        rbind(x[1:rowdots, 1, drop = FALSE], x[nrow(x), 1, drop = FALSE])
      } else {
        rbind(
          cbind(
            x[1:rowdots, 1:coldots, drop = FALSE],
            x[1:rowdots, ncol(x), drop = FALSE]
          ),
          cbind(
            x[nrow(x), 1:coldots, drop = FALSE],
            x[nrow(x), ncol(x), drop = FALSE]
          )
        )
      }
      charx <- as.character(x2)
      dim(charx) <- dim(x2)
      if (nrow(x) <= rowdots + 1 && ncol(x) <= coldots + 1) {
        res <- charx
      } else if (nrow(x) > rowdots + 1 && ncol(x) <= coldots + 1) {
        res <- rbind(
          as.matrix(charx[seq_len(rowdots - 1), ]),
          rep("...", ncol(charx)),
          charx[nrow(charx), ]
        )
        row_labs <- add_dots(row_labs, pos = rowdots)
      } else if (nrow(x) <= rowdots + 1 && ncol(x) > coldots + 1) {
        res <- t(apply(charx, 1, add_dots, pos = coldots))
        col_labs <- add_dots(col_labs, pos = coldots)
      } else if (nrow(x) > rowdots + 1 && ncol(x) > coldots + 1) {
        smallx <- t(apply(charx[seq_len(rowdots - 1), , drop = FALSE], 1,
                          add_dots,
                          pos = coldots
        ))
        res <- rbind(
          smallx,
          rep("...", ncol(smallx)),
          add_dots(charx[nrow(charx), ], pos = coldots)
        )
        row_labs <- add_dots(row_labs, pos = rowdots)
        col_labs <- add_dots(col_labs, pos = coldots)
      }
    }
    if (details) {
      cat(paste(dim(x), collapse = " x "), "matrix of", paste0(typeof(x), "s"), "\n")
    }
    if (simplify) {
      cat(paste("[", paste(apply(res, 1, paste, collapse = " "),
                           collapse = "; "
      ), "]"))
    } else {
      prmatrix(res,
               rowlab = row_labs, collab = col_labs, quote = FALSE,
               right = TRUE
      )
    }
  }
  return(invisible(x))
}

test_that("matrix printing works", {
  expect_snapshot(
    print_matrix(x = 1)
  )
  expect_snapshot(
    print_matrix(x = 1.5, label = "single numeric")
  )
  expect_snapshot(
    print_matrix(x = 1.5, label = "single numeric", simplify = TRUE)
  )
  expect_snapshot(
    print_matrix(x = LETTERS[1:26])
  )
  expect_snapshot(
    print_matrix(x = LETTERS[1:26], label = "letters")
  )
  expect_snapshot(
    print_matrix(x = LETTERS[1:26], label = "letters", details = FALSE)
  )
  expect_snapshot(
    print_matrix(x = LETTERS[1:26], label = "letters", simplify = TRUE)
  )
  expect_snapshot(
    print_matrix(x = LETTERS[1:26], rowdots = 1)
  )
  expect_snapshot(
    print_matrix(x = LETTERS[1:26], rowdots = 26)
  )
  expect_snapshot(
    print_matrix(x = LETTERS[1:26], coldots = 1)
  )
  expect_snapshot(
    print_matrix(x = LETTERS[1:26], coldots = 25)
  )
  expect_snapshot(
    print_matrix(x = LETTERS[1:26], coldots = 26)
  )
  expect_snapshot(
    print_matrix(x = LETTERS[1:26], coldots = 10)
  )
  expect_error(
    print_matrix(x = LETTERS[1:26], coldots = 0)
  )
  expect_snapshot(
    print_matrix(x = matrix(LETTERS[1:24], ncol = 6))
  )
  expect_snapshot(
    print_matrix(x = matrix(LETTERS[1:24], ncol = 6), label = "big matrix")
  )
  expect_snapshot(
    print_matrix(x = matrix(LETTERS[1:24], ncol = 6), rowdots = 2)
  )
  expect_snapshot(
    print_matrix(x = matrix(LETTERS[1:24], ncol = 6), details = FALSE)
  )
  expect_snapshot(
    print_matrix(x = matrix(LETTERS[1:24], ncol = 6), coldots = 1, rowdots = 1)
  )
  expect_snapshot(
    print_matrix(x = matrix(LETTERS[1:24], ncol = 6), simplify = TRUE, coldots = 2, rowdots = 2)
  )
  expect_snapshot(
    print_matrix(x = diag(5), coldots = 2, rowdots = 3, simplify = TRUE, digits = 0)
  )
  expect_snapshot(
    print_matrix(x = matrix(1:100, nrow = 1))
  )
  expect_snapshot(
    print_matrix(x = matrix(1:100, nrow = 1), label = "single row matrix")
  )
  expect_snapshot(
    print_matrix(x = matrix(1:100, nrow = 1), details = FALSE)
  )
  expect_snapshot(
    print_matrix(x = matrix(1:100, nrow = 1), coldots = 1)
  )
  expect_snapshot(
    print_matrix(x = matrix(1:100, nrow = 1), rowdots = 1)
  )
  expect_snapshot(
    print_matrix(x = matrix(1:100, nrow = 1), simplify = TRUE, coldots = 5)
  )
  expect_snapshot(
    print_matrix(x = matrix(1:100, ncol = 1))
  )
  expect_snapshot(
    print_matrix(x = matrix(1:100, ncol = 1), label = "single column matrix")
  )
  expect_snapshot(
    print_matrix(x = matrix(1:100, ncol = 1), details = FALSE)
  )
  expect_snapshot(
    print_matrix(x = matrix(1:100, ncol = 1), coldots = 1)
  )
  expect_snapshot(
    print_matrix(x = matrix(1:100, ncol = 1), rowdots = 1)
  )
  expect_snapshot(
    print_matrix(x = matrix(1:100, ncol = 1), simplify = TRUE, coldots = 5)
  )
  expect_snapshot(
    print_matrix(x = diag(2), simplify = TRUE)
  )
  expect_snapshot(
    print_matrix(x = diag(2), simplify = FALSE)
  )
  expect_snapshot({
    print_matrix(structure(diag(3), dimnames = list(c("la", "le", "lu"), c("x", "y", "z"))))
  })
})

#' Sample transition probability matrices
#'
#' @description
#' This function returns a random, squared matrix of dimension \code{dim}
#' that fulfills the properties of a transition probability matrix.
#'
#' @inheritParams sample_correlation_matrix
#'
#' @param state_persistent \[`logical(1)`\]\cr
#' Put more probability on the diagonal?
#'
#' @return
#' A transition probability \code{matrix}.
#'
#' @keywords simulation
#' @family matrix helpers
#' @export
#'
#' @examples
#' sample_transition_probability_matrix(dim = 3)

sample_transition_probability_matrix <- function(dim, state_persistent = TRUE) {
  checkmate::assert_int(dim, lower = 1)
  checkmate::assert_flag(state_persistent)
  Gamma <- matrix(stats::runif(dim^2), dim, dim)
  if (state_persistent) Gamma <- Gamma + diag(dim)
  Gamma / rowSums(Gamma)
}

#' Measure computation time
#'
#' @description
#' This function measures the computation time of a call.
#'
#' @details
#' This function is a wrapper for \code{\link[base]{do.call}}.
#'
#' @param what,args
#' Passed to \code{\link[base]{do.call}}.
#' @param units
#' Passed to \code{\link[base]{difftime}}.
#'
#' @return
#' A list of the two elements \code{"result"} (the results of the \code{do.call}
#' call) and \code{"time"} (the computation time).
#'
#' @keywords simulation
#' @family function helpers
#' @export do.call_timed
#' @usage do.call_timed(what, args, units = "secs")
#'
#' @examples
#' \dontrun{
#' what <- function(s) {
#'   Sys.sleep(s)
#'   return(s)
#' }
#' args <- list(s = 1)
#' do.call_timed(what = what, args = args)
#' }

do.call_timed <- function(what, args, units = "secs") {
  start <- Sys.time()
  res <- do.call(what = what, args = args)
  end <- Sys.time()
  list("result" = res, "time" = difftime(end, start, units = units))
}

test_that("measure computation time works", {
  what <- function(s) {
    Sys.sleep(s)
    return(s)
  }
  args <- list(s = 1)
  out <- do.call_timed(what = what, args = args)
  expect_type(out, "list")
  expect_length(out, 2)
  expect_equal(out[["result"]], 1)
  expect_type(out[["time"]], "double")
  expect_true(abs(out[["time"]] - 2) < 1)
})

#' Mixture of normal distributions
#'
#' @description
#' The function `dmixnorm()` computes the density of a mixture of multivariate
#' normal distribution.
#'
#' The function `pmixnorm()` computes the cumulative distribution function of a
#' mixture of multivariate normal distribution.
#'
#' The function `rmixnorm()` samples from a mixture of multivariate normal
#' distribution.
#'
#' The functions with suffix `_cpp` perform no input checks, hence are faster.
#'
#' The univariate normal mixture is available as the special case `p = 1`.
#'
#' @details
#' `pmixnorm()` is based on `mvtnorm::pmvnorm` with the randomized
#' Quasi-Monte-Carlo procedure by Genz and Bretz. The argument `abseps` controls
#' the accuracy of the Gaussian integral approximation.
#'
#' @param x \[`numeric(p)`\]\cr
#' A quantile vector of length `p`, where `p` is the dimension.
#'
#' @param mean \[`matrix(nrow = p, ncol = c)`\]\cr
#' The mean vectors for each component in columns.
#'
#' @param Sigma \[`matrix(nrow = p^2, ncol = c)`\]\cr
#' The vectorized covariance matrices for each component in columns.
#'
#' @param proportions \[`numeric(c)`\]\cr
#' The non-negative mixing proportions for each components.
#'
#' If proportions do not sum to unity, they are rescaled to do so.
#'
#' @param abseps \[`numeric(1)`\]\cr
#' The absolute error tolerance.
#'
#' @param n \[`integer(1)`\]\cr
#' The number of requested samples.
#'
#' @return
#' For `dmixnorm()`: The density value.
#'
#' For `pmixnorm()`: The value of the distribution function.
#'
#' For `rmixnorm()`: If \code{n = 1} a \code{vector} of length \code{p} (note
#' that it is a column vector for `rmixnorm_cpp()`), else
#' a \code{matrix} of dimension \code{n} times \code{p} with samples as rows.
#'
#' @keywords distribution
#' @family simulation helpers
#' @export
#'
#' @examples
#' x <- c(0, 0)
#' mean <- matrix(c(-1, -1, 0, 0), ncol = 2)
#' Sigma <- matrix(c(diag(2), diag(2)), ncol = 2)
#' proportions <- c(0.7, 0.3)
#'
#' # compute density
#' dmixnorm(x = x, mean = mean, Sigma = Sigma, proportions = proportions)
#'
#' # compute CDF
#' pmixnorm(x = x, mean = mean, Sigma = Sigma, proportions = proportions)
#'
#' # sample
#' rmixnorm(n = 3, mean = mean, Sigma = Sigma, proportions = proportions)

dmixnorm <- function(x, mean, Sigma, proportions) {
  input_check_response(
    check = check_numeric_vector(x, any.missing = FALSE, min.len = 1),
    var_name = "x"
  )
  dim <- length(x)
  input_check_response(
    check = checkmate::check_matrix(
      mean, mode = "numeric", any.missing = FALSE, nrows = dim, min.cols = 1
    ),
    var_name = "mean"
  )
  ncomp <- ncol(mean)
  input_check_response(
    check = checkmate::check_matrix(
      Sigma, mode = "numeric", any.missing = FALSE, nrows = dim^2, ncols = ncomp
    ),
    var_name = "Sigma"
  )
  for (c in seq_len(ncomp)) {
    input_check_response(
      check = check_covariance_matrix(
        matrix(Sigma[, c], nrow = dim, ncol = dim), dim = dim
      ),
      var_name = paste0("matrix(Sigma[, ", c, "], nrow = dim, ncol = dim)")
    )
  }
  input_check_response(
    check = check_numeric_vector(
      proportions, lower = 0, finite = TRUE, any.missing = FALSE, len = ncomp
    ),
    var_name = "proportions"
  )
  dmixnorm_cpp(x, mean, Sigma, proportions)
}

#' @rdname dmixnorm
#' @export

pmixnorm <- function(x, mean, Sigma, proportions, abseps = 1e-3) {
  input_check_response(
    check = check_numeric_vector(x, any.missing = FALSE, min.len = 1),
    var_name = "x"
  )
  dim <- length(x)
  input_check_response(
    check = checkmate::check_matrix(
      mean, mode = "numeric", any.missing = FALSE, nrows = dim, min.cols = 1
    ),
    var_name = "mean"
  )
  ncomp <- ncol(mean)
  input_check_response(
    check = checkmate::check_matrix(
      Sigma, mode = "numeric", any.missing = FALSE, nrows = dim^2, ncols = ncomp
    ),
    var_name = "Sigma"
  )
  for (c in seq_len(ncomp)) {
    input_check_response(
      check = check_covariance_matrix(
        matrix(Sigma[, c], nrow = dim, ncol = dim), dim = dim
      ),
      var_name = paste0("matrix(Sigma[, ", c, "], nrow = dim, ncol = dim)")
    )
  }
  input_check_response(
    check = check_numeric_vector(
      proportions, lower = 0, finite = TRUE, any.missing = FALSE, len = ncomp
    ),
    var_name = "proportions"
  )
  pmixnorm_cpp(x, mean, Sigma, proportions, abseps)
}

#' @rdname dmixnorm
#' @export

rmixnorm <- function(n = 1, mean, Sigma, proportions) {
  input_check_response(
    check = checkmate::check_int(n),
    var_name = "n"
  )
  input_check_response(
    check = checkmate::check_matrix(
      mean, mode = "numeric", any.missing = FALSE, min.rows = 1, min.cols = 1
    ),
    var_name = "mean"
  )
  dim <- nrow(mean)
  ncomp <- ncol(mean)
  input_check_response(
    check = checkmate::check_matrix(
      Sigma, mode = "numeric", any.missing = FALSE, nrows = dim^2, ncols = ncomp
    ),
    var_name = "Sigma"
  )
  for (c in seq_len(ncomp)) {
    input_check_response(
      check = check_covariance_matrix(
        matrix(Sigma[, c], nrow = dim, ncol = dim), dim = dim
      ),
      var_name = paste0("matrix(Sigma[, ", c, "], nrow = dim, ncol = dim)")
    )
  }
  input_check_response(
    check = check_numeric_vector(
      proportions, lower = 0, finite = TRUE, any.missing = FALSE, len = ncomp
    ),
    var_name = "proportions"
  )
  sample <- replicate(
    n = n, rmixnorm_cpp(mean, Sigma, proportions), simplify = TRUE
  )
  if (n == 1) {
    drop(sample)
  } else if (dim == 1) {
    as.matrix(sample)
  } else {
    t(sample)
  }
}

test_that("Mixture normal density can be computed", {

  ### univariate
  x <- 1
  mean <- matrix(c(-1, 0, 1), ncol = 3)
  Sigma <- matrix(c(0.5, 1, 1.5), ncol = 3)
  proportions <- c(0.5, 0.3, 0.2)
  p <- dmixnorm(x = x, mean = mean, Sigma = Sigma, proportions = proportions)
  expect_equal(
    p,
    proportions[1] * dnorm(x, mean[1], sqrt(Sigma[, 1])) +
      proportions[2] * dnorm(x, mean[2], sqrt(Sigma[, 2])) +
      proportions[3] * dnorm(x, mean[3], sqrt(Sigma[, 3]))
  )
  expect_equal(
    p,
    dmixnorm(x = x, mean = mean, Sigma = Sigma, proportions = proportions * 2)
  )

  ### multivariate
  x <- c(0, 0)
  mean <- matrix(c(-1, -1, 0, 0), ncol = 2)
  Sigma <- matrix(c(diag(2), diag(2)), ncol = 2)
  proportions <- c(0.7, 0.3)
  factor <- 1000
  expect_equal(
    round(dmixnorm(x = x, mean = mean, Sigma = Sigma, proportions = proportions) *
            factor) / factor,
    0.089
  )

  ### one component
  x <- c(0, 0)
  mean <- matrix(c(-1, -1), ncol = 1)
  Sigma <- matrix(c(diag(2)), ncol = 1)
  proportions <- 1
  expect_equal(
    dmixnorm(x = x, mean = mean, Sigma = Sigma, proportions = proportions),
    dmvnorm(x = x, mean = as.vector(mean), Sigma = matrix(Sigma, nrow = 2, ncol = 2))
  )
})

test_that("Mixture normal CDF can be computed", {

  ### univariate
  x <- 1
  mean <- matrix(c(-1, 0, 1), ncol = 3)
  Sigma <- matrix(c(0.5, 1, 1.5), ncol = 3)
  proportions <- c(0.5, 0.3, 0.2)
  p <- pmixnorm(x = x, mean = mean, Sigma = Sigma, proportions = proportions)
  expect_equal(
    p,
    proportions[1] * pnorm(x, mean[1], sqrt(Sigma[, 1])) +
      proportions[2] * pnorm(x, mean[2], sqrt(Sigma[, 2])) +
      proportions[3] * pnorm(x, mean[3], sqrt(Sigma[, 3]))
  )
  expect_equal(
    p,
    pmixnorm(x = x, mean = mean, Sigma = Sigma, proportions = proportions * 2)
  )

  ### multivariate
  x <- c(0, 0)
  mean <- matrix(c(-1, -1, 0, 0), ncol = 2)
  Sigma <- matrix(c(diag(2), diag(2)), ncol = 2)
  proportions <- c(0.7, 0.3)
  factor <- 1000
  expect_equal(
    round(pmixnorm(x = x, mean = mean, Sigma = Sigma, proportions = proportions) *
            factor) / factor,
    0.571
  )

  ### one component
  x <- c(0, 0)
  mean <- matrix(c(-1, -1), ncol = 1)
  Sigma <- matrix(c(diag(2)), ncol = 1)
  proportions <- 1
  expect_equal(
    pmixnorm(x = x, mean = mean, Sigma = Sigma, proportions = proportions),
    as.numeric(pmvnorm(
      x = x,
      mean = as.vector(mean),
      Sigma = matrix(Sigma, nrow = 2, ncol = 2)
    ))
  )
})

test_that("Mixture normal draws can be generated", {

  set.seed(1)

  ### univariate case
  mean <- matrix(0, nrow = 1, ncol = 1)
  Sigma <- matrix(1, ncol = 1)
  proportions <- 1
  x1 <- rmixnorm(n = 1, mean = mean, Sigma = Sigma, proportions = proportions)
  expect_true(is.numeric(x1) && length(x1) == 1 && !is.matrix(x1))
  x5 <- rmixnorm(n = 5, mean = mean, Sigma = Sigma, proportions = proportions)
  expect_true(is.matrix(x5))
  expect_equal(dim(x5), c(5L, 1L))
  mean <- matrix(c(-1, 0, 1), ncol = 3)
  Sigma <- matrix(c(0.5, 1, 1.5), ncol = 3)
  proportions <- c(0.5, 0.3, 0.2)
  w <- proportions / sum(proportions)
  mu <- as.numeric(mean)
  vars <- as.numeric(Sigma)
  m_theo <- sum(w * mu)
  v_theo <- sum(w * (vars + mu^2)) - m_theo^2
  n <- 10000
  samp <- rmixnorm(n = n, mean = mean, Sigma = Sigma, proportions = proportions)
  expect_true(is.matrix(samp))
  expect_equal(dim(samp), c(n, 1))
  m_emp <- mean(samp)
  v_emp <- var(drop(samp))
  expect_equal(m_emp, m_theo, tolerance = 0.1)
  expect_equal(as.numeric(v_emp), v_theo, tolerance = 0.1)
  samp2 <- rmixnorm(n = n, mean = mean, Sigma = Sigma, proportions = proportions * 2)
  m_emp2 <- mean(samp2)
  v_emp2 <- var(drop(samp2))
  expect_lt(abs(m_emp - m_emp2), 0.1)
  expect_lt(abs(as.numeric(v_emp) - as.numeric(v_emp2)), 0.1)

  ### bivariate case
  mean2 <- matrix(c(0, 0), nrow = 2, ncol = 1)
  Sigma2 <- matrix(c(diag(2)), ncol = 1)
  y1 <- rmixnorm(n = 1, mean = mean2, Sigma = Sigma2, proportions = 1)
  expect_true(is.numeric(y1) && length(y1) == 2 && !is.matrix(y1))
  y7 <- rmixnorm(n = 7, mean = mean2, Sigma = Sigma2, proportions = 1)
  expect_true(is.matrix(y7))
  expect_equal(dim(y7), c(7L, 2L))
  mean <- matrix(c(-1, -1, 0, 0), ncol = 2)
  Sigma <- matrix(c(diag(2), diag(2)), ncol = 2)
  proportions <- c(0.7, 0.3)
  w <- proportions / sum(proportions)
  m_theo <- drop(mean %*% w)
  dimx <- nrow(mean)
  comps <- ncol(mean)
  S_acc <- matrix(0, nrow = dimx, ncol = dimx)
  for (k in seq_len(comps)) {
    mu_k <- matrix(mean[, k], ncol = 1)
    Sig_k <- matrix(Sigma[, k], nrow = dimx, ncol = dimx)
    S_acc <- S_acc + w[k] * (Sig_k + mu_k %*% t(mu_k))
  }
  S_theo <- S_acc - m_theo %*% t(m_theo)
  n <- 10000
  samp <- rmixnorm(n = n, mean = mean, Sigma = Sigma, proportions = proportions)
  expect_true(is.matrix(samp))
  expect_equal(dim(samp), c(n, dimx))
  m_emp <- colMeans(samp)
  S_emp <- stats::cov(samp)
  expect_equal(m_emp, as.numeric(m_theo), tolerance = 0.1)
  expect_equal(S_emp, S_theo, tolerance = 0.1)
  samp2 <- rmixnorm(n = n, mean = mean, Sigma = Sigma, proportions = proportions * 10)
  m_emp2 <- colMeans(samp2)
  expect_lt(max(abs(m_emp - m_emp2)), 0.1)
})

#' Gaussian total variation
#'
#' @description
#' Computes the total variation (TV) between two multivariate Gaussian
#' distributions \eqn{f_1,f_2}: \deqn{\mathrm{TV}(f_1, f_2) = \tfrac{1}{2}
#' \int_{\mathbb{R}^p} \lvert f_1(x) - f_2(x) \rvert \, dx.}
#' The value ranges from 0 (identical distributions) to 1 (no overlap).
#'
#' @param mean1,mean2 \[`numeric(p)`\]\cr
#' The mean vectors.
#'
#' @param Sigma1,Sigma2 \[`matrix(nrow = p, ncol = p)`\]\cr
#' The covariance matrices.
#'
#' @param method \[`character(1)`\]\cr
#' Computation method. One of:
#'
#' * `"auto"`: use closed-form formula when covariances are equal, otherwise
#'   use `"cubature"` for \eqn{p \le 2} and `"mc"` for higher dimensions.
#'
#' * `"mc"`: estimate via Monte Carlo importance sampling from the mixture
#'   \eqn{0.5 (f_1 + f_2)}.
#'
#' * `"cubature"`: compute overlap via adaptive cubature integration over a
#'   bounding box, then convert to TV. Exact but slow for \eqn{p \ge 2}.
#'
#' @param n \[`integer(1)`\]\cr
#' Number of Monte Carlo samples to draw.
#'
#' @param tol_equal \[`numeric(1)`\]\cr
#' Numerical tolerance used to decide whether `Sigma1` and `Sigma2` are
#' considered equal (enabling the closed-form formula in `"auto"` mode).
#'
#' @param eps \[`numeric(1)`\]\cr
#' Only used when `method = "cubature"`. Specifies the total probability mass
#' allowed to lie outside the integration hyper-rectangle across all dimensions.
#' This determines the numerical integration bounds: the function chooses limits
#' so that the probability of a point from either Gaussian falling outside the
#' box is at most `eps`. The bound is split evenly across dimensions via a union
#' bound, so the per-dimension tail probability is approximately `eps / p`.
#' Smaller values produce wider bounds (slower but more accurate integration),
#' while larger values yield narrower bounds (faster but potentially less
#' accurate).
#'
#' @return
#' The total variation in \[0, 1\].
#'
#' @keywords distribution
#' @family simulation helpers
#' @export
#'
#' @examples
#' ### univariate case
#' mean1 <- 0
#' mean2 <- 1
#' Sigma1 <- Sigma2 <- matrix(1)
#' gaussian_tv(mean1, mean2, Sigma1, Sigma2)
#'
#' ### bivariate case
#' mean1 <- c(0, 0)
#' mean2 <- c(1, 1)
#' Sigma1 <- matrix(c(1, 0.2, 0.2, 1), ncol = 2)
#' Sigma2 <- matrix(c(1.5, -0.3, -0.3, 1), ncol = 2)
#' gaussian_tv(mean1, mean2, Sigma1, Sigma2, method = "mc", n = 1e3)

gaussian_tv <- function(
  mean1, mean2, Sigma1, Sigma2, method = c("auto", "mc", "cubature"),
  n = 1e4, tol_equal = 1e-6, eps = 1e-6
) {
  method <- match.arg(method)
  input_check_response(
    check = check_numeric_vector(mean1, finite = TRUE, any.missing = FALSE),
    var_name = "mean1"
  )
  p <- length(mean1)
  input_check_response(
    check = check_numeric_vector(
      mean2, finite = TRUE, any.missing = FALSE, len = p
    ),
    var_name = "mean2"
  )
  if (p == 1) {
    if (checkmate::test_number(Sigma1)) {
      Sigma1 <- as.matrix(Sigma1)
    }
    if (checkmate::test_number(Sigma2)) {
      Sigma2 <- as.matrix(Sigma2)
    }
  }
  input_check_response(
    check = check_covariance_matrix(Sigma1, dim = p),
    var_name = "Sigma1"
  )
  input_check_response(
    check = check_covariance_matrix(Sigma2, dim = p),
    var_name = "Sigma2"
  )
  if (identical(mean1, mean2) && identical(Sigma1, Sigma2)) return(0)
  input_check_response(
    check = checkmate::check_number(tol_equal, lower = 0, finite = TRUE),
    var_name = "tol_equal"
  )
  cov_equal <- isTRUE(all.equal(Sigma1, Sigma2, tolerance = tol_equal))
  if (method == "auto" && cov_equal) {
    mu_diff <- mean1 - mean2
    R <- chol(Sigma1)
    z <- backsolve(R, forwardsolve(t(R), mu_diff))
    delta <- sqrt(sum(z * z))
    return(1 - 2 * stats::pnorm(-delta / 2))
  }
  if (method == "cubature" || (method == "auto" && p <= 2)) {
    s1 <- sqrt(diag(Sigma1)); s2 <- sqrt(diag(Sigma2))
    k <- stats::qnorm(1 - eps / (2 * p))
    lower <- pmin(mean1 - k * s1, mean2 - k * s2)
    upper <- pmax(mean1 + k * s1, mean2 + k * s2)
    min_f1_f2 <- function(x) {
      f1 <- dmvnorm(x, mean = mean1, Sigma = Sigma1)
      f2 <- dmvnorm(x, mean = mean2, Sigma = Sigma2)
      pmin(f1, f2)
    }
    ovl <- cubature::hcubature(f = min_f1_f2, lower, upper)$integral
    return(1 - ovl)
  }
  input_check_response(
    check = checkmate::check_int(n, lower = 2),
    var_name = "n"
  )
  n1 <- n %/% 2
  n2 <- n - n1
  X1 <- rmvnorm(n = 1, mean1, Sigma1)
  X1 <- replicate(n1, rmvnorm(n = 1, mean1, Sigma1), simplify = TRUE)
  X2 <- replicate(n2, rmvnorm(n = 1, mean2, Sigma2), simplify = TRUE)
  X  <- cbind(X1, X2)
  lf1 <- apply(X, 2, dmvnorm, mean = mean1, Sigma = Sigma1, log = TRUE)
  lf2 <- apply(X, 2, dmvnorm, mean = mean2, Sigma = Sigma2, log = TRUE)
  log_min <- pmin(lf1, lf2)
  m <- pmax(lf1, lf2)
  log_q <- m + log(exp(lf1 - m) + exp(lf2 - m)) - log(2)
  w <- exp(log_min - log_q)
  ovl_hat <- mean(w)
  tv_hat  <- 1 - ovl_hat
  se_ovl <- stats::sd(w) / sqrt(length(w))
  structure(tv_hat, "se" = se_ovl)
}

test_that("Gaussian total variation can be computed", {

  ### identity
  m <- c(0, 0)
  S <- diag(2)
  tv <- gaussian_tv(m, m, S, S, method = "auto")
  expect_equal(tv, 0, tolerance = 1e-10)

  ### symmetry
  set.seed(1)
  m1 <- c(0, 0); m2 <- c(1, -1)
  S1 <- diag(2)
  S2 <- matrix(c(1.5, 0.3, 0.3, 1.2), 2)
  tv_a <- gaussian_tv(m1, m2, S1, S2, method = "mc", n = 1000)
  tv_b <- gaussian_tv(m2, m1, S2, S1, method = "mc", n = 1000)
  expect_equal(tv_a, tv_b, tolerance = 0.05)

  ### equal covariances
  s2 <- 1; Sig1D <- matrix(s2, 1, 1)
  for (d in c(0, 0.5, 1, 2, 3)) {
    mu1 <- 0; mu2 <- d
    tv_num <- gaussian_tv(mu1, mu2, Sig1D, Sig1D, method = "auto")
    tv_ref <- 1 - 2 * pnorm(-abs(mu1 - mu2) / (2 * sqrt(s2)))
    expect_equal(tv_num, tv_ref, tolerance = 1e-12)
  }

  ### cubature method
  mean1 <- 0
  mean2 <- 1
  Sigma1 <- 1
  Sigma2 <- 1.5
  tv_cub <- gaussian_tv(mean1, mean2, Sigma1, Sigma2, method = "cubature")
  expect_equal(tv_cub, 0.35, tolerance = 0.05)

  ### monotonicity
  ds <- c(0, 0.25, 0.5, 1, 2)
  tvs <- vapply(
    ds,
    function(d) gaussian_tv(0, d, Sig1D, Sig1D, method = "auto"), numeric(1)
  )
  expect_true(all(diff(tvs) >= -1e-10))
})

#' @keywords internal
#' @useDynLib oeli, .registration=TRUE
"_PACKAGE"

## usethis namespace: start
#' @importFrom benchmarkme get_cpu
#' @importFrom benchmarkme get_ram
#' @importFrom checkmate assert
#' @importFrom checkmate assert_atomic_vector
#' @importFrom checkmate assert_character
#' @importFrom checkmate assert_choice
#' @importFrom checkmate assert_count
#' @importFrom checkmate assert_data_frame
#' @importFrom checkmate assert_flag
#' @importFrom checkmate assert_function
#' @importFrom checkmate assert_int
#' @importFrom checkmate assert_integerish
#' @importFrom checkmate assert_list
#' @importFrom checkmate assert_matrix
#' @importFrom checkmate assert_names
#' @importFrom checkmate assert_number
#' @importFrom checkmate assert_numeric
#' @importFrom checkmate assert_string
#' @importFrom checkmate assert_subset
#' @importFrom checkmate assert_vector
#' @importFrom checkmate check_atomic_vector
#' @importFrom checkmate check_character
#' @importFrom checkmate check_choice
#' @importFrom checkmate check_class
#' @importFrom checkmate check_count
#' @importFrom checkmate check_data_frame
#' @importFrom checkmate check_directory_exists
#' @importFrom checkmate check_flag
#' @importFrom checkmate check_function
#' @importFrom checkmate check_int
#' @importFrom checkmate check_integerish
#' @importFrom checkmate check_list
#' @importFrom checkmate check_matrix
#' @importFrom checkmate check_names
#' @importFrom checkmate check_number
#' @importFrom checkmate check_numeric
#' @importFrom checkmate check_path_for_output
#' @importFrom checkmate check_string
#' @importFrom checkmate check_subset
#' @importFrom checkmate makeAssertion
#' @importFrom checkmate makeAssertionFunction
#' @importFrom checkmate makeTestFunction
#' @importFrom checkmate test_atomic_vector
#' @importFrom checkmate test_class
#' @importFrom checkmate test_int
#' @importFrom checkmate test_list
#' @importFrom checkmate test_matrix
#' @importFrom checkmate test_string
#' @importFrom checkmate vname
#' @importFrom cli cli_abort
#' @importFrom cli cli_inform
#' @importFrom cli cli_text
#' @importFrom cli cli_warn
#' @importFrom cli style_hyperlink
#' @importFrom cubature hcubature
#' @importFrom dplyr arrange
#' @importFrom dplyr bind_rows
#' @importFrom dplyr filter
#' @importFrom future multisession
#' @importFrom future plan
#' @importFrom future.apply future_lapply
#' @importFrom ggplot2 ggplot
#' @importFrom ggplot2 theme_void
#' @importFrom glue glue
#' @importFrom hexSticker sticker
#' @importFrom progressr handler_progress
#' @importFrom progressr handlers
#' @importFrom progressr progressor
#' @importFrom R6 R6Class
#' @importFrom Rcpp compileAttributes
#' @importFrom SimMultiCorrData calc_theory
#' @importFrom SimMultiCorrData rcorrvar
#' @importFrom SimMultiCorrData valid_corr
#' @importFrom stats cor
#' @importFrom stats dist
#' @importFrom stats dnorm
#' @importFrom stats rnorm
#' @importFrom stats runif
#' @importFrom tibble as_tibble
#' @importFrom tibble tibble
#' @importFrom tools package_native_routine_registration_skeleton
#' @importFrom utils combn
#' @importFrom utils head
#' @importFrom utils tail
## usethis namespace: end
NULL

#' Get function arguments
#'
#' @description
#' This function returns the names of function arguments.
#'
#' @param f \[`function`\]\cr
#' A \code{function}.
#'
#' @param with_default \[`logical(1)`\]\cr
#' Include function arguments that have default values?
#'
#' @param with_ellipsis \[`logical(1)`\]\cr
#' Include the \code{"..."} argument if present?
#'
#' @return
#' A \code{character} vector.
#'
#' @keywords functional
#' @family function helpers
#' @export
#'
#' @examples
#' f <- function(a, b = 1, c = "", ...) { }
#' function_arguments(f)
#' function_arguments(f, with_default = FALSE)
#' function_arguments(f, with_ellipsis = FALSE)

function_arguments <- function(f, with_default = TRUE, with_ellipsis = TRUE) {
  checkmate::assert_function(f)
  checkmate::assert_flag(with_default)
  args <- formals(f)
  if (is.null(args)) {
    return(character())
  }
  if (!with_default) {
    args <- args[sapply(seq_along(args), function(n) {
      any(!nzchar(args[[n]]) & is.name(args[[n]]))
    })]
  }
  if (!with_ellipsis) {
    args <- args[names(args) != "..."]
  }
  names(args)
}

test_that("names of function arguments can be extracted", {
  f <- function(a, b = 1, c = "", ...) { }
  expect_equal(
    function_arguments(f),
    c("a", "b", "c", "...")
  )
  expect_equal(
    function_arguments(f, with_default = FALSE),
    c("a", "...")
  )
  expect_equal(
    function_arguments(f, with_ellipsis = FALSE),
    c("a", "b", "c")
  )
  f <- function() { }
  expect_equal(
    function_arguments(f),
    character()
  )
  expect_equal(
    function_arguments(f, with_default = FALSE),
    character()
  )
  expect_equal(
    function_arguments(f, with_ellipsis = FALSE),
    character()
  )
  f <- function(x, a = 20, b = 0.2, c = 2 * pi) {
    -a * exp(-b * sqrt(mean(x^2))) - exp(mean(cos(c * x))) +
      a + exp(1)
  }
  expect_equal(
    function_arguments(f, with_default = TRUE, with_ellipsis = TRUE),
    c("x", "a", "b", "c")
  )
  expect_equal(
    function_arguments(f, with_default = TRUE, with_ellipsis = FALSE),
    c("x", "a", "b", "c")
  )
  expect_equal(
    function_arguments(f, with_default = FALSE, with_ellipsis = TRUE),
    c("x")
  )
  expect_equal(
    function_arguments(f, with_default = FALSE, with_ellipsis = FALSE),
    c("x")
  )
})

#' Merge named lists
#'
#' @description
#' This function merges \code{list}s based on their element names. Elements are
#' only included in the final output \code{list}, if no former \code{list} has
#' contributed an element with the same name.
#'
#' @param ...
#' One or more named \code{list}(s).
#'
#' @return
#' A \code{list}.
#'
#' @export
#' @keywords transformation
#' @family list helpers
#'
#' @examples
#' merge_lists(list("a" = 1, "b" = 2), list("b" = 3, "c" = 4, "d" = NULL))

merge_lists <- function(...) {
  inputs <- list(...)
  lapply(inputs, function(input) checkmate::assert_list(input, names = "unique"))
  final <- list()
  for (input in inputs) {
    for (element in names(input)) {
      if (!element %in% names(final)) {
        final[element] <- input[element]
      }
    }
  }
  return(final)
}

test_that("lists can be merged", {
  expect_equal(
    merge_lists(list("a" = 1, "b" = 2, "z" = NULL)),
    list(a = 1, b = 2, z = NULL)
  )
  expect_equal(
    merge_lists(list("a" = 1, "b" = 2, "z" = NULL), list()),
    list(a = 1, b = 2, z = NULL)
  )
  expect_equal(
    merge_lists(list("a" = 1, "b" = 2, "z" = NULL), list("b" = 3, "c" = 4)),
    list(a = 1, b = 2, z = NULL, c = 4)
  )
  expect_equal(
    merge_lists(list("a" = 1, "b" = 2), list("b" = 3, "c" = 4, "z" = NULL)),
    list(a = 1, b = 2, c = 4, z = NULL)
  )
  expect_equal(
    merge_lists(list("a" = 1, "b" = 2), list("b" = 3, "c" = 4), list("d" = 5)),
    list(a = 1, b = 2, c = 4, d = 5)
  )
  expect_equal(
    merge_lists(list("a" = 1, "b" = 2), list("b" = 3, "c" = 4), list("z" = NULL)),
    list(a = 1, b = 2, c = 4, z = NULL)
  )
})

#' Sample covariance matrix
#'
#' @description
#' This function samples a covariance matrix from an inverse Wishart
#' distribution.
#'
#' @inheritParams sample_correlation_matrix
#'
#' @param diag \[`logical(1)`\]\cr
#' Diagonal matrix?
#'
#' @return
#' A covariance \code{matrix}.
#'
#' @keywords simulation
#' @family matrix helpers
#' @export
#'
#' @examples
#' sample_covariance_matrix(dim = 3)

sample_covariance_matrix <- function(
    dim, df = dim, scale = diag(dim), diag = FALSE) {
  checkmate::assert_count(dim, positive = TRUE)
  checkmate::assert_int(df, lower = dim)
  assert_covariance_matrix(scale, dim = dim)
  checkmate::assert_flag(diag)
  cov <- rwishart(df = df, scale = scale, inv = TRUE)
  if (diag) cov[row(cov) != col(cov)] <- 0
  assert_covariance_matrix(cov)
  return(cov)
}

#' Grouping of a \code{data.frame}
#'
#' @description
#' This function groups a \code{data.frame} according to values of one column.
#'
#' @param df \[`data.frame`\]\cr
#' A \code{data.frame}.
#'
#' @param by \[`character(1)`\]\cr
#' The name of a column of \code{df} to group by.
#'
#' @param keep_by \[`logical(1)`\]\cr
#' Keep the grouping column \code{by}?
#'
#' @return
#' A \code{list} of \code{data.frame}s, subsets of \code{df}.
#'
#' @keywords transformation
#' @family data.frame helpers
#' @export group_data.frame
#' @usage group_data.frame(df, by, keep_by = TRUE)
#'
#' @examples
#' df <- data.frame("label" = c("A", "B"), "number" = 1:10)
#' group_data.frame(df = df, by = "label")
#' group_data.frame(df = df, by = "label", keep_by = FALSE)

group_data.frame <- function(df, by, keep_by = TRUE) {
  checkmate::assert_data_frame(df)
  checkmate::assert_names(by, subset.of = names(df))
  checkmate::assert_flag(keep_by)
  lapply(
    split(df, as.factor(df[, by])),
    delete_columns_data.frame,
    column_names = if (keep_by) character() else by
  )
}

test_that("can group data.frame", {
  df <- data.frame("label" = c("A", "B"), "number" = 1:10)
  expect_equal(
    group_data.frame(df = df, by = "label"),
    list(
      A = structure(
        list(label = c("A", "A", "A", "A", "A"), number = c(1L, 3L, 5L, 7L, 9L)),
        row.names = c(1L, 3L, 5L, 7L, 9L), class = "data.frame"
      ),
      B = structure(
        list(label = c("B", "B", "B", "B", "B"), number = c(2L, 4L, 6L, 8L, 10L)),
        row.names = c(2L, 4L, 6L, 8L, 10L), class = "data.frame"
      )
    )
  )
  expect_equal(
    group_data.frame(df = df, by = "label", keep_by = FALSE),
    list(
      A = structure(
        list(number = c(1L, 3L, 5L, 7L, 9L)),
        row.names = c(1L, 3L, 5L, 7L, 9L),
        class = "data.frame"
      ),
      B = structure(
        list(number = c(2L, 4L, 6L, 8L, 10L)),
        row.names = c(2L, 4L, 6L, 8L, 10L),
        class = "data.frame"
      )
    )
  )
})

#' Silence R code
#'
#' @description
#' This function silences warnings, messages and any `cat()` or `print()`
#' output from R expressions or functions.
#'
#' @param x \[`expression`\]\cr
#' Any function or expression or value assignment expression.
#'
#' @param print_cat \[`logical(1)`\]\cr
#' Silence `print()` and `cat()` outputs?
#'
#' @param message \[`logical(1)`\]\cr
#' Silence messages?
#'
#' @param warning \[`logical(1)`\]\cr
#' Silence warnings?
#'
#' @return
#' Invisibly the expression `x`.
#'
#' @keywords functional
#' @family function helpers
#' @export
#'
#' @references
#' This function is a modified version of \code{\link[spsUtil]{quiet}}.
#'
#' @examples
#' f <- function() {
#'   warning("warning")
#'   message("message")
#'   cat("cat")
#'   print("print")
#' }
#' quiet(f())

quiet <- function(
    x, print_cat = TRUE, message = TRUE, warning = TRUE
  ) {

  ### input checks
  input_check_response(checkmate::check_flag(print_cat), "print_cat")
  input_check_response(checkmate::check_flag(message), "message")
  input_check_response(checkmate::check_flag(warning), "warning")

  ### silence code
  if (print_cat) {
    sink(tempfile(), type = "output")
    on.exit(sink())
  }
  if (warning && message) {
    invisible(force(suppressMessages(suppressWarnings(x))))
  } else if (warning && !message) {
    invisible(suppressWarnings(force(x)))
  } else if (!warning && message) {
    invisible(suppressMessages(force(x)))
  } else {
    invisible(force(x))
  }
}

#' Difference and un-difference covariance matrix
#'
#' @description
#' These functions difference and un-difference random vectors and covariance
#' matrices.
#'
#' @param cov,cov_diff \[`matrix()`\]\cr
#' A (differenced) covariance matrix of dimension \code{dim}
#' (or \code{dim} - 1, respectively).
#'
#' @param ref \[`integer(1)`\]\cr
#' The reference row between \code{1} and \code{dim} for
#' differencing that maps \code{cov} to \code{cov_diff}, see details.
#'
#' @param dim \[`integer(1)`\]\cr
#' The matrix dimension.
#'
#' @param ranking \[`integer()`\]\cr
#' The integers `1` to `dim` in arbitrary order.
#'
#' @return
#' A (differenced or un-differenced) covariance \code{matrix}.
#'
#' @details
#' Assume \eqn{x \sim N(0, \Sigma)} is a multivariate normally distributed
#' random vector of dimension \eqn{n}. We may want to consider the differenced
#' vector \deqn{\tilde x = (x_1 - x_k, x_2 - x_k, \dots, x_n - x_k)',} excluding
#' the \eqn{k}th element (hence, \eqn{\tilde x} is of dimension
#' \eqn{(n - 1) \times 1}). Formally, \eqn{\tilde x = \Delta_k x}, where
#' \eqn{\Delta_k} is a difference operator that depends on the reference
#' row \eqn{k}. More precise, \eqn{\Delta_k} is the identity matrix of dimension
#' \eqn{n} without row \eqn{k} and with \eqn{-1}s in column \eqn{k}.
#' The difference operator \eqn{\Delta_k} can be computed via
#' \code{delta(ref = k, dim = n)}.
#'
#' Then, \eqn{\tilde x \sim N(0, \tilde \Sigma)}, where
#' \deqn{\tilde{\Sigma} = \Delta_k \Sigma \Delta_k'}
#' is the differenced covariance matrix with respect to row \eqn{k = 1,\dots,n}.
#' The differenced covariance matrix \eqn{\tilde \Sigma} can be computed via
#' \code{diff_delta(Sigma, ref = k)}.
#'
#' Since \eqn{\Delta_k} is a non-bijective mapping, \eqn{\Sigma} cannot be
#' uniquely restored from \eqn{\tilde \Sigma}. However, it is possible to
#' compute a non-unique solution \eqn{\Sigma_0}, such that
#' \eqn{\Delta_k \Sigma_0 \Delta_k = \tilde \Sigma}. For such a non-unique
#' solution, we add a column and a row of zeros
#' at column and row number \eqn{k} to \eqn{\tilde{\Sigma}}, respectively.
#' An "un-differenced" covariance matrix \eqn{\Sigma_0} can be computed via
#' \code{undiff_delta(Sigma_diff, ref = k)}.
#'
#' As a alternative to \eqn{\Delta_k}, the function `M()` returns a matrix for
#' taking differences such that the resulting vector is negative.
#'
#' @keywords transformation
#' @family matrix helpers
#' @export
#'
#' @examples
#' n <- 4
#' Sigma <- sample_covariance_matrix(dim = n)
#' k <- 2
#' x <- c(1, 3, 2, 4)
#'
#' # build difference operator
#' delta_k <- delta(ref = k, dim = n)
#'
#' # difference vector
#' delta_k %*% x
#'
#' # difference Sigma
#' (Sigma_diff <- diff_cov(Sigma, ref = k))
#'
#' # un-difference Sigma
#' (Sigma_0 <- undiff_cov(Sigma_diff, ref = k))
#'
#' # difference again
#' Sigma_diff_2 <- diff_cov(Sigma_0, ref = k)
#' all.equal(Sigma_diff, Sigma_diff_2)
#'
#' # difference such that the resulting vector is negative
#' M(ranking = order(x, decreasing = TRUE), dim = n) %*% x

diff_cov <- function(cov, ref = 1) {
  input_check_response(
    check = check_covariance_matrix(cov),
    var_name = "cov"
  )
  dim <- nrow(cov)
  input_check_response(
    check = checkmate::check_int(ref, lower = 1, upper = dim),
    var_name = "ref"
  )
  D <- delta(ref = ref, dim = dim)
  D %*% cov %*% t(D)
}

#' @rdname diff_cov
#' @export

undiff_cov <- function(cov_diff, ref = 1) {
  input_check_response(
    check = check_covariance_matrix(cov_diff),
    var_name = "cov_diff"
  )
  dim <- nrow(cov_diff) + 1
  input_check_response(
    check = checkmate::check_int(ref, lower = 1, upper = dim),
    var_name = "ref"
  )
  cov <- matrix(0, dim, dim)
  cov[row(cov) != ref & col(cov) != ref] <- cov_diff
  cov
}

#' @rdname diff_cov
#' @export

delta <- function(ref = 1, dim) {
  input_check_response(
    check = checkmate::check_int(dim, lower = 2),
    var_name = "dim"
  )
  input_check_response(
    check = checkmate::check_int(ref, lower = 1, upper = dim),
    var_name = "ref"
  )
  D <- diag(dim)
  D[, ref] <- -1
  D[-ref, , drop = FALSE]
}

#' @rdname diff_cov
#' @export

M <- function(ranking = seq_len(dim), dim) {
  input_check_response(
    check = checkmate::check_int(dim, lower = 2),
    var_name = "dim"
  )
  input_check_response(
    check = checkmate::check_integerish(
      ranking, lower = 1, upper = dim, len = dim, unique = TRUE,
      any.missing = FALSE
    ),
    var_name = "ranking"
  )
  M <- matrix(0, nrow = dim - 1, ncol = dim)
  for (i in seq_len(dim - 1)) {
    M[i, ranking[i]] <- -1
    M[i, ranking[i + 1]] <- 1
  }
  return(M)
}

test_that("differencing works", {
  expect_error(
    delta(0, 3),
    "Input `ref` is bad: Element 1 is not >= 1"
  )
  expect_equal(
    delta(1, 3),
    structure(c(-1, -1, 1, 0, 0, 1), dim = 2:3)
  )
  expect_equal(
    delta(2, 3),
    structure(c(1, 0, -1, -1, 0, 1), dim = 2:3)
  )
  expect_equal(
    delta(3, 3),
    structure(c(1, 0, 0, 1, -1, -1), dim = 2:3)
  )
  expect_error(
    delta(4, 3),
    "Input `ref` is bad: Element 1 is not <= 3"
  )
  n <- sample(2:10, 1)
  ref <- sample.int(n, 1)
  Sigma0 <- sample_covariance_matrix(dim = n)
  expect_silent(assert_covariance_matrix(Sigma0))
  Sigma_diff0 <- diff_cov(Sigma0, ref = ref)
  expect_silent(assert_covariance_matrix(Sigma_diff0))
  Sigma1 <- undiff_cov(Sigma_diff0, ref = ref)
  expect_silent(assert_covariance_matrix(Sigma1))
  Sigma_diff1 <- diff_cov(Sigma1, ref = ref)
  expect_silent(assert_covariance_matrix(Sigma_diff1))
  expect_equal(Sigma_diff0, Sigma_diff1)
})

test_that("matrix M works", {
  dim <- 4
  x <- rnorm(4)
  ranking <- order(x, decreasing = TRUE)
  expect_true(
    all(M(ranking = ranking, dim = dim) %*% x < 0)
  )
})

#' Namespace calls
#'
#' @description
#' This function searches for namespace calls in `.R` files, i.e., code lines of
#' the format `<package name>::<function name>`.
#'
#' @param path \[`character(1)`\]\cr
#' The path name to a folder. All `.R` files in this folder and sub-directories
#' will be searched.
#'
#' @param triple_colon \[`logical(1)`\]\cr
#' Also search for `:::`?
#'
#' @param as_list \[`logical(1)`\]\cr
#' Simplify the output into a `list` of unique function names per package?
#'
#' @return
#' A `data.frame`. If `as_list = TRUE`, a `list`.
#'
#' @keywords packaging
#' @family package helpers
#' @export
#'
#' @examples
#' \dontrun{
#' find_namespace_calls()
#' find_namespace_calls(as_list = TRUE)
#' }

find_namespace_calls <- function(
    path = "R", triple_colon = FALSE, as_list = FALSE
  ) {

  ### input checks
  input_check_response(
    check = checkmate::check_directory_exists(path, access = "r"),
    var_name = "path"
  )
  input_check_response(
    check = checkmate::check_flag(triple_colon),
    var_name = "triple_colon"
  )
  input_check_response(
    check = checkmate::check_flag(as_list),
    var_name = "as_list"
  )

  ### list .R files under path
  pattern <- "\\.R$"
  r_files <- list.files(
    path = path, pattern = pattern, full.names = TRUE, recursive = TRUE
  )

  ### regular expressions to match :: and ::: calls followed by a character that
  ### is NOT part of a valid function name
  pattern_double <- "\\b([a-zA-Z0-9\\.]+)::([a-zA-Z0-9_\\.]+)(?![a-zA-Z0-9_\\.])"
  pattern_triple <- "\\b([a-zA-Z0-9\\.]+):::+([a-zA-Z0-9_\\.]+)(?![a-zA-Z0-9_\\.])"
  pattern <- if (triple_colon) {
    paste0("(", pattern_double, "|", pattern_triple, ")")
  } else {
    pattern_double
  }

  ### search
  matches <- list()

  for (file in r_files) {
    lines <- readLines(file, warn = FALSE)

    for (i in seq_along(lines)) {
      line <- lines[i]
      found <- gregexpr(pattern, line, perl = TRUE)[[1]]

      if (found[1] != -1) {
        matched_texts <- regmatches(line, gregexpr(pattern, line, perl = TRUE))[[1]]
        for (match in matched_texts) {
          if (grepl(":::", match)) {
            pkg <- sub("^([a-zA-Z0-9\\.]+):::+.*", "\\1", match)
            fun <- sub("^[a-zA-Z0-9\\.]+:::+([a-zA-Z0-9_\\.]+).*", "\\1", match)
            full_call <- paste0(pkg, "::: ", fun)
          } else {
            pkg <- sub("^([a-zA-Z0-9\\.]+)::.*", "\\1", match)
            fun <- sub("^[a-zA-Z0-9\\.]+::([a-zA-Z0-9_\\.]+).*", "\\1", match)
            full_call <- paste0(pkg, "::", fun)
          }
          matches <- append(matches, list(list(
            file = file,
            line_number = i,
            call = full_call,
            pkg = pkg,
            fun = fun
          )))
        }
      }
    }
  }

  ### prepare output
  df <- if (length(matches) == 0) {
    data.frame(file = character(0), line_number = integer(0), call = character(0))
  } else {
    do.call(rbind, lapply(matches, as.data.frame, stringsAsFactors = FALSE))
  }
  if (as_list) {
    if (length(matches) == 0) {
      return(list())
    }
    return(lapply(split(df$fun, df$pkg), unique))
  }
  return(df)
}

#' Multivariate normal distribution
#'
#' @description
#' The function `dmvnorm()` computes the density of a multivariate normal
#' distribution.
#'
#' The function `pmvnorm()` computes the cumulative distribution function of a
#' multivariate normal distribution.
#'
#' The function `rmvnorm()` samples from a multivariate normal distribution.
#'
#' The functions with suffix `_cpp` perform no input checks, hence are faster.
#'
#' The univariate normal distribution is available as the special case `p = 1`.
#'
#' @details
#' `pmvnorm()` just calls `mvtnorm::pmvnorm` with the randomized
#' Quasi-Monte-Carlo procedure by Genz and Bretz. The argument `abseps` controls
#' the accuracy of the Gaussian integral approximation.
#'
#' @param x \[`numeric()`\]\cr
#' A quantile vector of length `p`.
#'
#' @param mean \[`numeric()`\]\cr
#' The mean vector of length `p`.
#'
#' For the functions without suffix `_cpp`, it can also be of length `1` for
#' convenience, then `rep(mean, p)` is considered.
#'
#' @param Sigma \[`matrix()`\]\cr
#' The covariance matrix of dimension `p`.
#'
#' For `rmvnorm()`, arbitrary dimensions (i.e., full rows and corresponding
#' columns) of `Sigma` can be `0`.
#'
#' For the functions without suffix `_cpp` and if `p = 1`, it can also be a
#' single `numeric` for convenience. Note that `Sigma` is this case is a
#' variance, which is a different format than in `stats::dnorm()` or
#' `stats::rnorm`, which require a standard deviation.
#'
#' @param log \[`logical(1)`\]\cr
#' Consider the log-normal distribution?
#'
#' @param abseps \[`numeric(1)`\]\cr
#' The absolute error tolerance.
#'
#' @param n \[`integer(1)`\]\cr
#' The number of requested samples.
#'
#' @return
#' For `dmvnorm()`: The density value.
#'
#' For `pmvnorm()`: The value of the distribution function.
#'
#' For `rmvnorm()`: If \code{n = 1} a \code{vector} of length \code{p} (note
#' that it is a column vector for `rmvnorm_cpp()`), else
#' a \code{matrix} of dimension \code{n} times \code{p} with samples as rows.
#'
#' @keywords distribution
#' @family simulation helpers
#' @export
#'
#' @examples
#' x <- c(0, 0)
#' mean <- c(0, 0)
#' Sigma <- diag(2)
#'
#' # compute density
#' dmvnorm(x = x, mean = mean, Sigma = Sigma)
#' dmvnorm(x = x, mean = mean, Sigma = Sigma, log = TRUE)
#'
#' # compute CDF
#' pmvnorm(x = x, mean = mean, Sigma = Sigma)
#'
#' # sample
#' rmvnorm(n = 3, mean = mean, Sigma = Sigma)
#' rmvnorm(mean = mean, Sigma = Sigma, log = TRUE)

dmvnorm <- function(x, mean, Sigma, log = FALSE) {
  input_check_response(
    check = check_numeric_vector(x, any.missing = FALSE, min.len = 1),
    var_name = "x"
  )
  dim <- length(x)
  if (checkmate::test_atomic_vector(mean, len = 1)) {
    mean <- rep(mean, dim)
  }
  input_check_response(
    check = check_numeric_vector(mean, any.missing = FALSE, len = dim),
    var_name = "mean"
  )
  if (dim == 1 && checkmate::test_atomic_vector(Sigma, len = 1)) {
    Sigma <- as.matrix(Sigma)
  }
  input_check_response(
    check = check_covariance_matrix(Sigma, dim = dim),
    var_name = "Sigma"
  )
  input_check_response(
    check = checkmate::check_flag(log),
    var_name = "log"
  )
  dmvnorm_cpp(x, mean, Sigma, log)
}

#' @rdname dmvnorm
#' @export

pmvnorm <- function(x, mean, Sigma, abseps = 1e-3) {
  input_check_response(
    check = check_numeric_vector(x, any.missing = FALSE, min.len = 1),
    var_name = "x"
  )
  dim <- length(x)
  if (checkmate::test_atomic_vector(mean, len = 1)) {
    mean <- rep(mean, dim)
  }
  input_check_response(
    check = check_numeric_vector(mean, any.missing = FALSE, len = dim),
    var_name = "mean"
  )
  if (dim == 1 && checkmate::test_atomic_vector(Sigma, len = 1)) {
    Sigma <- as.matrix(Sigma)
  }
  input_check_response(
    check = check_covariance_matrix(Sigma, dim = dim),
    var_name = "Sigma"
  )
  input_check_response(
    check = checkmate::check_number(abseps, lower = 0, finite = TRUE),
    var_name = "abseps"
  )
  pmvnorm_cpp(x, mean, Sigma, abseps)
}

#' @rdname dmvnorm
#' @export

rmvnorm <- function(n = 1, mean, Sigma, log = FALSE) {

  ### input checks
  input_check_response(
    check = checkmate::check_int(n),
    var_name = "n"
  )
  if (checkmate::test_atomic_vector(Sigma, len = 1)) {
    Sigma <- as.matrix(Sigma)
  }
  input_check_response(
    check = checkmate::check_matrix(Sigma, mode = "numeric"),
    var_name = "Sigma"
  )
  dim <- nrow(Sigma)
  if (checkmate::test_atomic_vector(mean, len = 1)) {
    mean <- rep(mean, dim)
  }
  input_check_response(
    check = check_numeric_vector(mean, any.missing = FALSE, len = dim),
    var_name = "mean"
  )
  input_check_response(
    check = checkmate::check_flag(log),
    var_name = "log"
  )

  ### zero dimensions case
  zero_dims <- which(
    apply(Sigma, 1, function(row) all(row == 0)) &
    apply(Sigma, 2, function(col) all(col == 0))
  )
  if (length(zero_dims) == dim) {
    ### the case where Sigma = 0 is handeled in rmvnorm_cpp
    zero_dims <- integer()
  }
  if (length(zero_dims) > 0) {
    mean_det <- mean[zero_dims]
    mean <- mean[-zero_dims]
    Sigma <- Sigma[-zero_dims, -zero_dims, drop = FALSE]
  }
  input_check_response(
    check = check_covariance_matrix(Sigma),
    var_name = "Sigma"
  )

  ### sample
  sample <- replicate(n = n, rmvnorm_cpp(mean, Sigma, log), simplify = TRUE)
  if (n == 1) {
    if (length(zero_dims) > 0) {
      out <- numeric(dim)
      out[-zero_dims] <- drop(sample)
      out[zero_dims] <- mean_det
      if (log) out[zero_dims] <- exp(out[zero_dims])
    } else {
      out <- drop(sample)
    }
  } else if (dim == 1) {
    out <- as.matrix(sample)
  } else {
    if (length(zero_dims) > 0) {
      out <- matrix(0, nrow = n, ncol = dim)
      out[, -zero_dims] <- t(sample)
      out[, zero_dims] <- matrix(
        mean_det, nrow = n, ncol = length(zero_dims), byrow = TRUE
      )
      if (log) out[, zero_dims] <- exp(out[, zero_dims])
    } else {
      out <- t(sample)
    }
  }

  ### return
  return(out)
}

test_that("Univariate normal density can be computed", {
  x <- 1
  mean <- 2
  Sigma <- 4
  expect_equal(dmvnorm(x, mean, Sigma), dnorm(x, mean, sqrt(Sigma)))
  expect_error(
    dmvnorm(x = 1:2, mean = 1:3, Sigma = diag(4)),
    "Input `mean` is bad: Must have length 2, but has length 3"
  )
})

test_that("Multivariate normal density can be computed", {
  x <- c(0, 0)
  mean <- c(0, 0)
  Sigma <- diag(2)
  factor <- 1000
  expect_equal(
    round(dmvnorm(x = x, mean = mean, Sigma = Sigma, log = FALSE) * factor) / factor,
    0.159
  )
  expect_equal(
    round(dmvnorm(x = x, mean = mean[1], Sigma = Sigma, log = FALSE) * factor) / factor,
    0.159
  )
  expect_equal(
    round(dmvnorm(x = x, mean = mean, Sigma = Sigma, log = TRUE) * factor) / factor,
    -1.838
  )
  expect_error(
    dmvnorm(x = 1:2, mean = 1:3, Sigma = diag(2)),
    "Input `mean` is bad: Must have length 2, but has length 3"
  )
})

test_that("Univariate normal CDF can be computed", {
  x <- 1
  mean <- 2
  Sigma <- 4
  expect_equal(pmvnorm(x, mean, Sigma), pnorm(x, mean, sqrt(Sigma)))
  expect_error(
    pmvnorm(x = 1:2, mean = 1:3, Sigma = diag(4)),
    "Input `mean` is bad: Must have length 2, but has length 3"
  )
})

test_that("Multivariate normal CDF can be computed", {
  x <- c(-0.78, -1.17)
  mean <- c(0.64, 0.97)
  Sigma <- matrix(c(3.35, 1.45, 1.45, 3.03), ncol = 2)
  factor <- 1000
  expect_equal(
    round(pmvnorm(x = x, mean = mean, Sigma = Sigma) * factor) / factor,
    0.055
  )
  expect_error(
    pmvnorm(x = 1:2, mean = 1:3, Sigma = diag(4)),
    "Input `mean` is bad: Must have length 2, but has length 3"
  )
})

test_that("Univariate normal can be drawn", {
  mean <- 0
  Sigma <- 1
  expect_length(rmvnorm(mean = mean, Sigma = Sigma, log = FALSE), 1)
  expect_length(rmvnorm(mean = mean, Sigma = Sigma, log = TRUE), 1)
  checkmate::expect_matrix(
    rmvnorm(n = 5, mean = mean, Sigma = Sigma),
    ncols = 1, nrows = 5
  )
})

test_that("Multivariate normal can be drawn", {
  mean <- c(0, 0)
  Sigma <- diag(2)
  expect_length(rmvnorm(mean = mean, Sigma = Sigma, log = FALSE), 2)
  expect_length(rmvnorm(mean = mean, Sigma = Sigma, log = TRUE), 2)
  checkmate::expect_matrix(
    rmvnorm(n = 5, mean = mean, Sigma = Sigma),
    ncols = 2, nrows = 5
  )
  expect_error(rmvnorm(mean = 1:3, Sigma = diag(4)))
})

test_that("Multivariate normal with zero dimensions can be drawn", {
  dim <- 4
  zero_dim <- c(2, 3)
  mean <- round(rnorm(dim), 2)
  Sigma <- sample_covariance_matrix(dim = dim)
  Sigma[zero_dim, ] <- Sigma[, zero_dim] <- 0

  ### n = 1
  sample <- rmvnorm(n = 1, mean = mean, Sigma = Sigma, log = FALSE)
  expect_equal(sample[zero_dim], mean[zero_dim])
  sample <- rmvnorm(n = 1, mean = mean, Sigma = Sigma, log = TRUE)
  expect_equal(sample[zero_dim], exp(mean[zero_dim]))

  ### n > 1
  n <- 3
  sample <- rmvnorm(n = n, mean = mean, Sigma = Sigma, log = FALSE)
  expect_equal(
    sample[, zero_dim],
    matrix(mean[zero_dim], nrow = n, ncol = length(zero_dim), byrow = TRUE)
  )
  sample <- rmvnorm(n = n, mean = mean, Sigma = Sigma, log = TRUE)
  expect_equal(
    sample[, zero_dim],
    exp(matrix(mean[zero_dim], nrow = n, ncol = length(zero_dim), byrow = TRUE))
  )

  ### dim = 1
  mean <- round(rnorm(1), 2)
  expect_equal(
    rmvnorm(n = 1, mean = mean, Sigma = 0, log = FALSE),
    mean
  )
  expect_equal(
    rmvnorm(n = 1, mean = mean, Sigma = 0, log = TRUE),
    exp(mean)
  )
  n <- 3
  expect_equal(
    rmvnorm(n = n, mean = mean, Sigma = 0, log = FALSE),
    matrix(mean, nrow = n)
  )
  expect_equal(
    rmvnorm(n = n, mean = mean, Sigma = 0, log = TRUE),
    matrix(exp(mean), nrow = n)
  )
})

# hermann_web <- rvest::read_html("https://de.wikipedia.org/wiki/Hermannslauf") |>
#   rvest::html_nodes("table") |>
#   rvest::html_table(fill = TRUE)
#
# hermann_men <- hermann_web[[3]] |>
#   dplyr::select(1:3) |>
#   dplyr::filter(Jahr != 2020) |>
#   dplyr::mutate(Zeit = lubridate::hms(Zeit)) |>
#   dplyr::rename("year" = "Jahr", "winner_men" = "Männer", "time_men" = "Zeit") |>
#   tsibble::as_tsibble(index = year) |>
#   tsibble::fill_gaps()
#
# hermann_women <- hermann_web[[3]] |>
#   dplyr::select(1, 4:5) |>
#   dplyr::filter(Jahr != 2020) |>
#   dplyr::mutate(Zeit = lubridate::hms(Zeit)) |>
#   dplyr::rename("year" = "Jahr", "winner_women" = "Frauen", "time_women" = "Zeit") |>
#   tsibble::as_tsibble(index = year) |>
#   tsibble::fill_gaps()
#
# hermann_weather <- tibble::tribble(
#   ~edition, ~year, ~date,    ~temp,
#   1,        1972,  "16.04.", NA,
#   2,        1973,  "29.04.", 14.0,
#   3,        1974,  "28.04.", 14.0,
#   4,        1975,  "27.04.", 11.0,
#   5,        1976,  "25.04.", 11.0,
#   6,        1977,  "24.04.", 9.0,
#   7,        1978,  "30.04.", 16.0,
#   8,        1979,  "29.04.", 7.0,
#   9,        1980,  "27.04.", 9.0,
#   10,       1981,  "26.04.", 16.0,
#   11,       1982,  "25.04.", 11.4,
#   12,       1983,  "24.04.", 10.0,
#   13,       1984,  "29.04.", 10.2,
#   14,       1985,  "28.04.", 6.9,
#   15,       1986,  "27.04.", 15.0,
#   16,       1987,  "26.04.", 15.9,
#   17,       1988,  "24.04.", 6.0,
#   18,       1989,  "30.04.", 11.9,
#   19,       1990,  "29.04.", 12.9,
#   20,       1991,  "28.04.", 7.3,
#   21,       1992,  "26.04.", 17.7,
#   22,       1993,  "25.04.", 20.1,
#   23,       1994,  "24.04.", 19.9,
#   24,       1995,  "30.04.", 12.2,
#   25,       1996,  "28.04.", 12.7,
#   26,       1997,  "27.04.", 11.0,
#   27,       1998,  "26.04.", 10.7,
#   28,       1999,  "25.04.", 15.0,
#   29,       2000,  "30.04.", 14.2,
#   30,       2001,  "29.04.", 14.4,
#   31,       2002,  "28.04.", 8.8,
#   32,       2003,  "27.04.", 13.7,
#   33,       2004,  "25.04.", 13.5,
#   34,       2005,  "24.04.", 16.1,
#   35,       2006,  "30.04.", 7.1,
#   36,       2007,  "29.04.", 15.8,
#   37,       2008,  "27.04.", 18.7,
#   38,       2009,  "26.04.", 20.6,
#   39,       2010,  "25.04.", 20.9,
#   40,       2011,  "17.04.", 15.5,
#   41,       2012,  "29.04.", 16.4,
#   42,       2013,  "28.04.", 7.5,
#   43,       2014,  "27.04.", 18.0,
#   44,       2015,  "26.04.", 14.0,
#   45,       2016,  "24.04.", 6.0,
#   46,       2017,  "30.04.", 12.0,
#   47,       2018,  "29.04.", 16.0,
#   48,       2019,  "28.04.", 10.0,
#   NA,       2020,  NA,       NA,
#   49,       2021,  "10.10.", 11.7,
#   50,       2022,  "24.04.", 12.7,
#   51,       2023,  "30.04.", 11.6,
#   52,       2024,  "28.04.", 17.6,
#   53,       2025,  "27.04.", 15.3
# )
#
# hermann <- hermann_weather |>
#   dplyr::left_join(hermann_men, by = "year") |>
#   dplyr::left_join(hermann_women, by = "year") |>
#   dplyr::mutate(date = lubridate::dmy(paste0(date, year))) |>
#   dplyr::mutate(
#     time_men = lubridate::period_to_seconds(time_men),
#     time_women = lubridate::period_to_seconds(time_women)
#   ) |>
#   dplyr::rename(seconds_men = time_men, seconds_women = time_women)
#
# usethis::use_data(hermann, overwrite = TRUE)

#' Hermannslauf
#'
#' @description
#' The Hermannslauf is an annual long-distance road and trail race in Germany
#' that runs from Detmold to Bielefeld through the Teutoburg Forest. It is one
#' of the best-known running events in the region and is especially noted for
#' its demanding, hilly course of about 31 kilometers. This dataset contains
#' historical information on editions of the race, including the date,
#' temperature, and winning times for men and women.
#'
#' From 1972 to 1976, the course followed the actual Hermannsweg for 30.4
#' kilometers, and in 1977 the first 13 kilometers were moved to a parallel
#' route. Since 2005, the total course length has been 31.1 kilometers; until
#' 2004, it was 30.6 kilometers. In 2020, the Hermannslauf was canceled due to
#' the pandemic, and for the same reason, in 2021 it was exceptionally moved
#' from April to October.
#'
#' @details
#' Temperature in degrees Celsius, measured at 12:00 noon, obtained from
#' different sources:
#' - from 1973 to 1992 and from 1998 to 2013 from
#' <https://de.weatherspark.com/>, Gütersloh Airport (approx. 30 km from start
#' and finish)
#' - from 1993 to 1997 from <https://de.weatherspark.com/>, Wunstorf Air Base
#' (approx. 100 km from start and finish)
#' - from 2014 to 2017 from <https://de.weatherspark.com/>, Hannover-Langenhagen
#' Airport (approx. 90 km from start and finish)
#' - from 2018 onward from <https://meteostat.net/de/station/D7106>,
#' Airfield Bielefeld-Windelsbleiche (approx. 20 km from start and finish)
#'
#' @format
#' A `tibble` with 54 rows and 8 columns:
#' \describe{
#'   \item{edition}{the edition of the Hermannslauf}
#'   \item{year}{the year}
#'   \item{date}{the date}
#'   \item{temp}{the temperature on that day at 12:00 noon; see details}
#'   \item{winner_men}{the men's winner}
#'   \item{seconds_men}{the men's winner's total time in seconds}
#'   \item{winner_women}{the women's winner}
#'   \item{seconds_women}{the women's winner's total time in seconds}
#' }
#'
#' @source <https://de.wikipedia.org/wiki/Hermannslauf>
#'
#' @keywords data
#' @family data

"hermann"

#' Get \code{matrix} indices
#'
#' @description
#' This function returns \code{matrix} indices as \code{character}.
#'
#' @param x \[`matrix`\]\cr
#' A \code{matrix}.
#'
#' @param prefix \[`character(1)`\]\cr
#' A prefix for the indices.
#'
#' @param exclude_diagonal \[`logical(1)`\]\cr
#' Exclude indices where row equals column?
#'
#' @return
#' A \code{character} \code{vector}.
#'
#' @export
#' @keywords indexing
#' @family matrix helpers
#'
#' @examples
#' M <- diag(3)
#' matrix_indices(M)
#' matrix_indices(M, "M_")
#' matrix_indices(M, "M_", TRUE)

matrix_indices <- function(x, prefix = "", exclude_diagonal = FALSE) {
  checkmate::assert_matrix(x)
  checkmate::assert_string(prefix)
  checkmate::assert_flag(exclude_diagonal)
  rows <- nrow(x)
  cols <- ncol(x)
  indices <- expand.grid(seq_len(rows), seq_len(cols))
  if (exclude_diagonal) {
    indices <- indices[apply(indices, 1, function(x) length(unique(x)) != 1), ]
  }
  indices <- apply(indices, 1, paste, collapse = "")
  paste0(prefix, indices)
}

#' Best-possible match of two numeric vectors
#'
#' @description
#' This function matches the indices of two numeric vectors as good as possible
#' (that means with the smallest possible sum of deviations).
#'
#' @param x,y \[`numeric()`\]\cr
#' Two vectors of the same length.
#'
#' @return
#' An \code{integer} vector of length \code{length(x)} with the positions of
#' \code{y} in \code{x}.
#'
#' @export
#' @keywords indexing
#' @family vector helpers
#'
#' @examples
#' x <- c(-1, 0, 1)
#' y <- c(0.1, 1.5, -1.2)
#' match_numerics(x, y)

match_numerics <- function(x, y) {
  checkmate::assert_atomic_vector(x)
  checkmate::assert_numeric(x)
  checkmate::assert_atomic_vector(y, len = length(x))
  checkmate::assert_numeric(y)
  matches <- numeric(length(x))
  distances <- unique(sort(stats::dist(c(x, y)))) + sqrt(.Machine$double.eps)
  for (d in distances) {
    if (any(c(!is.na(x), !is.na(y)))) {
      for (i_x in 1:length(x)) {
        for (i_y in 1:length(y)) {
          if (!is.na(x[i_x]) && !is.na(y[i_y])) {
            if (isTRUE(all.equal(x[i_x], y[i_y], d))) {
              matches[i_y] <- i_x
              x[i_x] <- NA_integer_
              y[i_y] <- NA_integer_
            }
          }
        }
      }
    }
  }
  return(matches)
}

#' Generate vector subsets
#'
#' @description
#' This function generates subsets of a vector.
#'
#' @param v \[atomic()`\]\cr
#' A vector of elements.
#'
#' @param n \[integer(1)`\]\cr
#' The requested subset sizes.
#'
#' @return
#' A \code{list}, each element is a subset of \code{v}.
#'
#' @export
#' @keywords indexing
#' @family vector helpers
#'
#' @examples
#' v <- 1:3
#' subsets(v)
#' subsets(v, c(1, 3)) # only subsets of length 1 or 3
#' subsets(integer())  # trivial case works

subsets <- function(v, n = seq_along(v)) {
  input_check_response(
    check = checkmate::check_atomic_vector(v),
    var_name = "v"
  )
  input_check_response(
    check = checkmate::check_integerish(
      n, lower = 1, upper = length(v), any.missing = FALSE, unique = TRUE
    ),
    var_name = "n"
  )
  if (length(v) == 0 || length(n) == 0) {
    return(list())
  }
  unlist(lapply(n, function(i) {
    utils::combn(v, i, simplify = FALSE)
  }), recursive = FALSE)
}

test_that("generating subsets works", {
  v <- 1:3
  expect_equal(
    subsets(v),
    list(1L, 2L, 3L, 1:2, c(1L, 3L), 2:3, 1:3)
  )
  expect_equal(
    subsets(v, n = c(1, 3)),
    list(1L, 2L, 3L, 1:3)
  )
  expect_equal(
    subsets(v = integer()),
    list()
  )
  expect_equal(
    subsets(v, n = integer()),
    list()
  )
  expect_equal(
    subsets(LETTERS[1:3]),
    list("A", "B", "C", c("A", "B"), c("A", "C"), c("B", "C"), c("A", "B", "C"))
  )
})

#' Insert column in matrix
#'
#' @description
#' This function inserts a column into a matrix.
#'
#' @param A \[`matrix()`\]\cr
#' A \code{matrix}.
#'
#' @param x \[`atomic()`\]\cr
#' The column to be added, of length \code{nrow(A)}.
#'
#' Can also be a single value.
#'
#' @param p \[`integer())`\]\cr
#' The position(s) where to add the column, one or more of:
#' - \code{p = 0} appends the column left
#' - \code{p = ncol(A)} appends the column right
#' - \code{p = n} inserts the column between the \code{n}-th and
#'   \code{(n + 1)}-th column of \code{A}.
#'
#' @return
#' A \code{matrix}.
#'
#' @export
#' @keywords transformation
#' @family matrix helpers
#'
#' @examples
#' A <- diag(3)
#' x <- 1:3
#' insert_matrix_column(A, x, 0)
#' insert_matrix_column(A, x, 1)
#' insert_matrix_column(A, x, 2)
#' insert_matrix_column(A, x, 3)
#'
#' ### also single value
#' x <- 2
#' insert_matrix_column(A, x, 0)
#'
#' ### also multiple positions
#' insert_matrix_column(A, x, 0:3)
#'
#' ### also trivial case
#' insert_matrix_column(matrix(nrow = 0, ncol = 0), integer(), integer())

insert_matrix_column <- function(A, x, p) {
  checkmate::assert_matrix(A)
  n <- ncol(A)
  checkmate::assert_atomic_vector(x)
  stopifnot(
    "'x' must be of length 1 or nrow(A)" = length(x) == 1 || length(x) == nrow(A)
  )
  checkmate::assert_integerish(
    p, lower = 0, upper = n, unique = TRUE, max.len = n + 1
  )
  if (length(p) == 0) {
    return(A)
  }
  p <- sort(p)
  for (i in seq_along(p)) {
    if (p[i] == 0) {
      A <- cbind(x, A, deparse.level = 0)
    } else if (p[i] == n) {
      A <- cbind(A, x, deparse.level = 0)
    } else {
      A <- cbind(
        A[, seq_len(p[i]), drop = FALSE], x, A[, (p[i] + 1):n, drop = FALSE],
        deparse.level = 0
      )
    }
    p <- p + 1
    n <- ncol(A)
  }
  return(A)
}

test_that("insert matrix column works", {
  A <- diag(3)
  x <- numeric(3)
  expect_equal(
    insert_matrix_column(A, x, 0),
    structure(c(0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 1), dim = 3:4)
  )
  expect_equal(
    insert_matrix_column(A, x, 1),
    structure(c(1, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 1), dim = 3:4)
  )
  expect_equal(
    insert_matrix_column(A, x, 2),
    structure(c(1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 1), dim = 3:4)
  )
  expect_equal(
    insert_matrix_column(A, x, 3),
    structure(c(1, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0), dim = 3:4)
  )
})

test_that("insert matrix column at multiple position works", {
  A <- diag(3)
  x <- 1:3
  expect_equal(
    insert_matrix_column(A, x, 0:3),
    structure(
      c(1, 2, 3, 1, 0, 0, 1, 2, 3, 0, 1, 0, 1, 2, 3, 0, 0, 1, 1, 2, 3),
      dim = c(3L, 7L)
    )
  )
})

test_that("insert single value as matrix column works", {
  A <- diag(3)
  x <- 2
  expect_equal(
    insert_matrix_column(A, x, 0),
    structure(c(2, 2, 2, 1, 0, 0, 0, 1, 0, 0, 0, 1), dim = 3:4)
  )
})

test_that("insert matrix column trivial case works", {
  expect_equal(
    insert_matrix_column(matrix(nrow = 0, ncol = 0), integer(), integer()),
    structure(logical(0), dim = c(0L, 0L))
  )
})

#' Standardized response to input check
#'
#' @description
#' This function provides a standardized response to input checks, ensuring
#' consistency.
#'
#' @param check \[`TRUE` | `character(1)` | `list()`\]\cr
#' Matches the return value of the `check*` functions from the `{checkmate}`
#' package, i.e., either `TRUE` if the check was successful, or a `character`
#' (the error message) else.
#'
#' Can also be a `list` of multiple such values for alternative criteria, where
#' at least one must be `TRUE` for a successful check.
#'
#' @param var_name \[`NULL` | `character(1)`\]\cr
#' Optionally specifies the name of the input being checked. This name will be
#' used for the default value of the `prefix` argument.
#'
#' @param error \[`logical(1)`\]\cr
#' If `check` is not `TRUE` (or no element in `check` is `TRUE`, if `check` is
#' a `list`), throw an error?
#'
#' @param prefix \[`character(1)`\]\cr
#' A prefix for the thrown error message, only relevant if `error` is `TRUE`.
#'
#' @return
#' `TRUE` if `check` is `TRUE` (or any element in `check` is `TRUE`, if `check`
#' is a `list`) . Else, depending on `error`:
#'
#' - If `error` is `TRUE`, throws an error.
#' - If `error` is `FALSE`, returns `FALSE`.
#'
#' @export
#' @keywords packaging
#' @family package helpers
#'
#' @examples
#' x <- "1"
#' y <- 1
#'
#' ### check is successful
#' input_check_response(
#'   check = checkmate::check_character(x),
#'   var_name = "x",
#'   error = TRUE
#' )
#'
#' ### alternative checks
#' input_check_response(
#'   check = list(
#'     checkmate::check_character(x),
#'     checkmate::check_character(y)
#'   ),
#'   var_name = "x",
#'   error = TRUE
#' )
#'
#' ### standardized check response
#' \dontrun{
#' input_check_response(
#'   check = checkmate::check_character(y),
#'   var_name = "y",
#'   error = TRUE
#' )
#'
#' input_check_response(
#'   check = list(
#'     checkmate::check_flag(x),
#'     checkmate::check_character(y)
#'   ),
#'   var_name = "y",
#'   error = TRUE
#' )
#' }

input_check_response <- function(
    check,
    var_name = NULL,
    error = TRUE,
    prefix = "Input {.var {var_name}} is bad:"
  ) {
  if (is.list(check)) {
    check_true <- sapply(check, isTRUE)
    check_result <- any(check_true)
    check_msg <- check[!check_true]
  } else {
    check_result <- check_msg <- check
  }
  if (!isTRUE(check_result)) {
    if (isTRUE(error)) {
      checkmate::assert_string(prefix, null.ok = TRUE)
      if (length(check_msg) > 1) {
        cli::cli_abort(
          c(paste(prefix, "Either:"), paste0("- ", check_msg)),
          call = NULL
        )
      } else {
        cli::cli_abort(
          paste(prefix, "{check_msg}"),
          call = NULL
        )
      }
    } else {
      FALSE
    }
  } else {
    TRUE
  }
}

test_that("response to input check works", {
  x <- "1"
  y <- 1
  expect_true(
    input_check_response(
      check = checkmate::check_character(x),
      var_name = "x",
      error = TRUE
    )
  )
  expect_true(
    input_check_response(
      check = checkmate::check_character(x),
      var_name = "x",
      error = FALSE
    )
  )
  expect_false(
    input_check_response(
      check = checkmate::check_character(y),
      var_name = "y",
      error = FALSE
    )
  )
  expect_error(
    input_check_response(
      check = checkmate::check_character(y),
      var_name = "y",
      error = TRUE
    ),
    "Input `y` is bad: Must be of type 'character', not 'double'"
  )
  expect_error(
    input_check_response(
      check = list(
        checkmate::check_flag(x),
        checkmate::check_character(y)
      ),
      var_name = "y",
      error = TRUE
    ),
    "Input `y` is bad: Either:"
  )
})

#' Check numeric vector
#'
#' @description
#' These functions check whether the input is a numeric vector.
#'
#' @inheritParams checkmate::check_numeric
#' @inheritParams checkmate::check_atomic_vector
#'
#' @return
#' Same as documented in \code{\link[checkmate]{check_numeric}}.
#'
#' @keywords validation
#' @family vector helpers
#' @export
#'
#' @examples
#' x <- c(1, 2, "3")
#' check_numeric_vector(x)
#' test_numeric_vector(x)
#' \dontrun{
#' assert_numeric_vector(x)
#' }

check_numeric_vector <- function(
    x, lower = -Inf, upper = Inf, finite = FALSE, any.missing = TRUE,
    all.missing = TRUE, len = NULL, min.len = NULL, max.len = NULL,
    unique = FALSE, sorted = FALSE, names = NULL, typed.missing = FALSE,
    null.ok = FALSE
) {
  if (isTRUE(null.ok) && is.null(x)) {
    return(TRUE)
  }
  res1 <- checkmate::check_atomic_vector(
    x, any.missing = any.missing, all.missing = all.missing, len = len,
    min.len = min.len, max.len = max.len, unique = unique, names = names
  )
  if (!isTRUE(res1)) {
    return(res1)
  }
  res2 <- checkmate::check_numeric(
    x, lower = lower, upper = upper, finite = finite, any.missing = any.missing,
    all.missing = all.missing, len = len, min.len = min.len, max.len = max.len,
    unique = unique, sorted = sorted, names = names,
    typed.missing = typed.missing, null.ok = null.ok
  )
  if (!isTRUE(res2)) {
    return(res2)
  }
  return(TRUE)
}

#' @rdname check_numeric_vector
#' @inheritParams checkmate::assert_numeric
#' @inheritParams checkmate::assert_atomic_vector
#' @export

assert_numeric_vector <- checkmate::makeAssertionFunction(
  check_numeric_vector
)

#' @rdname check_numeric_vector
#' @inheritParams checkmate::assert_numeric
#' @inheritParams checkmate::assert_atomic_vector
#' @export

test_numeric_vector <- checkmate::makeTestFunction(
  check_numeric_vector
)

test_that("check for numeric vector works", {
  # Test a valid numeric vector
  valid_numeric_vector <- 1:9
  expect_true(check_numeric_vector(valid_numeric_vector))

  # Test a non-numeric vector
  non_numeric_vector <- letters[1:9]
  expect_error(
    assert_numeric_vector(non_numeric_vector),
    "Must be of type 'numeric', not 'character'."
  )

  # Test a vector with bad elements
  vector_with_bad_elements <- 0:2
  expect_error(
    assert_numeric_vector(vector_with_bad_elements, lower = 1),
    "Assertion on 'vector_with_bad_elements' failed: Element 1 is not >= 1."
  )

  # Test a vector with bad class
  vector_with_bad_class <- diag(2)
  expect_error(
    assert_numeric_vector(vector_with_bad_class),
    "Assertion on 'vector_with_bad_class' failed: Must be of type 'atomic vector', not 'matrix'."
  )

  # Test NULL
  expect_true(
    test_numeric_vector(NULL, null.ok = TRUE)
  )
  expect_false(
    test_numeric_vector(NULL, null.ok = FALSE)
  )
})

#' Print (abbreviated) \code{data.frame}
#'
#' @description
#' This function prints a (possibly abbreviated) \code{data.frame}.
#'
#' @param x \[`data.frame`\]\cr
#' A \code{data.frame}.
#'
#' @param rows,cols \[`integer(1)` | `NULL` \]\cr
#' The number of rows or columns to be printed, greater or equal `2`.
#'
#' Printing is abbreviated in the middle.
#'
#' Can be `NULL` to print everything.
#'
#' @param row.names,col.names \[`logical(1)`\]\cr
#' Print row names or column names?
#'
#' @inheritParams round_data.frame
#'
#' @return
#' Invisibly returns \code{x}.
#'
#' @export
#' @keywords packaging
#' @family package helpers
#'
#' @examples
#' x <- data.frame(1:10, LETTERS[1:10], stats::rnorm(10))
#' print_data.frame(x, rows = 7)
#' print_data.frame(x, rows = 7, cols = 2)
#' print_data.frame(x, rows = 7, cols = 2, digits = 1)
#' print_data.frame(x, rows = 7, cols = 2, digits = 1, row.names = FALSE)
#' print_data.frame(x, rows = 7, cols = 2, digits = 1, col.names = FALSE)

print_data.frame <- function(
    x, rows = NULL, cols = NULL, digits = NULL,
    row.names = TRUE, col.names = TRUE
  ) {

  ### input checks
  input_check_response(
    check = checkmate::check_data_frame(x),
    var_name = "x"
  )
  input_check_response(
    check = checkmate::check_int(rows, lower = 2, null.ok = TRUE),
    var_name = "rows"
  )
  input_check_response(
    check = checkmate::check_int(cols, lower = 2, null.ok = TRUE),
    var_name = "cols"
  )
  input_check_response(
    check = checkmate::check_int(digits, lower = 0, null.ok = TRUE),
    var_name = "digits"
  )
  input_check_response(
    check = checkmate::check_flag(row.names),
    var_name = "row.names"
  )
  input_check_response(
    check = checkmate::check_flag(col.names),
    var_name = "col.names"
  )

  ### printing
  r <- nrow(x)
  c <- ncol(x)
  if (c == 0L) {
    cli::cli_text("< data.frame with 0 columns and {r} row{?s} >\n")
  } else if (r == 0L) {
    cli::cli_text("< data.frame with {c} column{?s} and 0 rows >\n")
  } else {
    if (!col.names) {
      colnames(x) <- NULL
    }
    if (!is.null(digits)) {
      x <- round_data.frame(x, digits = digits)
    }
    m <- as.matrix(format.data.frame(x, na.encode = FALSE))
    if (!is.null(cols) && cols < c) {
      lc <- ceiling(cols / 2)
      rc <- floor(cols / 2)
      hc <- c - (lc + rc)
      hc_colname <- paste0(
        "<", hc, " col", ifelse(hc > 1, "s", ""), " hidden>"
      )
      hc_value <- strrep(" ", nchar(hc_colname))
      hc_value_bar <- floor(nchar(hc_value) / 2)
      substr(hc_value, hc_value_bar, hc_value_bar) <- "-"
      m <- cbind(
        m[, seq_len(lc), drop = FALSE],
        hc_value,
        m[, utils::tail(seq_len(c), rc), drop = FALSE]
      )
      colnames(m)[lc + 1L] <- hc_colname
    }
    if (isFALSE(row.names)) {
      dimnames(m)[[1L]] <- rep.int("", r)
    }
    if (is.null(rows) || rows >= r) {
      print(m, quote = FALSE)
    } else {
      ur <- ceiling(rows / 2)
      br <- floor(rows / 2)
      hr <- r - (ur + br)
      rowlab <- sprintf(paste0("%-", nchar(r), "s"), seq_len(r))
      prmatrix(
        utils::head(m, ur), quote = FALSE,
        rowlab = if (isFALSE(row.names)) rep("", ur) else rowlab[seq_len(ur)]
      )
      cat("\n"); cli::cli_text("<{hr} row{?s} hidden>\n")
      prmatrix(
        utils::tail(m, br), quote = FALSE,
        collab = sapply(sapply(colnames(m), nchar), function(x) strrep(" ", x))
      )
    }
  }
  invisible(x)
}

test_that("printing abbreviated data.frames works", {
  set.seed(1)
  x <- data.frame(1:10, LETTERS[1:10], stats::rnorm(10))
  expect_snapshot(
    print_data.frame(x, rows = 7)
  )
  expect_snapshot(
    print_data.frame(x, rows = 7, cols = 2)
  )
  expect_snapshot(
    print_data.frame(x, rows = 7, cols = 2, digits = 1)
  )
  expect_snapshot(
    print_data.frame(x, rows = 7, cols = 2, digits = 1, row.names = FALSE)
  )
  expect_snapshot(
    print_data.frame(x, rows = 7, cols = 2, digits = 1, col.names = FALSE)
  )
  expect_snapshot(
    print_data.frame(data.frame())
  )
  expect_snapshot(
    print_data.frame(data.frame(x = numeric()))
  )
})

#' Try an expression silently
#'
#' @description
#' This function tries to execute \code{expr} and returns a string with the
#' error message if the execution failed.
#'
#' @details
#' This function is a wrapper for \code{\link[base]{try}}.
#'
#' @param expr \[`expression`\]\cr
#' An R expression to be evaluated.
#'
#' @return
#' Either the value of \code{expr} or in case of a failure an object of class
#' \code{fail}, which contains the error message.
#'
#' @keywords functional
#' @family function helpers
#' @export
#'
#' @examples
#' \dontrun{
#' try_silent(1 + 1)
#' try_silent(1 + "1")
#' }

try_silent <- function(expr) {
  out <- suppressWarnings(try(expr, silent = TRUE))
  if ("try-error" %in% class(out)) {
    out <- structure(as.character(out[1]), class = "fail")
  }
  return(out)
}

#' Simulate Markov chain
#'
#' @description
#' This function simulates a Markov chain.
#'
#' @param Gamma \[`matrix()`\]\cr
#' A transition probability matrix.
#'
#' @param T \[`integer(1)`\]\cr
#' The length of the Markov chain.
#'
#' @param delta \[`numeric()`\]\cr
#' A probability vector, the initial distribution.
#'
#' The stationary distribution is used by default.
#'
#' @return
#' A \code{numeric} vector of length \code{T} with states.
#'
#' @keywords simulation
#' @family simulation helpers
#' @export
#'
#' @examples
#' Gamma <- matrix(c(0.8, 0.2, 0.3, 0.7), byrow = TRUE, nrow = 2)
#' delta <- c(0.6, 0.4)
#' simulate_markov_chain(Gamma = Gamma, T = 20, delta = delta)

simulate_markov_chain <- function(
    Gamma, T, delta = oeli::stationary_distribution(Gamma)
  ) {
  input_check_response(
    check = check_transition_probability_matrix(Gamma),
    var_name = "Gamma"
  )
  input_check_response(
    check = checkmate::check_int(T, lower = 1),
    var_name = "T"
  )
  input_check_response(
    check = check_probability_vector(delta, len = nrow(Gamma)),
    var_name = "delta"
  )
  N <- length(delta)
  markov_chain <- numeric(T)
  markov_chain[1] <- sample(1:N, 1, prob = delta)
  for (t in 2:T) {
    markov_chain[t] <- sample(1:N, 1, prob = Gamma[markov_chain[t - 1], ])
  }
  markov_chain
}

#' Interrupt long evaluations
#'
#' @description
#' This function interrupts an evaluation after a certain number of seconds.
#' Note the limitations documented in \code{\link[base]{setTimeLimit}}.
#'
#' @param expression \[`expression`\]\cr
#' An R expression to be evaluated.
#'
#' @param seconds \[`numeric(1)`\]\cr
#' The number of seconds.
#'
#' @param on_time_out \[`character(1)`\]\cr
#' Defines what action to take if the evaluation time exceeded, either:
#' \itemize{
#'   \item \code{"error"} to throw an error exception
#'   \item \code{"warning"} to return \code{NULL} along with a warning
#'   \item \code{"silent"} (the default) to just return \code{NULL}
#' }
#'
#' @return
#' The value of \code{expression} or, if the evaluation time exceeded, whatever
#' is specified for \code{on_time_out}.
#'
#' @keywords simulation
#' @family function helpers
#' @export
#'
#' @examples
#' foo <- function(x) {
#'   for (i in 1:10) Sys.sleep(x / 10)
#'   return(x)
#' }
#' timed(foo(0.5), 1)
#' timed(foo(1.5), 1)

timed <- function(
    expression, seconds = Inf, on_time_out = "silent"
  ) {
  checkmate::assert_number(seconds, lower = 0)
  checkmate::assert_choice(on_time_out, c("error", "warning", "silent"))
  setTimeLimit(cpu = seconds, elapsed = seconds, transient = TRUE)
  on.exit({
    setTimeLimit(cpu = Inf, elapsed = Inf, transient = FALSE)
  })
  tryCatch(
    expression,
    error = function(e) {
      msg <- e$message
      time_out <- grepl("reached elapsed time limit|reached CPU time limit", msg)
      if (time_out) {
        if (on_time_out == "error") {
          stop("time limit exceeded", call. = FALSE)
        } else if (on_time_out == "warning") {
          warning("time limit exceeded", call. = FALSE)
        }
        return(NULL)
      } else {
        stop(msg, call. = FALSE)
      }
    }
  )
}

test_that("interruption of long evaluations works", {
  foo <- function(x) {
    for (i in 1:10) Sys.sleep(x / 10)
    return(x)
  }
  expect_error(
    timed(foo(0.5), "1"),
    "Assertion on 'seconds' failed: Must be of type 'number', not 'character'."
  )
  expect_equal(
    timed(foo(0.5), 5),
    0.5
  )
  expect_null(
    timed(foo(5), 1, "silent")
  )
  expect_warning(
    timed(foo(5), 1, "warning"),
    "time limit exceeded"
  )
  expect_error(
    timed(foo(5), 1, "error"),
    "time limit exceeded"
  )
})

#' Simulator R6 Object
#'
#' @description
#' Creates a simulation setup, where a function `f` is evaluated `runs` times,
#' optionally at each combination of input values.
#'
#' Provides some convenience (see below for more details):
#'
#' - Simulation results can be restored from a backup if the R session crashes.
#' - More simulation runs can be conducted after the initial simulation, failed
#'   simulation cases can be re-run.
#' - Parallel computation and progress updates are supported.
#'
#' @details
#' ## Backup
#' Simulation results can be saved to disk, allowing you to restore the results
#' if the R session is interrupted or crashes before the simulation completes.
#' To enable backup, set `backup = TRUE` in the `$go()` method, which will
#' create a backup directory at the location specified by `path`.
#' To restore, use `Simulator$initialize(use_backup = path)`.
#'
#' ## More runs and re-run
#' If additional simulation runs are needed, simply call the `$go()` method
#' again. Any cases that were not successfully completed in previous runs will
#' be attempted again.
#'
#' ## Parallel computation
#' By default, simulations run sequentially. But since they are independent,
#' they can be parallelized to decrease computation time. To enable parallel
#' computation, use the [`{future}` framework](https://future.futureverse.org/).
#' For example, run
#' \preformatted{future::plan(future::multisession, workers = 4)}
#' in advance for computation in 4 parallel R sessions.
#'
#' ## Progress updates
#' Use the [`{progressr}` framework](https://progressr.futureverse.org/) to
#' get progress updates. For example, run the following in advance:
#' \preformatted{progressr::handlers(global = TRUE)
#' progressr::handlers(
#'   progressr::handler_progress(format = ">> :percent, :eta to go :message")
#' )
#' }
#'
#' @keywords simulation
#' @family simulation helpers
#' @export
#'
#' @examples
#' # 1. Initialize a new simulation setup:
#' object <- Simulator$new(verbose = TRUE)
#'
#' # 2. Define function `f` and arguments (if any):
#' f <- function(x, y = 1) {
#'   Sys.sleep(runif(1)) # to see progress updates
#'   x + y
#' }
#' x_args <- list(1, 2)
#' object$define(f = f, x = x_args)
#' print(object)
#'
#' # 3. Define 'future' and 'progress' (optional):
#' \dontrun{
#' future::plan(future::sequential)
#' progressr::handlers(global = TRUE)}
#'
#' # 4. Evaluate `f` `runs` times at each parameter combination (backup is optional):
#' path <- file.path(tempdir(), paste0("backup_", format(Sys.time(), "%Y-%m-%d-%H-%M-%S")))
#' object$go(runs = 2, backup = TRUE, path = path)
#'
#' # 5. Access the results:
#' object$results
#'
#' # 6. Check if cases are pending or if an error occurred:
#' object$cases
#'
#' # 7. Restore simulation results from backup:
#' object_restored <- Simulator$new(use_backup = path)
#' print(object_restored)
#' \dontrun{all.equal(object, object_restored)}
#'
#' # 8. Run more simulations and pending simulations (if any):
#' object_restored$go(runs = 2)

Simulator <- R6::R6Class(

  classname = "Simulator",
  cloneable = FALSE,

  public = list(

    #' @description
    #' Initialize a `Simulator` object, either a new one or from backup.
    #'
    #' @param use_backup \[`NULL` | `character(1)`\]\cr
    #' Optionally a path to a backup folder previously used in `$go()`.
    #'
    #' @param verbose \[`logical(1)`\]\cr
    #' Provide info? Does not include progress updates. For that, see details.

    initialize = function(
      use_backup = NULL, verbose = getOption("verbose", default = FALSE)
    ) {

      private$.verbose <- isTRUE(verbose)

      if (is.null(use_backup)) {
        private$.status("Created {.cls Simulator}, call {.fun $define} next.")
      } else {
        input_check_response(
          check = checkmate::check_directory_exists(use_backup, access = "rw"),
          var_name = "use_backup"
        )
        input_check_response(
          check = checkmate::check_file_exists(
            file.path(use_backup, "Simulator_object.rds"), extension = ".rds",
            access = "r"
          )
        )
        self_old <- readRDS(file.path(use_backup, "Simulator_object.rds"))
        list2env(
          mget(
            c(".verbose", ".f", ".args", ".results", ".cases"),
            envir = self_old$.__enclos_env__$private
          ),
          envir = self$.__enclos_env__$private
        )
        case_files <- list.files(
          use_backup, pattern = "^case_.*\\.rds$", full.names = TRUE
        )
        cases <- lapply(case_files, readRDS)
        private$.results <- c(private$.results, cases)
        lapply(cases, function(case) {
          private$.update_case_pending(case = case$.case, error = case$.error)
        })
        private$.status(
          "Loaded {.cls Simulator} and {length(cases)} case{?s} from backup."
        )
      }
    },

    #' @description
    #' Print method.

    print = function() {
      cli::cli_bullets(c(
        "*"  = "Total cases: {nrow(private$.cases)}",
        ">"  = "Pending cases: {sum(private$.cases$.pending)}",
        "v"  = "Successful cases: {length(private$.results)}",
        "x"  = "Failed cases: {sum(private$.cases$.error)}"
      ))
      cli::cli_end()
      invisible(self)
    },

    #' @description
    #' Define function and arguments for a new `Simulator` object.
    #'
    #' @param f \[`function`\]\cr
    #' A `function` to evaluate.
    #'
    #' @param ...
    #' Arguments for `f`. Each value must be
    #'
    #' 1. named after an argument of `f`, and
    #' 2. a `list`, where each element is a variant of that argument for `f`.

    define = function(f, ...) {

      ### already defined?
      if (!is.null(private$.f)) {
        cli::cli_abort(
          "Definition already provided.",
          call = NULL
        )
      }

      ### input checks
      args <- list(...)
      arg_names <- names(args)
      input_check_response(
        check = list(
          checkmate::check_list(args, len = 0),
          checkmate::check_names(
            arg_names, type = "strict", disjunct.from = private$.reserved_vars
          )
        ),
        var_name = "..."
      )
      Map(function(arg, name) {
        input_check_response(
          check = checkmate::check_list(arg),
          var_name = name
        )
      }, args, arg_names)
      input_check_response(
        check = checkmate::check_function(f, args = arg_names),
        var_name = "f"
      )

      ### save simulation setting
      private$.f <- f
      private$.args <- args
      private$.status(
        "Simulation details defined, call {.fun $go} next to run simulations."
      )
      invisible(self)

    },

    #' @description
    #' Run simulations.
    #'
    #' @param runs \[`integer(1)`\]\cr
    #' The number of (additional) simulation runs.
    #'
    #' If `runs = 0`, only pending cases (if any) are solved.
    #'
    #' @param backup \[`logical(1)`\]\cr
    #' Create a backup under `path`?
    #'
    #' @param path \[`character(1)`\]\cr
    #' Only relevant, if `backup = TRUE`.
    #'
    #' In this case, a path for a new folder, which does not yet exist and
    #' allows reading and writing.

    go = function(
      runs = 0, backup = FALSE,
      path = paste0("backup_", format(Sys.time(), "%Y-%m-%d-%H-%M-%S"))
    ) {

      ### input checks
      input_check_response(
        check = check_missing(runs),
        var_name = "runs"
      )
      input_check_response(
        check = checkmate::check_count(runs, positive = FALSE),
        var_name = "runs"
      )
      input_check_response(
        check = checkmate::check_flag(backup),
        var_name = "backup"
      )

      ### prepare simulation
      private$.new_cases(runs = runs)
      cases <- self$cases |> dplyr::filter(.pending)
      if (backup) {
        input_check_response(
          check = checkmate::check_path_for_output(path, overwrite = FALSE),
          var_name = "path"
        )
        dir.create(path)
        saveRDS(self, file = sprintf("%s/Simulator_object.rds", path))
        private$.status("Saving backup to path {.path {normalizePath(path)}}.")
      }
      private$.status("Started simulation with {nrow(cases)} cases...")
      progress_step <- progressr::progressor(steps = nrow(cases))
      f <- private$.f # need to put f in environment

      ### run simulation
      results <- future.apply::future_lapply(cases$.case, function(case) {

        progress_step(glue::glue("[case {case}] started"), amount = 0)

        args <- lapply(
          cases[cases$.case == case, -(1:4), drop = FALSE],
          function(x) x[[1]]
        )

        start <- Sys.time()
        f_out <- try_silent(do.call(what = f, args = args))
        end <- Sys.time()
        error <- inherits(f_out, "fail")

        progress_step(glue::glue("[case {case}] finished"), class = "sticky")

        case_out <- list(
          ".case" = case,
          ".error" = error,
          ".seconds" = difftime(end, start, units = "secs"),
          ".input" = args,
          ".output" = f_out
        )

        if (backup && !error) {
          saveRDS(case_out, file = sprintf("%s/case_%03d.rds", path, case))
        }

        return(case_out)

      }, future.seed = TRUE)

      ### save results
      valid_cases <- Filter(function(case) !case$.error, results)
      ncases_failed <- nrow(cases) - length(valid_cases)
      private$.results <- c(private$.results, valid_cases)
      lapply(results, function(case) {
        private$.update_case_pending(case = case$.case, error = case$.error)
      })
      private$.status("Simulation complete, {ncases_failed} case{?s} failed.")
      invisible(self)

    }

  ),

  active = list(

    #' @field results \[`tibble`, read-only\]\cr
    #' The simulation results.

    results = function(value) {
      if (missing(value)) {
        if (length(private$.results) == 0) {
          tibble::tibble(
            .case = integer(),
            .seconds = difftime(numeric(), numeric(), units = "secs"),
            .input = list(),
            .output = list()
          )
        } else {
          private$.results |>
            lapply(function(x) tibble::tibble(
              .case = x$.case,
              .seconds = x$.seconds,
              .input = list(x$.input),
              .output = list(x$.output)
            )) |>
            dplyr::bind_rows() |>
            dplyr::arrange(.case)
        }
      } else {
        cli::cli_abort(
          "Field {.var $results} is read-only.",
          call = NULL
        )
      }
    },

    #' @field cases \[`tibble`, read-only\]\cr
    #' The simulation cases.

    cases = function(value) {
      if (missing(value)) {
        private$.cases |>
          dplyr::arrange(.case)
      } else {
        cli::cli_abort(
          "Field {.var $cases} is read-only.",
          call = NULL
        )
      }
    }

  ),

  private = list(

    .verbose = FALSE,
    .f = NULL,
    .args = list(),
    .results = list(),
    .cases = tibble::tibble(
      .case = integer(),
      .pending = logical(),
      .error = logical(),
      .run = integer()
    ),
    .reserved_vars = c(".case", ".pending", ".error", ".run"),

    .status = function(msg, verbose = private$.verbose) {
      checkmate::assert_string(msg)
      checkmate::assert_flag(verbose)
      if (verbose) cli::cli_inform(msg, .envir = parent.frame())
    },

    .new_cases = function(runs) {
      run_seq <- length(unique(private$.cases$.run)) + seq_len(runs)
      grid <- c(list(.run = run_seq), private$.args) |>
        expand.grid(stringsAsFactors = FALSE) |>
        tibble::as_tibble() |>
        dplyr::arrange(.run)
      new_cases <- cbind(
        tibble::tibble(
          .case = nrow(self$cases) + seq_len(nrow(grid)),
          .pending = TRUE,
          .error = FALSE
        ),
        grid
      )
      private$.cases <- dplyr::bind_rows(private$.cases, new_cases)
    },

    .update_case_pending = function(case, error) {
      private$.cases[private$.cases$.case == case, ".pending"] <- error
      private$.cases[private$.cases$.case == case, ".error"] <- error
    }
  )

)

test_that("simulation works", {
  f <- function() rnorm(1, sd = 0.1)
  sim <- Simulator$new(verbose = FALSE)
  checkmate::expect_r6(sim, "Simulator")
  expect_identical(
    sim$results,
    structure(
      list(
        .case = integer(0),
        .seconds = structure(numeric(0), class = "difftime", units = "secs"),
        .input = list(), .output = list()
      ),
      class = c("tbl_df", "tbl", "data.frame"),
      row.names = integer(0)
    )
  )
  sim$define(f = f)
  sim$go(runs = 2, backup = FALSE)
  expect_snapshot(print(sim))
  checkmate::expect_tibble(sim$cases, nrows = 2, ncols = 4)
  expect_error(sim$cases <- "bad", "read-only")
  checkmate::expect_tibble(sim$results, nrows = 2, ncols = 4)
  expect_error(sim$results <- "bad", "read-only")
})

test_that("re-run works", {
  throw_error <- TRUE
  g <- function(x, y) {
    if (x + y == 5 && throw_error) stop()
    x + y
  }
  sim <- Simulator$new(verbose = FALSE)
  sim$define(f = g, x = as.list(1:2), y = as.list(2:3))
  sim$go(runs = 1, backup = FALSE)
  expect_identical(
    sim$cases,
    structure(
      list(
        .case = 1:4, .pending = c(FALSE, FALSE, FALSE, TRUE),
        .error = c(FALSE, FALSE, FALSE, TRUE), .run = c(1L, 1L, 1L, 1L),
        x = list(1L, 2L, 1L, 2L), y = list(2L, 2L, 3L, 3L)
      ),
      row.names = c(NA, -4L), class = c("tbl_df", "tbl", "data.frame")
    )
  )
  expect_identical(
    sim$results |> dplyr::select(-.seconds),
    structure(
      list(
        .case = 1:3,
        .input = list(
          list(x = 1L, y = 2L), list(x = 2L, y = 2L), list(x = 1L, y = 3L)
        ),
        .output = list(3L, 4L, 4L)
      ),
      row.names = c(NA, -3L), class = c("tbl_df", "tbl", "data.frame")
    )
  )
  throw_error <- FALSE
  sim$go(backup = FALSE)
  expect_identical(
    sim$cases,
    structure(
      list(
        .case = 1:4, .pending = c(FALSE, FALSE, FALSE, FALSE),
        .error = c(FALSE, FALSE, FALSE, FALSE), .run = c(1L, 1L, 1L, 1L),
        x = list(1L, 2L, 1L, 2L), y = list(2L, 2L, 3L, 3L)
      ),
      row.names = c(NA, -4L), class = c("tbl_df", "tbl", "data.frame")
    )
  )
  expect_identical(
    sim$results |> dplyr::select(-.seconds),
    structure(
      list(
        .case = 1:4,
        .input = list(
          list(x = 1L, y = 2L), list(x = 2L, y = 2L), list(x = 1L, y = 3L), list(x = 2L, y = 3L)
        ),
        .output = list(3L, 4L, 4L, 5L)
      ),
      row.names = c(NA, -4L), class = c("tbl_df", "tbl", "data.frame")
    )
  )
})

test_that("backup works", {
  h <- function(x, y = 1) x + y
  sim <- Simulator$new(verbose = FALSE)
  sim$define(f = h, x = as.list(1:2))
  path <- file.path(tempdir(), paste0(sample(letters, size = 26), collapse = ""))
  sim$go(runs = 2, backup = TRUE, path = path)
  sim_restored <- Simulator$new(use_backup = path)
  expect_error(sim_restored$define(), "Definition already provided.")
  expect_identical(sim, sim_restored)
})

#' Check if two objects have identical structure
#'
#' @description
#' This function determines whether two objects have the same structure,
#' - which includes the \code{\link[base]{mode}}, \code{\link[base]{class}} and
#'   dimension
#' - but does *not* include concrete values or attributes.
#'
#' @param x,y \[`any`\]\cr
#' Two objects.
#'
#' @return
#' Either \code{TRUE} if \code{x} and \code{y} have the same structure, and
#' \code{FALSE}, else.
#'
#' @keywords validation
#' @family package helpers
#' @export
#'
#' @references
#' Inspired by \url{https://stackoverflow.com/a/45548885/15157768}.
#
#' @examples
#' identical_structure(integer(1), 1L)
#' identical_structure(diag(2), matrix(rnorm(4), 2, 2))
#' identical_structure(diag(2), data.frame(diag(2)))

identical_structure <- function(x, y) {
  compare_type <- function(x, y) {
    if (length(x) == length(y)) {
      all(mapply(x, y,
        FUN = function(x, y) {
          if (is.list(x) && is.list(y)) {
            all(compare_type(x, y))
          } else if (is.list(x) == is.list(y)) {
            identical(x, y)
          } else {
            FALSE
          }
        }
      ))
    } else {
      FALSE
    }
  }
  if (is.list(x) && is.list(y)) {
    compare_type(
      rapply(x, function(values) c(mode(values), length(values)), how = "list"),
      rapply(y, function(values) c(mode(values), length(values)), how = "list")
    )
  } else {
    if (!identical(class(x), class(y))) {
      FALSE
    } else {
      compare_type(
        if (is.null(dim(x))) length(x) else dim(x),
        if (is.null(dim(y))) length(y) else dim(y)
      )
    }
  }
}

test_that("check for identical structures works", {
  expect_true(identical_structure(1, 2))
  expect_true(identical_structure(list(1), list(2)))
  expect_false(identical_structure(list(1), list()))
  x <- structure(1, class = c("a", "b"))
  y <- structure(1, class = c("b", "a"))
  expect_false(identical_structure(x, y))
  x <- list(list(1, 2, "3"))
  y <- list(list(1, 2, 3))
  expect_false(identical_structure(x, y))
  x <- list(1, 2, 3)
  expect_false(identical_structure(x, y))
})

#' Dictionary R6 Object
#'
#' @description
#' Provides a simple key-value interface based on R6.
#'
#' @param key_name \[`character(1)`\]\cr
#' The name for the key variable.
#'
#' @param alias_name \[`NULL` | `character(1)`\]\cr
#' Optionally the name for the alias variable.
#'
#' @param value_names \[`character(0)`\]\cr
#' The names of the values connected to a key.
#'
#' @param value_assert \[`alist(1)`\]\cr
#' For each element in \code{value_names}, \code{values_assert} *can* have an
#' identically named element of the form \code{checkmate::assert*(...)}, where
#' \code{...} can be any argument for the assertion function except for the
#' \code{x} argument.
#'
#' @param allow_overwrite \[`logical(1)`\]\cr
#' Allow overwriting existing keys with new values?
#' Duplicate keys are never allowed.
#'
#' @param keys_reserved \[`character()`\]\cr
#' Names that must not be used as keys.
#'
#' @param alias_choices \[`NULL` or `character()`\]\cr
#' Optionally possible values for the alias. Can also be \code{NULL}, then all
#' alias values are allowed.
#'
#' @param dictionary_name \[`NULL` or `character()`\]\cr
#' Optionally the name for the dictionary.
#'
#' @param key \[`character(1)`\]\cr
#' A value for the key variable \code{key_name}. Use the \code{$keys} method for
#' available keys.
#'
#' @keywords indexing
#' @family package helpers
#' @export
#'
#' @examples
#' # Managing variable metadata for a dataset
#'
#' meta_dict <- Dictionary$new(
#'   key_name = "var_name",
#'   alias_name = "category",
#'   value_names = c("label", "type"),
#'   value_assert = alist(
#'     label = checkmate::assert_string(),
#'     type = checkmate::assert_choice(choices = c("numeric", "factor", "character"))
#'   ),
#'   allow_overwrite = FALSE,
#'   keys_reserved = c("id"),
#'   alias_choices = c("demographics", "outcome", "other"),
#'   dictionary_name = "Variable Metadata"
#' )
#'
#' # Add entries to the dictionary
#' meta_dict$add(
#'   var_name = "age",
#'   label = "Age of respondent",
#'   type = "numeric",
#'   category = "demographics"
#' )
#'
#' meta_dict$add(
#'   var_name = "gender",
#'   label = "Gender identity",
#'   type = "factor",
#'   category = "demographics"
#' )
#'
#' meta_dict$add(
#'   var_name = "income",
#'   label = "Annual income in USD",
#'   type = "numeric",
#'   category = c("demographics", "outcome")
#' )
#'
#' # Print dictionary
#' meta_dict$print()
#'
#' # Retrieve full metadata for a variable
#' meta_dict$get("income")
#'
#' # Retrieve a specific piece of metadata
#' meta_dict$get("income", value = "label")
#'
#' # Show variables by category
#' meta_dict$alias

Dictionary <- R6::R6Class(
  classname = "Dictionary",
  lock_class = TRUE,
  cloneable = FALSE,
  public = list(

    #' @description
    #' Initializing a new \code{Dictionary} object.
    initialize = function(
      key_name,
      alias_name = NULL,
      value_names = character(),
      value_assert = alist(),
      allow_overwrite = TRUE,
      keys_reserved = character(),
      alias_choices = NULL,
      dictionary_name = NULL
    ) {
      checkmate::assert_string(key_name)
      checkmate::assert_string(alias_name, null.ok = TRUE)
      checkmate::assert_character(
        value_names,
        any.missing = FALSE, unique = TRUE
      )
      checkmate::assert_list(
        value_assert,
        types = "call", any.missing = FALSE
      )
      checkmate::assert_names(
        names(value_assert),
        type = "unique", subset.of = value_names
      )
      checkmate::assert_flag(allow_overwrite)
      checkmate::assert_character(
        keys_reserved,
        any.missing = FALSE, unique = TRUE
      )
      checkmate::assert_character(
        alias_choices,
        any.missing = FALSE, unique = TRUE, null.ok = TRUE
      )
      checkmate::assert_string(dictionary_name, null.ok = TRUE)
      private$.key_name <- key_name
      if (is.null(alias_name)) {
        private$.alias_activated <- FALSE
      } else {
        private$.alias_activated <- TRUE
        private$.alias_name <- alias_name
      }
      private$.value_names <- value_names
      private$.value_assert <- value_assert
      private$.allow_overwrite <- allow_overwrite
      private$.keys_reserved <- keys_reserved
      private$.alias_choices <- alias_choices
      private$.dictionary_name <- dictionary_name
    },

    #' @description
    #' Adding an element to the dictionary.
    #' @param ...
    #' Values for
    #' - the key variable \code{key_name} (must be a single \code{character}),
    #' - the alias variable \code{alias_name} (optionally, must then be a
    #'   \code{character} \code{vector}),
    #' - all the variables specified for \code{value_names} (if any, they must
    #'   comply to the \code{value_assert} checks).

    add = function(...) {
      inputs <- list(...)
      private$.check_inputs(inputs)
      private$.add_to_storage(inputs)
      invisible(self)
    },

    #' @description
    #' Getting elements from the dictionary.
    #' @param value \[`NULL` | `character(1)`\]\cr
    #' One of the elements in \code{value_names}, selecting the required value.
    #' Can also be \code{NULL} (default) for all values connected to the
    #' \code{key}, returned as a \code{list}.

    get = function(key, value = NULL) {
      if (!private$.key_exists(key)) {
        stop("The key '", key, "' does not exist.", call. = FALSE)
      }
      if (is.null(value)) {
        private$.storage[[key]]
      } else {
        if (value %in% private$.value_names) {
          private$.storage[[key]][[value]]
        } else {
          stop("The value name '", value, "' does not exist.", call. = FALSE)
        }
      }
    },

    #' @description
    #' Removing elements from the dictionary (and associated alias, if any).

    remove = function(key) {
      if (!private$.key_exists(key)) {
        warning("The key '", key, "' does not exist.", call. = FALSE)
      } else {
        private$.storage <- within(private$.storage, rm(list = key))
        if (private$.alias_activated) {
          private$.alias_list <- lapply(
            private$.alias_list, function(alias) alias[!key %in% alias]
          )
        }
      }
      invisible(self)
    },

    #' @description
    #' Printing details of the dictionary.

    print = function() {
      cat("<Dictionary>", private$.dictionary_name, "\n")
      if (length(self$keys) == 0) {
        cat("No elements contained.")
      } else {
        cat("Keys: \n")
        cat(paste("-", self$keys), sep = "\n")
      }
      invisible(self)
    }
  ),
  active = list(

    #' @field keys \[`character()`\]\cr
    #' Available keys.
    keys = function(value) {
      if (missing(value)) {
        names(private$.storage)
      } else {
        stop("read only")
      }
    },

    #' @field alias \[`list()`\]\cr
    #' Available keys per alias value.

    alias = function(value) {
      if (missing(value)) {
        private$.alias_list
      } else {
        stop("read only")
      }
    }
  ),

  private = list(
    .storage = list(),
    .alias_list = list(),
    .key_name = NA_character_,
    .alias_name = NULL,
    .alias_activated = NA,
    .value_names = character(),
    .value_assert = alist(),
    .allow_overwrite = NA,
    .keys_reserved = character(),
    .alias_choices = NULL,
    .dictionary_name = NULL,

    .check_inputs = function(inputs) {
      input_names_required <- c(private$.key_name, private$.value_names)
      checkmate::assert_names(
        names(inputs),
        subset.of = c(input_names_required, private$.alias_name)
      )
      checkmate::assert_string(inputs[[private$.key_name]])
      if (private$.alias_activated && !is.null(private$.alias_choices)) {
        checkmate::assert_subset(
          inputs[[private$.alias_name]],
          choices = private$.alias_choices
        )
      }
      if (inputs[[private$.key_name]] %in% private$.keys_reserved) {
        stop(
          "The key '", inputs[[private$.key_name]], "' must not be used.",
          call. = FALSE
        )
      }
      if (private$.key_exists(inputs[[private$.key_name]]) &&
        !private$.allow_overwrite
      ) {
        stop(
          "The key '", inputs[[private$.key_name]], "' already exists.",
          call. = FALSE
        )
      }
      for (value_name in private$.value_names) {
        if (value_name %in% names(private$.value_assert)) {
          assert_function <- private$.value_assert[[value_name]]
          assert_function[["x"]] <- inputs[[value_name]]
          eval(assert_function)
        }
      }
    },

    .add_to_storage = function(inputs) {
      key <- inputs[[private$.key_name]]
      private$.storage[[key]] <- inputs[private$.value_names]
      if (private$.alias_activated && !is.null(inputs[[private$.alias_name]])) {
        for (alias in inputs[[private$.alias_name]]) {
          private$.alias_list[[alias]] <- c(private$.alias_list[[alias]], key)
        }
      }
    },

    .key_exists = function(key) {
      key %in% self$keys
    }
  )
)

#' Split a vector into chunks
#'
#' @description
#' This function either
#' - splits a vector into \code{n} chunks of equal size (\code{type = 1}),
#' - splits a vector into chunks of size \code{n} (\code{type = 2}).
#'
#' @param x \[atomic()`\]\cr
#' A vector of elements.
#'
#' @param n \[`integer(1)`\]\cr
#' A number smaller or equal \code{length(x)}.
#'
#' @param type \[`1` | `2`\]\cr
#' Either
#'
#' - \code{1} (default) to split \code{x} into \code{n} chunks of equal size,
#' - or \code{2} to split \code{x} into chunks of size \code{n}.
#'
#' @param strict \[`logical(1)`\]\cr
#' Set to \code{TRUE} to fail if \code{length(x)} is not a multiple of \code{n},
#' or \code{FALSE} (default), else.
#'
#' @return
#' A \code{list}.
#'
#' @keywords transformation
#' @family vector helpers
#' @export
#'
#' @examples
#' x <- 1:12
#' chunk_vector(x, n = 3, type = 1)
#' chunk_vector(x, n = 3, type = 2)
#' try(chunk_vector(x, n = 5, strict = TRUE))

chunk_vector <- function(x, n, type = 1, strict = FALSE) {
  input_check_response(
    checkmate::check_atomic_vector(x),
    "x"
  )
  input_check_response(
    checkmate::check_int(n, lower = 1),
    "n"
  )
  input_check_response(
    checkmate::check_choice(type, c(1, 2)),
    "type"
  )
  input_check_response(
    checkmate::check_flag(strict),
    "strict"
  )
  if (strict) {
    input_check_response(
      if (length(x) %% n == 0) TRUE else "Not a multiple of 'length(x)'",
      "n"
    )
  }
  if (type == 1) {
    split(x, cut(seq_along(x), n, labels = FALSE))
  } else {
    split(x, ceiling(seq_along(x) / n))
  }
}

test_that("chunking a vector works", {
  checkmate::expect_list(
    chunk_vector(LETTERS, n = 4), len = 4
  )
  expect_equal(
    chunk_vector(1:12, n = 3, type = 1),
    list("1" = 1:4, "2" = 5:8, "3" = 9:12)
  )
  expect_equal(
    chunk_vector(1:12, n = 3, type = 2),
    list("1" = 1:3, "2" = 4:6, "3" = 7:9, "4" = 10:12)
  )
  expect_error(
    chunk_vector(1:10, n = 4, strict = TRUE),
    "Not a multiple"
  )
})

#' Split a vector at positions
#'
#' @description
#' This function splits a vector at specific positions.
#'
#' @param x \[atomic()`\]\cr
#' A vector of elements.
#'
#' @param at \[`integer()`\]\cr
#' Index position(s) just before to split.
#'
#' For example, \code{at = n} splits before the \code{n}th element of \code{x}.
#'
#' @return
#' A \code{list}.
#'
#' @keywords transformation
#' @family vector helpers
#' @export
#'
#' @references
#' Based on https://stackoverflow.com/a/19274414.
#'
#' @examples
#' x <- 1:10
#' split_vector_at(x, c(2, 3, 5, 7))

split_vector_at <- function(x, at) {
  input_check_response(
    check = checkmate::check_atomic_vector(x),
    var_name = "x"
  )
  input_check_response(
    check = checkmate::check_integerish(
      at, lower = 1, upper = length(x), unique = TRUE, any.missing = FALSE
    ),
    var_name = "at"
  )
  if (length(x) %in% c(0, 1)) {
    return(list(x))
  }
  at <- unique(c(1L, at, length(x) + 1L))
  Map(
    function(x, i, j) x[i:j], list(x),
    utils::head(at, -1L), utils::tail(at, -1L) - 1L
  )
}

test_that("splitting a vector works", {
  expect_equal(
    split_vector_at(1:10, 1:10),
    as.list(1:10)
  )
  expect_equal(
    split_vector_at(1:10, c(2, 3, 5, 7)),
    list(1L, 2L, 3:4, 5:6, 7:10)
  )
  expect_equal(
    split_vector_at(1:10, numeric()),
    list(1:10)
  )
  expect_error(
    split_vector_at(1:10, 0),
    "Input `at` is bad: Element 1 is not >= 1"
  )
  expect_error(
    split_vector_at(1:10, 11),
    "Input `at` is bad: Element 1 is not <= 10"
  )
  expect_equal(
    split_vector_at(numeric(), numeric()),
    list(numeric())
  )
  expect_equal(
    split_vector_at(1, 1),
    list(1)
  )
})

# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

#' @rdname ddirichlet
#' @export
ddirichlet_cpp <- function(x, concentration, log = FALSE) {
    .Call(`_oeli_ddirichlet_cpp`, x, concentration, log)
}

#' @rdname ddirichlet
#' @export
rdirichlet_cpp <- function(concentration) {
    .Call(`_oeli_rdirichlet_cpp`, concentration)
}

#' @rdname dmixnorm
#' @export
dmixnorm_cpp <- function(x, mean, Sigma, proportions) {
    .Call(`_oeli_dmixnorm_cpp`, x, mean, Sigma, proportions)
}

#' @rdname dmixnorm
#' @export
pmixnorm_cpp <- function(x, mean, Sigma, proportions, abseps = 1e-3) {
    .Call(`_oeli_pmixnorm_cpp`, x, mean, Sigma, proportions, abseps)
}

#' @rdname dmixnorm
#' @export
rmixnorm_cpp <- function(mean, Sigma, proportions) {
    .Call(`_oeli_rmixnorm_cpp`, mean, Sigma, proportions)
}

#' @rdname dmvnorm
#' @export
dmvnorm_cpp <- function(x, mean, Sigma, log = FALSE) {
    .Call(`_oeli_dmvnorm_cpp`, x, mean, Sigma, log)
}

#' @rdname dmvnorm
#' @export
pmvnorm_cpp <- function(x, mean, Sigma, abseps = 1e-3) {
    .Call(`_oeli_pmvnorm_cpp`, x, mean, Sigma, abseps)
}

#' @rdname dmvnorm
#' @export
rmvnorm_cpp <- function(mean, Sigma, log = FALSE) {
    .Call(`_oeli_rmvnorm_cpp`, mean, Sigma, log)
}

#' @rdname dtnorm
#' @export
dtnorm_cpp <- function(x, mean, sd, point, above, log = FALSE) {
    .Call(`_oeli_dtnorm_cpp`, x, mean, sd, point, above, log)
}

#' @rdname dtnorm
#' @export
dttnorm_cpp <- function(x, mean, sd, lower, upper, log = FALSE) {
    .Call(`_oeli_dttnorm_cpp`, x, mean, sd, lower, upper, log)
}

#' @rdname dtnorm
#' @export
rtnorm_cpp <- function(mean, sd, point, above, log = FALSE) {
    .Call(`_oeli_rtnorm_cpp`, mean, sd, point, above, log)
}

#' @rdname dtnorm
#' @export
rttnorm_cpp <- function(mean, sd, lower, upper, log = FALSE) {
    .Call(`_oeli_rttnorm_cpp`, mean, sd, lower, upper, log)
}

#' @rdname dwishart
#' @export
dwishart_cpp <- function(x, df, scale, log = FALSE, inv = FALSE) {
    .Call(`_oeli_dwishart_cpp`, x, df, scale, log, inv)
}

#' @rdname dwishart
#' @export
rwishart_cpp <- function(df, scale, inv = FALSE) {
    .Call(`_oeli_rwishart_cpp`, df, scale, inv)
}

# Register entry points for exported C++ functions
methods::setLoadAction(function(ns) {
    .Call(`_oeli_RcppExport_registerCCallable`)
})

#' Wishart distribution
#'
#' @description
#' The function `dwishart()` computes the density of a Wishart distribution.
#'
#' The function `rwishart()` samples from a Wishart distribution.
#'
#' The functions with suffix `_cpp` perform no input checks, hence are faster.
#'
#' @param x \[`matrix()`\]\cr
#' A covariance matrix of dimension `p`.
#'
#' @param df \[`integer()`\]\cr
#' The degrees of freedom greater of equal `p`.
#'
#' @param scale \[`matrix()`\]\cr
#' The scale covariance matrix of dimension `p`.
#'
#' @param log \[`logical(1)`\]\cr
#' Return the logarithm of the density value?
#'
#' @param inv \[`logical(1)`\]\cr
#' Use this inverse Wishart distribution?
#'
#' @return
#' For `dwishart()`: The density value.
#'
#' For `rwishart()`: A `matrix`, the random draw.
#'
#' @keywords distribution
#' @family simulation helpers
#' @export
#'
#' @examples
#' x <- diag(2)
#' df <- 6
#' scale <- matrix(c(1, -0.3, -0.3, 0.8), ncol = 2)
#'
#' # compute density
#' dwishart(x = x, df = df, scale = scale)
#' dwishart(x = x, df = df, scale = scale, log = TRUE)
#' dwishart(x = x, df = df, scale = scale, inv = TRUE)
#'
#' # sample
#' rwishart(df = df, scale = scale)
#' rwishart(df = df, scale = scale, inv = TRUE)
#'
#' # expectation of Wishart is df * scale
#' n <- 100
#' replicate(n, rwishart(df = df, scale = scale), simplify = FALSE) |>
#'   Reduce(f = "+") / n
#' df * scale
#'
#' # expectation of inverse Wishart is scale / (df - p - 1)
#' n <- 100
#' replicate(n, rwishart(df = df, scale = scale, TRUE), simplify = FALSE) |>
#'   Reduce(f = "+") / n
#' scale / (df - 2 - 1)

dwishart <- function(x, df, scale, log = FALSE, inv = FALSE) {
  assert_covariance_matrix(x)
  checkmate::assert_int(df, lower = nrow(x))
  assert_covariance_matrix(scale, dim = nrow(x))
  checkmate::assert_flag(log)
  checkmate::assert_flag(inv)
  dwishart_cpp(x, df, scale, log, inv)
}

#' @rdname dwishart
#' @export

rwishart <- function(df, scale, inv = FALSE) {
  assert_covariance_matrix(scale)
  checkmate::assert_int(df, lower = nrow(scale))
  checkmate::assert_flag(inv)
  rwishart_cpp(df, scale, inv)
}

test_that("(Inverse-) Wishart density can be computed", {
  x <- diag(2)
  df <- 4
  scale <- matrix(c(1, 0.2, 0.2, 2), 2, 2)
  factor <- 1000
  expect_equal(
    round(dwishart(x = x, df = df, scale = scale, log = FALSE, inv = FALSE) * factor) / factor,
    0.005
  )
  expect_equal(
    round(dwishart(x = x, df = df, scale = scale, log = TRUE, inv = FALSE) * factor) / factor,
    -5.335
  )
  expect_equal(
    round(dwishart(x = x, df = df, scale = scale, log = FALSE, inv = TRUE) * factor) / factor,
    0.034
  )
  expect_equal(
    round(dwishart(x = x, df = df, scale = scale, log = TRUE, inv = TRUE) * factor) / factor,
    -3.378
  )
  expect_error(
    dmvnorm(dwishart(x = x, df = 1, scale = scale)),
    "Assertion on 'df' failed: Element 1 is not >= 2."
  )
})

test_that("(Inverse-) Wishart can be drawn", {
  df <- 4
  scale <- matrix(c(1, 0.2, 0.2, 2), 2, 2)
  expect_equal(dim(rwishart(df = df, scale = scale, inv = FALSE)), c(2, 2))
  expect_equal(dim(rwishart(df = df, scale = scale, inv = TRUE)), c(2, 2))
  expect_error(dmvnorm(rwishart(df = 1, scale = numeric(3))))
})

#' Find the positions of first or last occurrence of unique vector elements
#'
#' @description
#' This function finds the positions of first or last occurrence of unique
#' vector elements.
#'
#' @param x \[`atomic()`\]\cr
#' A \code{vector}.
#'
#' @param type \[`character(1)`\]\cr
#' Either \code{"first"} for the first or \code{"last"} for the last occurrence.
#'
#' @return
#' An \code{integer} \code{vector}, the positions of the unique vector elements.
#' The ordering corresponds to \code{unique(x)}, i.e., the \eqn{i}-th element in
#' the output is the (first or last) occurrence of the \eqn{i}-th element from
#' \code{unique(x)}.
#'
#' @keywords indexing
#' @family vector helpers
#' @export
#'
#' @examples
#' x <- c(1, 1, 1, 2, 2, 2, 3, 3, 3)
#' unique(x)
#' vector_occurrence(x, "first")
#' vector_occurrence(x, "last")

vector_occurrence <- function(x, type = "first") {
  checkmate::assert_atomic_vector(x)
  checkmate::assert_choice(type, c("first", "last"))
  if (type == "first") {
    which(!duplicated(x))
  } else {
    length(x) - match(unique(x), rev(x)) + 1L
  }
}

test_that("finding occurrence in vector works", {
  x <- c("A", "A", "B", "C", "A")
  expect_equal(vector_occurrence(x, "first"), c(1L, 3L, 4L))
  expect_equal(vector_occurrence(x, "last"), c(5L, 3L, 4L))
})

#' Extract function body
#'
#' @description
#' This function extracts the body of a function as a single \code{character}.
#'
#' @param fun \[`function`\]\cr
#' A \code{function}.
#'
#' @param braces \[`logical(1)`\]\cr
#' Remove \code{"{"} and \code{"}"} at start and end (if any)?
#'
#' @param nchar \[`integer(1)`\]\cr
#' The maximum number of characters before abbreviation, at least \code{3}.
#'
#' @return
#' A \code{character}, the body of \code{f}.
#'
#' @keywords functional
#' @family function helpers
#' @export
#'
#' @examples
#' fun <- mean.default
#' function_body(fun)
#' function_body(fun, braces = TRUE)
#' function_body(fun, nchar = 30)

function_body <- function(fun, braces = FALSE, nchar = getOption("width") - 4) {
  checkmate::assert_function(fun)
  checkmate::assert_flag(braces)
  checkmate::assert_int(nchar, lower = 3)
  out <- deparse1(body(fun))
  if (!braces) out <- gsub("^[{]|[}]$", "", out)
  out <- trimws(gsub("\\s+", " ", out))
  if (nchar(out) > nchar) out <- paste0(strtrim(out, nchar - 3), "...")
  out
}

test_that("extraction of function body as character works", {
  test_fun <- function(x) {
    stopifnot(is.numeric(x))
    {
      x + 1
    }
  }
  expect_equal(
    function_body(test_fun),
    "stopifnot(is.numeric(x)) { x + 1 }"
  )
  expect_equal(
    function_body(test_fun, braces = TRUE),
    "{ stopifnot(is.numeric(x)) { x + 1 } }"
  )
  expect_equal(
    function_body(test_fun, nchar = 20),
    "stopifnot(is.nume..."
  )
})

#' Generate equidistant vectors in Euclidean space
#'
#' @description
#' This function constructs the coordinates of vertices of a regular simplex
#' in \eqn{\mathbb{R}^{\code{dim}}} and returns the first \code{n} of them,
#'
#' * scaled so that the pairwise Euclidean distance between any two vertices
#'   equals \code{dist},
#'
#' * and centered so their centroid is at \code{center}.
#'
#' @param dim \[`integer(1)`\]\cr
#' The dimension.
#'
#' @param n \[`integer(1)`\]\cr
#' The number of vertices to return. Cannot be larger than \code{dim + 1}.
#'
#' @param dist \[`numeric(1)`\]\cr
#' Desired pairwise Euclidean distance between any two vertices.
#'
#' @param center \[`numeric(dim)`\]\cr
#' Desired center.
#'
#' @return
#' A matrix, where each column is a vertex of the simplex.
#'
#' @keywords simulation
#' @family vector helpers
#' @export
#'
#' @examples
#' dim <- n <- 3
#' (dist <- runif(1))
#' (center <- rnorm(dim))
#' (V <- equidistant_vectors(dim = dim, n = n, dist = dist, center = center))
#' rowMeans(V)
#' dist(t(V))

equidistant_vectors <- function(
    dim, n = dim + 1, dist = 1, center = rep(0, dim)
  ) {
  input_check_response(
    check = checkmate::check_count(dim, positive = TRUE),
    var_name = "dim"
  )
  input_check_response(
    check = checkmate::check_int(n, lower = 1, upper = dim + 1),
    var_name = "n"
  )
  input_check_response(
    check = checkmate::check_number(dist, lower = 0, finite = TRUE),
    var_name = "dist"
  )
  input_check_response(
    check = check_numeric_vector(
      center, finite = TRUE, any.missing = FALSE, len = dim
    ),
    var_name = "center"
  )
  b <- matrix(c(1, -1), ncol = 2)
  for (k in seq_len(dim)[-1]) {
    b <- rbind(
      cbind(1, matrix(-1 / k, ncol = k)),
      cbind(0, sqrt(1 - 1 / k^2) * b)
    )
  }
  b <- b * dist / mean(as.vector(dist(t(b))))
  b_sub <- b[, seq_len(n), drop = FALSE]
  b_sub <- sweep(b_sub, 1, rowMeans(b_sub))
  b_sub <- sweep(b_sub, 1, center, FUN = "+")
  return(b_sub)
}

test_that("equidistant vectors in Euclidean space can be generated", {

  ### basic cases
  V <- equidistant_vectors(dim = 1, n = 2)
  expect_true(is.matrix(V))
  expect_identical(dim(V), c(1L, 2L))
  V <- equidistant_vectors(dim = 3, n = 4)
  expect_identical(dim(V), c(3L, 4L))

  ### center is respected
  center <- c(0.5, -1, 2)
  for (n in 1:4) {
    V <- equidistant_vectors(dim = 3, n = n, dist = 1, center = center)
    expect_lt(sqrt(sum((rowMeans(V) - center)^2)), 1e-10)
  }

  ### distance is respected
  set.seed(123)
  for (d in 1:5) {
    for (n in seq_len(d + 1)[-1]) {
      dist_target <- runif(1, min = 0.1, max = 3)
      center <- rnorm(d)
      V <- equidistant_vectors(
        dim = d, n = n, dist = dist_target, center = center
      )
      pw <- as.vector(stats::dist(t(V)))
      expect_lt(abs(mean(pw) - dist_target), 1e-8)
    }
  }
})

#' Build permutations
#'
#' @description
#' This function creates all permutations of a given \code{vector}.
#'
#' @references
#' Modified version of <https://stackoverflow.com/a/20199902/15157768>.
#'
#' @param x \[`atomic()`\]\cr
#' Any \code{vector}.
#'
#' @return
#' A \code{list} of all permutations of \code{x}.
#'
#' @export
#' @keywords indexing
#' @family vector helpers
#'
#' @examples
#' permutations(1:3)
#' permutations(LETTERS[1:3])

permutations <- function(x) {
  checkmate::assert_atomic_vector(x, min.len = 1)
  permutation_index <- function(n) {
    if (n == 1) {
      return(matrix(1))
    } else {
      sp <- permutation_index(n - 1)
      p <- nrow(sp)
      A <- matrix(nrow = n * p, ncol = n)
      for (i in 1:n) {
        A[(i - 1) * p + 1:p, ] <- cbind(i, sp + (sp >= i))
      }
      return(A)
    }
  }
  p <- permutation_index(length(x))
  apply(p, 1, function(p) x[p], simplify = FALSE)
}

test_that("computation of permutations works", {
  expect_equal(
    permutations(1:3),
    list(1:3, c(1L, 3L, 2L), c(2L, 1L, 3L), c(2L, 3L, 1L), c(3L, 1L, 2L), 3:1)
  )
  expect_equal(
    permutations(LETTERS[1:3]),
    list(
      c("A", "B", "C"), c("A", "C", "B"), c("B", "A", "C"), c("B", "C", "A"),
      c("C", "A", "B"), c("C", "B", "A")
    )
  )
})

#' Stationary distribution
#'
#' @description
#' This function computes the stationary distribution corresponding to a
#' transition probability matrix.
#'
#' @param tpm \[`matrix()`\]\cr
#' A transition probability matrix.
#'
#' @param soft_fail \[`logical(1)`\]\cr
#' Return the discrete uniform distribution if the computation of the stationary
#' distribution fails for some reason? Else, throw an error.
#'
#' @return
#' A \code{numeric} vector.
#'
#' @keywords transformation
#' @family matrix helpers
#' @export
#'
#' @examples
#' tpm <- matrix(0.05, nrow = 3, ncol = 3)
#' diag(tpm) <- 0.9
#' stationary_distribution(tpm)

stationary_distribution <- function(tpm, soft_fail = FALSE) {
  input_check_response(
    check = check_transition_probability_matrix(tpm),
    var_name = "tpm"
  )
  input_check_response(
    check = checkmate::check_flag(soft_fail),
    var_name = "soft_fail"
  )
  dim <- nrow(tpm)
  stat_distr <- try(solve(t(diag(dim) - tpm + 1), rep(1, dim)), silent = TRUE)
  if (inherits(stat_distr, "try-error")) {
    if (soft_fail) {
      stat_distr <- rep(1, dim) / dim
    } else {
      stop("singular system")
    }
  }
  return(stat_distr)
}

test_that("stationary distribution can be computed", {
  tpm <- matrix(0.05, nrow = 3, ncol = 3)
  diag(tpm) <- 0.9
  expect_equal(
    stationary_distribution(tpm),
    c(1, 1, 1) / 3
  )
})

test_that("soft fail works", {
  tpm <- matrix(c(
    1, 0, 0,
    0.5, 0.5, 0,
    0, 0, 1
  ), nrow = 3, byrow = TRUE)
  expect_equal(
    stationary_distribution(tpm, soft_fail = TRUE),
    c(1, 1, 1) / 3
  )
  expect_error(
    stationary_distribution(tpm),
    "singular system"
  )
})

#' Provide information about occurrences
#'
#' @description
#' This function provides verbose information about absolute or relative
#' element occurrences in `data.frame` columns.
#'
#' @param x \[`data.frame`\]\cr
#' The object to check for occurrences.
#'
#' @param relative \[`logical(1)` \]\cr
#' The number of rows or columns to be printed, greater or equal `2`.
#'
#' @param named \[`logical(1)` \]\cr
#' Prepend column names of `x` (if not `NA`)?
#'
#' @return
#' A `character()`.
#'
#' @export
#' @keywords packaging
#' @family data.frame helpers
#'
#' @examples
#' occurrence_info(datasets::warpbreaks, relative = FALSE, named = TRUE)

occurrence_info <- function(x, relative = FALSE, named = FALSE) {

  ### input checks
  input_check_response(
    check = checkmate::check_data_frame(x),
    var_name = "x"
  )
  input_check_response(
    check = checkmate::check_flag(relative),
    var_name = "relative"
  )
  input_check_response(
    check = checkmate::check_flag(named),
    var_name = "named"
  )

  ### extract unique values
  r <- nrow(x)
  n <- ncol(x)
  m <- colnames(x)
  out <- character(n)
  table_values <- lapply(x, table, useNA = "ifany")
  for (i in seq_len(n)) {
    tab <- table_values[[i]] |> sort(decreasing = TRUE)
    occ <- if (relative) {
      paste0(ceiling(tab / r * 100), "% ", names(tab), collapse = ", ")
    } else {
      paste0(tab, "x ", names(tab), collapse = ", ")
    }
    out[i] <- if (named & checkmate::test_string(m[i])) {
      paste0(m[i], ": ", occ)
    } else {
      occ
    }
  }
  return(out)
}

test_that("multiplication works", {
  x <- data.frame("A" = c(1, 1, 2), "B" = LETTERS[1:3], c(NA, NA, NA))
  colnames(x)[3] <- NA
  expect_equal(
    occurrence_info(x, relative = FALSE, named = FALSE),
    c("2x 1, 1x 2", "1x A, 1x B, 1x C", "3x NA")
  )
  expect_equal(
    occurrence_info(x, relative = TRUE, named = FALSE),
    c("67% 1, 34% 2", "34% A, 34% B, 34% C", "100% NA")
  )
  expect_equal(
    occurrence_info(x, relative = FALSE, named = TRUE),
    c("A: 2x 1, 1x 2", "B: 1x A, 1x B, 1x C", "3x NA")
  )
  expect_equal(
    occurrence_info(x, relative = TRUE, named = TRUE),
    c("A: 67% 1, 34% 2", "B: 34% A, 34% B, 34% C", "100% NA")
  )
  expect_equal(
    occurrence_info(data.frame()),
    character()
  )
})

#' Check list of lists
#'
#' @description
#' These functions check whether the input is a list that contains list
#' elements.
#'
#' @inheritParams checkmate::check_list
#'
#' @return
#' Same as documented in \code{\link[checkmate]{check_list}}.
#'
#' @keywords validation
#' @family list helpers
#' @export
#'
#' @examples
#' L <- list(list(1), list(2), 3)
#' check_list_of_lists(L)
#' test_list_of_lists(L)
#' \dontrun{
#' assert_list_of_lists(L)
#' }

check_list_of_lists <- function(x, len = NULL) {
  res <- checkmate::check_list(x, len = len)
  if (!isTRUE(res)) {
    return(res)
  }
  for (i in seq_along(x)) {
    res <- checkmate::check_list(x[[i]], len = len)
    if (!isTRUE(res)) {
      return(paste("Check for element", i, "failed:", res))
    }
  }
  return(TRUE)
}

#' @rdname check_list_of_lists
#' @inheritParams checkmate::assert_list
#' @export

assert_list_of_lists <- checkmate::makeAssertionFunction(
  check_list_of_lists
)

#' @rdname check_list_of_lists
#' @inheritParams checkmate::assert_list
#' @export

test_list_of_lists <- checkmate::makeTestFunction(
  check_list_of_lists
)

#' Creating a basic logo for an R package
#'
#' @description
#' This function creates a basic R package logo. The logo has a white
#' background and the package name (with or without curly brackets) in the
#' center. The font size for the package name is scaled such that it fits inside
#' the logo. Type \code{?oeli} to see an example.
#'
#' @references
#' - This function builds upon \code{\link[hexSticker]{sticker}}.
#' - Use \code{\link[usethis]{use_logo}} to set up the logo for a package.
#'
#' @param package_name \[`character(1)`\]\cr
#' The package name.
#'
#' @param brackets \[`logical(1)`\]\cr
#' Curly brackets around the package name?
#'
#' @param background
#' A `ggplot` object, the background of the sticker.
#'
#' @param s_x,s_y,s_width,s_height,white_around_sticker
#' Passed on to \code{\link[hexSticker]{sticker}}.
#'
#' @return
#' A \code{ggplot} object.
#'
#' @export
#' @keywords packaging
#' @family package helpers
#'
#' @examples
#' print(package_logo("my_package", brackets = TRUE))

package_logo <- function(
    package_name, brackets = FALSE,
    background = ggplot2::ggplot() + ggplot2::theme_void(),
    s_x = 1,
    s_y = 1,
    s_width = 1,
    s_height = 1,
    white_around_sticker = FALSE
  ) {

  ### input checks
  input_check_response(
    check = checkmate::check_string(package_name),
    var_name = "package_name"
  )
  input_check_response(
    check = checkmate::check_flag(brackets),
    var_name = "brackets"
  )
  input_check_response(
    check = checkmate::check_class(background, "ggplot"),
    var_name = "background"
  )

  ### define path
  filename <- tempfile(
    pattern = paste("logo", package_name, sep = "_"),
    fileext = ".png"
  )

  ### optionally add brackets to package name
  if (brackets) {
    package_name <- paste0("{", package_name, "}")
  }

  ### determine font size for package name
  chars <- nchar(package_name)
  p_size <- 56 * exp(-0.09 * chars)

  ### build logo
  suppressWarnings(
    hexSticker::sticker(

      ### image
      subplot = background,
      s_x = s_x,
      s_y = s_y,
      s_width = s_width,
      s_height = s_height,

      ### package name
      package = package_name,
      p_x = 1,
      p_y = 1,
      p_color = "black",
      p_fontface = "plain",
      p_size = p_size,

      ### sticker
      h_size = 1.2,
      h_fill = "white",
      h_color = "black",
      spotlight = FALSE,
      white_around_sticker = white_around_sticker,

      ### URL
      url = "",
      u_x = 1,
      u_y = 0.1,
      u_color = "black",
      u_family = "my_font",
      u_size = 7,
      u_angle = 30,

      ### save file
      filename = filename,
      asp = 1,
      dpi = 300
    )
  )
}

test_that("package logo can be created", {
  expect_true(ggplot2::is_ggplot(package_logo("my_package")))
})

test_that("custom background can be used", {
  background <- ggplot2::ggplot(data = cars) +
    ggplot2::geom_point(ggplot2::aes(x = speed, y = dist))
  expect_true(
    ggplot2::is_ggplot(
      package_logo("my_package", background = background, brackets = TRUE)
    )
  )
})

#' User confirmation
#'
#' @description
#' This function asks in an interactive question a binary question.
#'
#' @param question \[`character(1)`\]\cr
#' The binary question to ask. It should end with a question mark.
#'
#' @param default \[`logical(1)`\]\cr
#' The default decision.
#'
#' @return
#' Either \code{TRUE} or \code{FALSE}.
#'
#' @keywords validation
#' @family package helpers
#' @export

user_confirm <- function(question = "Question?", default = FALSE) {
  checkmate::assert_flag(default)
  if (!interactive()) {
    return(default)
  }
  checkmate::assert_string(question)
  repeat {
    selection <- ifelse(default, "[Y/n]", "[y/N]")
    cli::cli_inform("{question} {selection}:")
    response <- tryCatch(
      tolower(trimws(readline())),
      interrupt = identity
    )
    if (inherits(response, "interrupt")) {
      stop()
    }
    if (!nzchar(response)) {
      return(default)
    }
    if (response %in% c("y", "yes")) {
      return(TRUE)
    }
    if (response %in% c("n", "no")) {
      return(FALSE)
    }
    cli::cli_inform("Please enter 'y' or 'n'.")
  }
}

#' Get default function arguments
#'
#' @description
#' This function returns the default function arguments (if any).
#'
#' @param f \[`function`\]\cr
#' A \code{function}.
#'
#' @param exclude \[`NULL` | `character()`\]\cr
#' Argument names to exclude.
#'
#' Can be \code{NULL} (default) to not exclude any argument names.
#'
#' @return
#' A named \code{list}.
#'
#' @keywords functional
#' @family function helpers
#' @export
#'
#' @examples
#' f <- function(a, b = 1, c = "", ...) { }
#' function_defaults(f)
#' function_defaults(f, exclude = "b")

function_defaults <- function(f, exclude = NULL) {
  input_check_response(
    checkmate::check_function(f),
    "f"
  )
  input_check_response(
    checkmate::check_character(exclude, null.ok = TRUE),
    "exclude"
  )
  formals_f <- formals(f)
  formals_f <- formals_f[!sapply(formals_f, is.symbol)]
  formals_f <- formals_f[!names(formals_f) %in% exclude]
  return(formals_f)
}

test_that("default values of a function can be returned", {
  f <- function(a, b = 1, c = "", ...) { }
  expect_equal(
    function_defaults(f),
    list(b = 1, c = "")
  )
  expect_equal(
    function_defaults(f, exclude = "b"),
    list(c = "")
  )
})

#' Get indices of matrix diagonal
#'
#' @description
#' This function returns the indices of the diagonal elements of a quadratic
#' matrix.
#'
#' @param n \[`integer(1)`\]\cr
#' The matrix dimension.
#'
#' @param triangular \[`NULL` or `character(1)`\]\cr
#' If \code{NULL} (default), all elements of the matrix are considered. If
#' \code{"lower"} (\code{"upper"}), only the lower- (upper-) triangular matrix
#' is considered.
#'
#' @return
#' An \code{integer} \code{vector}.
#'
#' @export
#' @keywords indexing
#' @family matrix helpers
#'
#' @examples
#' # indices of diagonal elements
#' n <- 3
#' matrix(1:n^2, n, n)
#' matrix_diagonal_indices(n)
#'
#' # indices of diagonal elements of lower-triangular matrix
#' L <- matrix(0, n, n)
#' L[lower.tri(L, diag=TRUE)] <- 1:((n * (n + 1)) / 2)
#' L
#' matrix_diagonal_indices(n, triangular = "lower")
#'
#' # indices of diagonal elements of upper-triangular matrix
#' U <- matrix(0, n, n)
#' U[upper.tri(U, diag=TRUE)] <- 1:((n * (n + 1)) / 2)
#' U
#' matrix_diagonal_indices(n, triangular = "upper")

matrix_diagonal_indices <- function(n, triangular = NULL) {
  checkmate::assert_int(n, lower = 1)
  if (is.null(triangular)) {
    M <- matrix(1:n^2, n, n)
  } else {
    M <- matrix(0, n, n)
    triangular <- match_arg(triangular, c("lower", "upper"))
    if (triangular == "lower") {
      ind <- lower.tri(M, diag = TRUE)
    } else {
      ind <- upper.tri(M, diag = TRUE)
    }
    M[ind] <- 1:((n * (n + 1)) / 2)
  }
  as.integer(diag(M))
}

test_that("can get indices of matrix diagonal", {
  expect_identical(matrix_diagonal_indices(3), c(1L, 5L, 9L))
  expect_identical(matrix_diagonal_indices(3, triangular = "lower"), c(1L, 4L, 6L))
  expect_identical(matrix_diagonal_indices(3, triangular = "upper"), c(1L, 3L, 6L))
})

#' Check one-hot matrix
#'
#' @description
#' These functions check whether the input is a one-hot matrix, i.e., a
#' `numeric` matrix where each row contains exactly one entry equal to `1` and
#' all other entries equal to `0`.
#'
#' @inheritParams check_covariance_matrix
#' @inheritParams checkmate::check_matrix
#'
#' @return
#' Same as documented in \code{\link[checkmate]{check_matrix}}.
#'
#' @keywords validation
#' @family matrix helpers
#' @export
#'
#' @examples
#' x <- matrix(c(
#'   1, 0, 0, 0,
#'   0, 1, 0, 0,
#'   0, 0, 0, 0
#' ), nrow = 3, byrow = TRUE)
#'
#' check_one_hot_matrix(x)
#' test_one_hot_matrix(x)
#' \dontrun{
#' assert_one_hot_matrix(x)
#' }

check_one_hot_matrix <- function(
    x, nrows = NULL, ncols = NULL, tolerance = sqrt(.Machine$double.eps)
) {
  input_check_response(
    check = checkmate::check_number(tolerance, lower = 0),
    var_name = "tolerance"
  )
  res <- checkmate::check_matrix(
    x, mode = "numeric", nrows = nrows, ncols = ncols
  )
  if (!isTRUE(res)) {
    return(res)
  }
  if (anyNA(x)) {
    return("Must not contain missing values")
  }
  if (any(!is.finite(x))) {
    return("Must not contain infinite values")
  }
  is_zero <- abs(x) <= tolerance
  is_one <- abs(x - 1) <= tolerance
  if (any(!(is_zero | is_one))) {
    return("All elements must be 0 or 1")
  }
  if (any(rowSums(is_one) != 1L)) {
    return("Each row must contain exactly one 1")
  }
  return(TRUE)
}

#' @rdname check_one_hot_matrix
#' @inheritParams checkmate::assert_matrix
#' @export

assert_one_hot_matrix <- checkmate::makeAssertionFunction(
  check_one_hot_matrix
)

#' @rdname check_one_hot_matrix
#' @inheritParams checkmate::test_matrix
#' @export

test_one_hot_matrix <- checkmate::makeTestFunction(
  check_one_hot_matrix
)

#' Round \code{numeric} columns of a \code{data.frame}
#'
#' @description
#' This function rounds (only) the \code{numeric} columns of a
#' \code{data.frame}.
#'
#' @param df \[`data.frame`\]\cr
#' A \code{data.frame}.
#'
#' @param digits \[`integer(1)` | `NULL` \]\cr
#' The number of decimal places to be used.
#'
#' Negative values are allowed, resulting in rounding to a power of ten.
#'
#' Can be `NULL` to not round.
#'
#' @return
#' A \code{data.frame}.
#'
#' @keywords transformation
#' @family data.frame helpers
#' @export
#'
#' @examples
#' df <- data.frame("label" = c("A", "B"), "number" = rnorm(10))
#' round_data.frame(df, digits = 1)

round_data.frame <- function(df, digits = 0) {

  ### input checks
  input_check_response(
    check = checkmate::check_data_frame(df),
    var_name = "df"
  )
  input_check_response(
    check = checkmate::check_int(digits, null.ok = TRUE),
    var_name = "digits"
  )

  ### rounding
  if (!is.null(digits)) {
    df[] <- lapply(df, function(x) if (is.numeric(x)) round(x, digits) else x)
  }
  return(df)
}

test_that("rounding numeric data.frame columns works", {
  set.seed(1)
  df <- data.frame("label" = c("A", "B"), "number" = rnorm(10) * 10)
  expect_equal(
    round_data.frame(df, NULL), df
  )
  expect_equal(
    round_data.frame(df),
    structure(
      list(
        label = c("A", "B", "A", "B", "A", "B", "A", "B", "A", "B"),
        number = c(-6, 2, -8, 16, 3, -8, 5, 7, 6, -3)
      ),
      row.names = c(NA, -10L),
      class = "data.frame"
    )
  )
})

test_that("check for correlation matrix works", {
  # Test a non-numeric matrix
  non_numeric_matrix <- matrix(letters[1:9], nrow = 3)
  expect_error(
    assert_correlation_matrix(non_numeric_matrix),
    "Must store numerics"
  )

  # Test a non-square matrix
  non_square_matrix <- matrix(1:6, nrow = 2)
  expect_error(
    assert_correlation_matrix(non_square_matrix),
    "Must be square"
  )

  # Test a non-symmetric matrix
  non_symmetric_matrix <- matrix(c(1, 2, 3, 4), nrow = 2)
  expect_error(
    assert_correlation_matrix(non_symmetric_matrix),
    "Must be symmetric"
  )

  # Test a matrix with bad diagonal
  matrix_with_bad_diagonal <- matrix(c(1, 0, 0, 0), nrow = 2)
  expect_error(
    assert_correlation_matrix(matrix_with_bad_diagonal),
    "Assertion on 'matrix_with_bad_diagonal' failed: Must have ones on the diagonal."
  )

  # Test a matrix with bad correlations
  matrix_with_bad_correlations <- matrix(c(1, 2, 2, 1), nrow = 2)
  expect_error(
    assert_correlation_matrix(matrix_with_bad_correlations),
    "Assertion on 'matrix_with_bad_correlations' failed: Must have values between -1 and 1."
  )

  # Test a matrix with bad eigenvalues
  matrix_with_bad_eigenvalues <- matrix(c(
    1,  0.9,  0.9,
    0.9,  1,  -0.9,
    0.9,  -0.9,  1), nrow = 3, byrow = TRUE)
  expect_error(
    assert_correlation_matrix(matrix_with_bad_eigenvalues),
    "Assertion on 'matrix_with_bad_eigenvalues' failed: Must have positive eigenvalues only."
  )

  # Test a matrix with the wrong dimensions
  matrix_with_wrong_dimensions <- diag(3)
  expect_error(
    assert_correlation_matrix(matrix_with_wrong_dimensions, dim = 2),
    "Must be of dimension 2"
  )

  # Test a matrix with the correct dimensions
  matrix_with_correct_dimensions <- diag(2)
  expect_silent(assert_correlation_matrix(matrix_with_correct_dimensions, dim = 2))

  # Test a matrix with NAs
  matrix_with_nas <- matrix(c(1, NA, NA, 1), 2, 2)
  expect_error(
    assert_correlation_matrix(matrix_with_nas),
    "Assertion on 'matrix_with_nas' failed: Must not have NA values."
  )

  # Test a matrix with infinite values
  matrix_with_inf <- matrix(c(1, Inf, Inf, 1), 2, 2)
  expect_error(
    assert_correlation_matrix(matrix_with_inf),
    "Assertion on 'matrix_with_inf' failed: Must not have infinite values."
  )
})

test_that("check for probability vector works", {
  # Test a non-numeric vector
  non_numeric_vector <- letters[1:9]
  expect_error(
    assert_probability_vector(non_numeric_vector),
    "Must be of type 'numeric', not 'character'."
  )

  # Test a vector with bad elements
  vector_with_bad_elements <- 0:2
  expect_error(
    assert_probability_vector(vector_with_bad_elements),
    "Assertion on 'vector_with_bad_elements' failed: Element 3 is not <= 1."
  )

  # Test a vector with bad sum
  vector_with_bad_sum <- c(1, 1:5 / sum(1:5))
  expect_error(
    assert_probability_vector(vector_with_bad_sum),
    "Assertion on 'vector_with_bad_sum' failed: Must add up to 1."
  )

  # Test a vector with the wrong dimensions
  vector_with_wrong_length <- 1:5 / sum(1:5)
  expect_error(
    assert_probability_vector(vector_with_wrong_length, len = 4),
    "Assertion on 'vector_with_wrong_length' failed: Must have length 4, but has length 5."
  )

  # Test a vector with the correct dimensions
  vector_with_correct_length <- 1:5 / sum(1:5)
  expect_silent(assert_probability_vector(vector_with_correct_length, len = 5))
})

test_that("check for transition probability matrix works", {
  # Test a non-numeric matrix
  non_numeric_matrix <- matrix(letters[1:9], nrow = 3)
  expect_error(
    assert_transition_probability_matrix(non_numeric_matrix),
    "Must store numerics"
  )

  # Test a non-square matrix
  non_square_matrix <- matrix(1:6, nrow = 2)
  expect_error(
    assert_transition_probability_matrix(non_square_matrix),
    "Must be square"
  )

  # Test a matrix with bad rows
  matrix_with_bad_rows <- matrix(c(1, 2, 2, 1), nrow = 2)
  expect_error(
    assert_transition_probability_matrix(matrix_with_bad_rows),
    "Assertion on 'matrix_with_bad_rows' failed: Must have values between 0 and 1."
  )

  # Test a matrix with bad row sum
  matrix_with_bad_row_sums <- matrix(c(1, 1, 0, 1), nrow = 2)
  expect_error(
    assert_transition_probability_matrix(matrix_with_bad_row_sums),
    "Assertion on 'matrix_with_bad_row_sums' failed: Must have row sums equal to 1."
  )

  # Test a matrix with the wrong dimensions
  matrix_with_wrong_dimensions <- diag(3)
  expect_error(
    assert_transition_probability_matrix(matrix_with_wrong_dimensions, dim = 2),
    "Must be of dimension 2"
  )

  # Test a matrix with the correct dimensions
  matrix_with_correct_dimensions <- diag(2)
  expect_silent(assert_transition_probability_matrix(matrix_with_correct_dimensions, dim = 2))

  # Deal with machine epsilon
  matrix_rounding_issues <- structure(
    c(0.96413163700029, 0.342824884561103, 0.0358683629997102, 0.657175115438897),
    dim = c(2L, 2L)
  )
  expect_false(
    test_transition_probability_matrix(matrix_rounding_issues, tolerance = 0)
  )
  expect_true(
    test_transition_probability_matrix(matrix_rounding_issues)
  )

  # Test a matrix with NAs
  matrix_with_nas <- matrix(c(0.5, NA, NA, 0.5), 2, 2)
  expect_error(
    assert_transition_probability_matrix(matrix_with_nas),
    "Assertion on 'matrix_with_nas' failed: Must not have NA values."
  )

  # Test a matrix with infinite values
  matrix_with_inf <- matrix(c(0.5, Inf, Inf, 0.5), 2, 2)
  expect_error(
    assert_transition_probability_matrix(matrix_with_inf),
    "Assertion on 'matrix_with_inf' failed: Must not have infinite values."
  )
})

test_that("check for one-hot matrix works", {
  # Test a non-numeric matrix
  non_numeric_matrix <- matrix(letters[1:9], nrow = 3)
  expect_error(
    assert_one_hot_matrix(non_numeric_matrix),
    "Must store numerics"
  )

  # Test a matrix with missing values
  matrix_with_nas <- matrix(c(1, 0, NA, 1), nrow = 2)
  expect_error(
    assert_one_hot_matrix(matrix_with_nas),
    "Assertion on 'matrix_with_nas' failed: Must not contain missing values."
  )

  # Test a matrix with infinite values
  matrix_with_inf <- matrix(c(1, 0, Inf, 1), nrow = 2)
  expect_error(
    assert_one_hot_matrix(matrix_with_inf),
    "Assertion on 'matrix_with_inf' failed: Must not contain infinite values."
  )

  # Test a matrix with values different from 0 and 1
  matrix_with_bad_values <- matrix(c(1, 0, 0.2, 0.8), nrow = 2)
  expect_error(
    assert_one_hot_matrix(matrix_with_bad_values, tolerance = 0),
    "Assertion on 'matrix_with_bad_values' failed: All elements must be 0 or 1."
  )

  # Test a matrix with no 1 in a row
  matrix_with_no_one <- matrix(c(1, 0, 0, 0), nrow = 2)
  expect_error(
    assert_one_hot_matrix(matrix_with_no_one),
    "Assertion on 'matrix_with_no_one' failed: Each row must contain exactly one 1."
  )

  # Test a matrix with more than one 1 in a row
  matrix_with_multiple_ones <- matrix(c(1, 0, 1, 1), nrow = 2)
  expect_error(
    assert_one_hot_matrix(matrix_with_multiple_ones),
    "Assertion on 'matrix_with_multiple_ones' failed: Each row must contain exactly one 1."
  )

  # Test a matrix with the wrong number of rows
  matrix_with_wrong_nrow <- diag(2)
  expect_error(
    assert_one_hot_matrix(matrix_with_wrong_nrow, nrows = 3),
    "Must have exactly 3 rows, but has 2 rows."
  )

  # Test a matrix with the wrong number of columns
  matrix_with_wrong_ncol <- diag(2)
  expect_error(
    assert_one_hot_matrix(matrix_with_wrong_ncol, ncols = 3),
    "Must have exactly 3 cols, but has 2 cols."
  )

  # Deal with machine epsilon
  matrix_rounding_issues <- structure(
    c(1, 0, sqrt(.Machine$double.eps) / 2, 1 - sqrt(.Machine$double.eps) / 2),
    dim = c(2L, 2L)
  )
  expect_false(
    test_one_hot_matrix(matrix_rounding_issues, tolerance = 0)
  )
  expect_true(
    test_one_hot_matrix(matrix_rounding_issues)
  )

  # Test a valid one-hot matrix
  valid_one_hot_matrix <- matrix(c(
    1, 0, 0,
    0, 1, 0,
    0, 0, 1
  ), nrow = 3, byrow = TRUE)
  expect_silent(
    assert_one_hot_matrix(valid_one_hot_matrix)
  )
  expect_true(
    test_one_hot_matrix(valid_one_hot_matrix)
  )
})

test_that("check for list of lists works", {
  # Test an empty list
  empty_list <- list()
  expect_true(test_list_of_lists(empty_list))
  expect_error(
    assert_list_of_lists(empty_list, len = 1),
    "Assertion on 'empty_list' failed: Must have length 1, but has length 0."
  )

  # Test a list with non-list elements
  non_list_elements <- list(1, 2)
  expect_error(
    assert_list_of_lists(non_list_elements),
    "Assertion on 'non_list_elements' failed: Check for element 1 failed: Must be of type 'list', not 'double'."
  )

  # Test a list with only list elements
  only_list_elements <- list(list(), list(1))
  expect_true(test_list_of_lists(only_list_elements))
})

# matrix printing works

    Code
      print_matrix(x = 1)
    Output
      1

---

    Code
      print_matrix(x = 1.5, label = "single numeric")
    Output
      single numeric : 1.5

---

    Code
      print_matrix(x = 1.5, label = "single numeric", simplify = TRUE)
    Output
      single numeric : 1.5

---

    Code
      print_matrix(x = LETTERS[1:26])
    Output
      character vector of length 26 
      A B C ... Z

---

    Code
      print_matrix(x = LETTERS[1:26], label = "letters")
    Output
      letters : character vector of length 26 
      A B C ... Z

---

    Code
      print_matrix(x = LETTERS[1:26], label = "letters", details = FALSE)
    Output
      letters : A B C ... Z

---

    Code
      print_matrix(x = LETTERS[1:26], label = "letters", simplify = TRUE)
    Output
      letters : A B C ... Z

---

    Code
      print_matrix(x = LETTERS[1:26], rowdots = 1)
    Output
      character vector of length 26 
      A B C ... Z

---

    Code
      print_matrix(x = LETTERS[1:26], rowdots = 26)
    Output
      character vector of length 26 
      A B C ... Z

---

    Code
      print_matrix(x = LETTERS[1:26], coldots = 1)
    Output
      character vector of length 26 
      ... Z

---

    Code
      print_matrix(x = LETTERS[1:26], coldots = 25)
    Output
      character vector of length 26 
      A B C D E F G H I J K L M N O P Q R S T U V W X ... Z

---

    Code
      print_matrix(x = LETTERS[1:26], coldots = 26)
    Output
      character vector of length 26 
      A B C D E F G H I J K L M N O P Q R S T U V W X Y Z

---

    Code
      print_matrix(x = LETTERS[1:26], coldots = 10)
    Output
      character vector of length 26 
      A B C D E F G H I ... Z

---

    Code
      print_matrix(x = matrix(LETTERS[1:24], ncol = 6))
    Output
      4 x 6 matrix of characters 
           [,1] [,2] [,3] ... [,6]
      [1,]    A    E    I ...    U
      [2,]    B    F    J ...    V
      [3,]    C    G    K ...    W
      [4,]    D    H    L ...    X

---

    Code
      print_matrix(x = matrix(LETTERS[1:24], ncol = 6), label = "big matrix")
    Output
      big matrix : 4 x 6 matrix of characters 
           [,1] [,2] [,3] ... [,6]
      [1,]    A    E    I ...    U
      [2,]    B    F    J ...    V
      [3,]    C    G    K ...    W
      [4,]    D    H    L ...    X

---

    Code
      print_matrix(x = matrix(LETTERS[1:24], ncol = 6), rowdots = 2)
    Output
      4 x 6 matrix of characters 
           [,1] [,2] [,3] ... [,6]
      [1,]    A    E    I ...    U
      ...   ...  ...  ... ...  ...
      [4,]    D    H    L ...    X

---

    Code
      print_matrix(x = matrix(LETTERS[1:24], ncol = 6), details = FALSE)
    Output
           [,1] [,2] [,3] ... [,6]
      [1,]    A    E    I ...    U
      [2,]    B    F    J ...    V
      [3,]    C    G    K ...    W
      [4,]    D    H    L ...    X

---

    Code
      print_matrix(x = matrix(LETTERS[1:24], ncol = 6), coldots = 1, rowdots = 1)
    Output
      4 x 6 matrix of characters 
           ... [,6]
      ...  ...  ...
      [4,] ...    X

---

    Code
      print_matrix(x = matrix(LETTERS[1:24], ncol = 6), simplify = TRUE, coldots = 2,
      rowdots = 2)
    Output
      [ A ... U; ... ... ...; D ... X ]

---

    Code
      print_matrix(x = diag(5), coldots = 2, rowdots = 3, simplify = TRUE, digits = 0)
    Output
      [ 1 ... 0; 0 ... 0; ... ... ...; 0 ... 1 ]

---

    Code
      print_matrix(x = matrix(1:100, nrow = 1))
    Output
      1 x 100 matrix of doubles 
           [,1] [,2] [,3] ... [,100]
      [1,]    1    2    3 ...    100

---

    Code
      print_matrix(x = matrix(1:100, nrow = 1), label = "single row matrix")
    Output
      single row matrix : 1 x 100 matrix of doubles 
           [,1] [,2] [,3] ... [,100]
      [1,]    1    2    3 ...    100

---

    Code
      print_matrix(x = matrix(1:100, nrow = 1), details = FALSE)
    Output
           [,1] [,2] [,3] ... [,100]
      [1,]    1    2    3 ...    100

---

    Code
      print_matrix(x = matrix(1:100, nrow = 1), coldots = 1)
    Output
      1 x 100 matrix of doubles 
           ... [,100]
      [1,] ...    100

---

    Code
      print_matrix(x = matrix(1:100, nrow = 1), rowdots = 1)
    Output
      1 x 100 matrix of doubles 
           [,1] [,2] [,3] ... [,100]
      [1,]    1    2    3 ...    100

---

    Code
      print_matrix(x = matrix(1:100, nrow = 1), simplify = TRUE, coldots = 5)
    Output
      [ 1 2 3 4 ... 100 ]

---

    Code
      print_matrix(x = matrix(1:100, ncol = 1))
    Output
      100 x 1 matrix of doubles 
             [,1]
      [1,]      1
      [2,]      2
      [3,]      3
      ...     ...
      [100,]  100

---

    Code
      print_matrix(x = matrix(1:100, ncol = 1), label = "single column matrix")
    Output
      single column matrix : 100 x 1 matrix of doubles 
             [,1]
      [1,]      1
      [2,]      2
      [3,]      3
      ...     ...
      [100,]  100

---

    Code
      print_matrix(x = matrix(1:100, ncol = 1), details = FALSE)
    Output
             [,1]
      [1,]      1
      [2,]      2
      [3,]      3
      ...     ...
      [100,]  100

---

    Code
      print_matrix(x = matrix(1:100, ncol = 1), coldots = 1)
    Output
      100 x 1 matrix of doubles 
             [,1]
      [1,]      1
      [2,]      2
      [3,]      3
      ...     ...
      [100,]  100

---

    Code
      print_matrix(x = matrix(1:100, ncol = 1), rowdots = 1)
    Output
      100 x 1 matrix of doubles 
             [,1]
      ...     ...
      [100,]  100

---

    Code
      print_matrix(x = matrix(1:100, ncol = 1), simplify = TRUE, coldots = 5)
    Output
      [ 1; 2; 3; ...; 100 ]

---

    Code
      print_matrix(x = diag(2), simplify = TRUE)
    Output
      [ 1 0; 0 1 ]

---

    Code
      print_matrix(x = diag(2), simplify = FALSE)
    Output
      2 x 2 matrix of doubles 
           [,1] [,2]
      [1,]    1    0
      [2,]    0    1

---

    Code
      print_matrix(structure(diag(3), dimnames = list(c("la", "le", "lu"), c("x", "y",
        "z"))))
    Output
      3 x 3 matrix of doubles 
         x y z
      la 1 0 0
      le 0 1 0
      lu 0 0 1

# printing abbreviated data.frames works

    Code
      print_data.frame(x, rows = 7)
    Output
         X1.10 LETTERS.1.10. stats..rnorm.10.
      1   1    A             -0.6264538      
      2   2    B              0.1836433      
      3   3    C             -0.8356286      
      4   4    D              1.5952808      
      
    Message
      <3 rows hidden>
    Output
                                             
      8   8    H              0.7383247      
      9   9    I              0.5757814      
      10 10    J             -0.3053884      

---

    Code
      print_data.frame(x, rows = 7, cols = 2)
    Output
         X1.10 <1 col hidden> stats..rnorm.10.
      1   1          -        -0.6264538      
      2   2          -         0.1836433      
      3   3          -        -0.8356286      
      4   4          -         1.5952808      
      
    Message
      <3 rows hidden>
    Output
                                              
      8   8          -         0.7383247      
      9   9          -         0.5757814      
      10 10          -        -0.3053884      

---

    Code
      print_data.frame(x, rows = 7, cols = 2, digits = 1)
    Output
         X1.10 <1 col hidden> stats..rnorm.10.
      1   1          -        -0.6            
      2   2          -         0.2            
      3   3          -        -0.8            
      4   4          -         1.6            
      
    Message
      <3 rows hidden>
    Output
                                              
      8   8          -         0.7            
      9   9          -         0.6            
      10 10          -        -0.3            

---

    Code
      print_data.frame(x, rows = 7, cols = 2, digits = 1, row.names = FALSE)
    Output
       X1.10 <1 col hidden> stats..rnorm.10.
        1          -        -0.6            
        2          -         0.2            
        3          -        -0.8            
        4          -         1.6            
      
    Message
      <3 rows hidden>
    Output
                                            
        8          -         0.7            
        9          -         0.6            
       10          -        -0.3            

---

    Code
      print_data.frame(x, rows = 7, cols = 2, digits = 1, col.names = FALSE)
    Output
            <1 col hidden>     
      1   1       -        -0.6
      2   2       -         0.2
      3   3       -        -0.8
      4   4       -         1.6
      
    Message
      <3 rows hidden>
    Output
                               
      8   8       -         0.7
      9   9       -         0.6
      10 10       -        -0.3

---

    Code
      print_data.frame(data.frame())
    Message
      < data.frame with 0 columns and 0 rows >

---

    Code
      print_data.frame(data.frame(x = numeric()))
    Message
      < data.frame with 1 column and 0 rows >

# dictionary with optimizers works

    Code
      print(optimizer_dictionary)
    Output
      <Dictionary> 
      No elements contained.

---

    Code
      print(optimizer_dictionary)
    Output
      <Dictionary> 
      Keys: 
      - stats::nlm
      - stats::optim

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/occurrence_info.R
\name{occurrence_info}
\alias{occurrence_info}
\title{Provide information about occurrences}
\usage{
occurrence_info(x, relative = FALSE, named = FALSE)
}
\arguments{
\item{x}{[\code{data.frame}]\cr
The object to check for occurrences.}

\item{relative}{[\code{logical(1)} ]\cr
The number of rows or columns to be printed, greater or equal \code{2}.}

\item{named}{[\code{logical(1)} ]\cr
Prepend column names of \code{x} (if not \code{NA})?}
}
\value{
A \code{character()}.
}
\description{
This function provides verbose information about absolute or relative
element occurrences in \code{data.frame} columns.
}
\examples{
occurrence_info(datasets::warpbreaks, relative = FALSE, named = TRUE)
}
\seealso{
Other data.frame helpers: 
\code{\link{delete_columns_data.frame}()},
\code{\link{group_data.frame}()},
\code{\link{round_data.frame}()}
}
\concept{data.frame helpers}
\keyword{packaging}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insert_matrix_column.R
\name{insert_matrix_column}
\alias{insert_matrix_column}
\title{Insert column in matrix}
\usage{
insert_matrix_column(A, x, p)
}
\arguments{
\item{A}{[\code{matrix()}]\cr
A \code{matrix}.}

\item{x}{[\code{atomic()}]\cr
The column to be added, of length \code{nrow(A)}.

Can also be a single value.}

\item{p}{[\verb{integer())}]\cr
The position(s) where to add the column, one or more of:
\itemize{
\item \code{p = 0} appends the column left
\item \code{p = ncol(A)} appends the column right
\item \code{p = n} inserts the column between the \code{n}-th and
\code{(n + 1)}-th column of \code{A}.
}}
}
\value{
A \code{matrix}.
}
\description{
This function inserts a column into a matrix.
}
\examples{
A <- diag(3)
x <- 1:3
insert_matrix_column(A, x, 0)
insert_matrix_column(A, x, 1)
insert_matrix_column(A, x, 2)
insert_matrix_column(A, x, 3)

### also single value
x <- 2
insert_matrix_column(A, x, 0)

### also multiple positions
insert_matrix_column(A, x, 0:3)

### also trivial case
insert_matrix_column(matrix(nrow = 0, ncol = 0), integer(), integer())
}
\seealso{
Other matrix helpers: 
\code{\link{check_correlation_matrix}()},
\code{\link{check_covariance_matrix}()},
\code{\link{check_one_hot_matrix}()},
\code{\link{check_transition_probability_matrix}()},
\code{\link{cov_to_chol}()},
\code{\link{diff_cov}()},
\code{\link{matrix_diagonal_indices}()},
\code{\link{matrix_indices}()},
\code{\link{sample_correlation_matrix}()},
\code{\link{sample_covariance_matrix}()},
\code{\link{sample_transition_probability_matrix}()},
\code{\link{stationary_distribution}()}
}
\concept{matrix helpers}
\keyword{transformation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulate_markov_chain.R
\name{simulate_markov_chain}
\alias{simulate_markov_chain}
\title{Simulate Markov chain}
\usage{
simulate_markov_chain(Gamma, T, delta = oeli::stationary_distribution(Gamma))
}
\arguments{
\item{Gamma}{[\code{matrix()}]\cr
A transition probability matrix.}

\item{T}{[\code{integer(1)}]\cr
The length of the Markov chain.}

\item{delta}{[\code{numeric()}]\cr
A probability vector, the initial distribution.

The stationary distribution is used by default.}
}
\value{
A \code{numeric} vector of length \code{T} with states.
}
\description{
This function simulates a Markov chain.
}
\examples{
Gamma <- matrix(c(0.8, 0.2, 0.3, 0.7), byrow = TRUE, nrow = 2)
delta <- c(0.6, 0.4)
simulate_markov_chain(Gamma = Gamma, T = 20, delta = delta)
}
\seealso{
Other simulation helpers: 
\code{\link{Simulator}},
\code{\link{correlated_regressors}()},
\code{\link{ddirichlet_cpp}()},
\code{\link{dmixnorm_cpp}()},
\code{\link{dmvnorm_cpp}()},
\code{\link{dtnorm_cpp}()},
\code{\link{dwishart_cpp}()},
\code{\link{gaussian_tv}()}
}
\concept{simulation helpers}
\keyword{simulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/function_body.R
\name{function_body}
\alias{function_body}
\title{Extract function body}
\usage{
function_body(fun, braces = FALSE, nchar = getOption("width") - 4)
}
\arguments{
\item{fun}{[\code{function}]\cr
A \code{function}.}

\item{braces}{[\code{logical(1)}]\cr
Remove \code{"{"} and \code{"}"} at start and end (if any)?}

\item{nchar}{[\code{integer(1)}]\cr
The maximum number of characters before abbreviation, at least \code{3}.}
}
\value{
A \code{character}, the body of \code{f}.
}
\description{
This function extracts the body of a function as a single \code{character}.
}
\examples{
fun <- mean.default
function_body(fun)
function_body(fun, braces = TRUE)
function_body(fun, nchar = 30)
}
\seealso{
Other function helpers: 
\code{\link{do.call_timed}()},
\code{\link{function_arguments}()},
\code{\link{function_defaults}()},
\code{\link{quiet}()},
\code{\link{timed}()},
\code{\link{try_silent}()},
\code{\link{variable_name}()}
}
\concept{function helpers}
\keyword{functional}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/matrix_diagonal_indices.R
\name{matrix_diagonal_indices}
\alias{matrix_diagonal_indices}
\title{Get indices of matrix diagonal}
\usage{
matrix_diagonal_indices(n, triangular = NULL)
}
\arguments{
\item{n}{[\code{integer(1)}]\cr
The matrix dimension.}

\item{triangular}{[\code{NULL} or \code{character(1)}]\cr
If \code{NULL} (default), all elements of the matrix are considered. If
\code{"lower"} (\code{"upper"}), only the lower- (upper-) triangular matrix
is considered.}
}
\value{
An \code{integer} \code{vector}.
}
\description{
This function returns the indices of the diagonal elements of a quadratic
matrix.
}
\examples{
# indices of diagonal elements
n <- 3
matrix(1:n^2, n, n)
matrix_diagonal_indices(n)

# indices of diagonal elements of lower-triangular matrix
L <- matrix(0, n, n)
L[lower.tri(L, diag=TRUE)] <- 1:((n * (n + 1)) / 2)
L
matrix_diagonal_indices(n, triangular = "lower")

# indices of diagonal elements of upper-triangular matrix
U <- matrix(0, n, n)
U[upper.tri(U, diag=TRUE)] <- 1:((n * (n + 1)) / 2)
U
matrix_diagonal_indices(n, triangular = "upper")
}
\seealso{
Other matrix helpers: 
\code{\link{check_correlation_matrix}()},
\code{\link{check_covariance_matrix}()},
\code{\link{check_one_hot_matrix}()},
\code{\link{check_transition_probability_matrix}()},
\code{\link{cov_to_chol}()},
\code{\link{diff_cov}()},
\code{\link{insert_matrix_column}()},
\code{\link{matrix_indices}()},
\code{\link{sample_correlation_matrix}()},
\code{\link{sample_covariance_matrix}()},
\code{\link{sample_transition_probability_matrix}()},
\code{\link{stationary_distribution}()}
}
\concept{matrix helpers}
\keyword{indexing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/package_logo.R
\name{package_logo}
\alias{package_logo}
\title{Creating a basic logo for an R package}
\usage{
package_logo(
  package_name,
  brackets = FALSE,
  background = ggplot2::ggplot() + ggplot2::theme_void(),
  s_x = 1,
  s_y = 1,
  s_width = 1,
  s_height = 1,
  white_around_sticker = FALSE
)
}
\arguments{
\item{package_name}{[\code{character(1)}]\cr
The package name.}

\item{brackets}{[\code{logical(1)}]\cr
Curly brackets around the package name?}

\item{background}{A \code{ggplot} object, the background of the sticker.}

\item{s_x, s_y, s_width, s_height, white_around_sticker}{Passed on to \code{\link[hexSticker]{sticker}}.}
}
\value{
A \code{ggplot} object.
}
\description{
This function creates a basic R package logo. The logo has a white
background and the package name (with or without curly brackets) in the
center. The font size for the package name is scaled such that it fits inside
the logo. Type \code{?oeli} to see an example.
}
\examples{
print(package_logo("my_package", brackets = TRUE))
}
\references{
\itemize{
\item This function builds upon \code{\link[hexSticker]{sticker}}.
\item Use \code{\link[usethis]{use_logo}} to set up the logo for a package.
}
}
\seealso{
Other package helpers: 
\code{\link{Dictionary}},
\code{\link{Storage}},
\code{\link{check_missing}()},
\code{\link{find_namespace_calls}()},
\code{\link{identical_structure}()},
\code{\link{input_check_response}()},
\code{\link{match_arg}()},
\code{\link{print_data.frame}()},
\code{\link{print_matrix}()},
\code{\link{system_information}()},
\code{\link{unexpected_error}()},
\code{\link{user_confirm}()}
}
\concept{package helpers}
\keyword{packaging}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R, R/dirichlet.R
\name{ddirichlet_cpp}
\alias{ddirichlet_cpp}
\alias{rdirichlet_cpp}
\alias{ddirichlet}
\alias{rdirichlet}
\title{Dirichlet distribution}
\usage{
ddirichlet_cpp(x, concentration, log = FALSE)

rdirichlet_cpp(concentration)

ddirichlet(x, concentration, log = FALSE)

rdirichlet(n = 1, concentration)
}
\arguments{
\item{x}{[\code{numeric()}]\cr
A probability vector.}

\item{concentration}{[\code{numeric()}]\cr
A concentration vector of the same length as \code{x}.}

\item{log}{[\code{logical(1)}]\cr
Return the logarithm of the density value?}

\item{n}{[\code{integer(1)}]\cr
The number of samples.}
}
\value{
For \code{ddirichlet()}: The density value.

For \code{rdirichlet()}: If \code{n = 1} a \code{vector} of length \code{p}, else
a \code{matrix} of dimension \code{n} times \code{p} with samples as rows.
}
\description{
The function \code{ddirichlet()} computes the density of a Dirichlet distribution.

The function \code{rdirichlet()} samples from a Dirichlet distribution.

The functions with suffix \verb{_cpp} perform no input checks, hence are faster.
}
\examples{
x <- c(0.5, 0.3, 0.2)
concentration <- 1:3

# compute density
ddirichlet(x = x, concentration = concentration)
ddirichlet(x = x, concentration = concentration, log = TRUE)

# sample
rdirichlet(concentration = 1:3)
rdirichlet(n = 4, concentration = 1:2)
}
\seealso{
Other simulation helpers: 
\code{\link{Simulator}},
\code{\link{correlated_regressors}()},
\code{\link{dmixnorm_cpp}()},
\code{\link{dmvnorm_cpp}()},
\code{\link{dtnorm_cpp}()},
\code{\link{dwishart_cpp}()},
\code{\link{gaussian_tv}()},
\code{\link{simulate_markov_chain}()}
}
\concept{simulation helpers}
\keyword{distribution}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/storage.R
\name{Storage}
\alias{Storage}
\title{Storage R6 Object}
\value{
The output depends on the method:
\itemize{
\item \code{$new()} returns a \code{Storage} object.
\item \code{$add()}, \code{$remove()}, and \code{$print()} invisibly return the
\code{Storage} object (to allow for method chaining)
\item \code{$get()} returns the requested element(s)
\item \code{$number()} returns an \code{integer}
\item \code{$indices()} return an \code{integer} \code{vector}
}
}
\description{
Provides a simple indexing interface for list elements based on R6.
Basically, it allows to store items in a list and to regain them based on
identifiers defined by the user.
}
\section{Setting identifiers}{

An identifier is a \code{character}, typically a binary property. Identifiers
can be negated by placing an exclamation mark (\code{"!"}) in front of them.
Identifiers that have been assigned to other elements previously do not need
to be specified again for new elements; instead, a default value can be used.
This default value can be defined either globally for all cases (via the
\code{$missing_identifier} field) or separately for each specific case (via
the method argument).
}

\section{User confirmation}{

If desired, the user can be asked for confirmation when adding, extracting,
or removing elements using identifiers. This behavior can be set globally
through the \code{$confirm} field or customized separately for each specific
case via the method argument.
}

\examples{
### 1. Create a `Storage` object:
my_storage <- Storage$new()

# 2. Add elements along with identifiers:
my_storage$
  add(42, c("number", "rational"))$
  add(pi, c("number", "!rational"))$
  add("fear of black cats", c("text", "!rational"))$
  add("wearing a seat belt", c("text", "rational"))$
  add(mean, "function")

# 3. What elements are stored?
print(my_storage)

# 4. Extract elements based on identifiers:
my_storage$get("rational")
my_storage$get("!rational")
my_storage$get(c("text", "!rational"))
my_storage$get("all") # get all elements
my_storage$get(c("text", "!text"))
my_storage$get(c("text", "!text"), logical = "or")

# 5. Extract elements based on ids:
my_storage$get(ids = 4:5)
my_storage$get(ids = 4:5, id_names = TRUE) # add the ids as names
}
\seealso{
Other package helpers: 
\code{\link{Dictionary}},
\code{\link{check_missing}()},
\code{\link{find_namespace_calls}()},
\code{\link{identical_structure}()},
\code{\link{input_check_response}()},
\code{\link{match_arg}()},
\code{\link{package_logo}()},
\code{\link{print_data.frame}()},
\code{\link{print_matrix}()},
\code{\link{system_information}()},
\code{\link{unexpected_error}()},
\code{\link{user_confirm}()}
}
\concept{package helpers}
\keyword{indexing}
\section{Active bindings}{
\if{html}{\out{<div class="r6-active-bindings">}}
\describe{
\item{\code{identifier}}{[\code{character()}]\cr
The identifiers used.}

\item{\code{confirm}}{[\code{logical(1)}]\cr
The default value for confirmations.}

\item{\code{missing_identifier}}{[\code{logical(1)}]\cr
The default value for not specified identifiers.}

\item{\code{hide_warnings}}{[\code{logical(1)}]\cr
Hide warnings (for example if unknown identifiers are selected)?}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Storage-new}{\code{Storage$new()}}
\item \href{#method-Storage-add}{\code{Storage$add()}}
\item \href{#method-Storage-get}{\code{Storage$get()}}
\item \href{#method-Storage-remove}{\code{Storage$remove()}}
\item \href{#method-Storage-number}{\code{Storage$number()}}
\item \href{#method-Storage-indices}{\code{Storage$indices()}}
\item \href{#method-Storage-print}{\code{Storage$print()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Storage-new"></a>}}
\if{latex}{\out{\hypertarget{method-Storage-new}{}}}
\subsection{Method \code{new()}}{
Initializing a \code{Storage} object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Storage$new()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Storage-add"></a>}}
\if{latex}{\out{\hypertarget{method-Storage-add}{}}}
\subsection{Method \code{add()}}{
Adding an element.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Storage$add(
  x,
  identifier,
  confirm = interactive() & self$confirm,
  missing_identifier = self$missing_identifier
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{x}}{[\code{any}]\cr
An object to be saved.}

\item{\code{identifier}}{[\code{character()}]\cr
Pne or more identifiers (the identifier \code{"all"} is reserved to select
all elements).}

\item{\code{confirm}}{[\code{logical(1)}]\cr
Prompted for confirmation?}

\item{\code{missing_identifier}}{[\code{logical(1)} | NA]\cr
The value for not specified identifiers.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Storage-get"></a>}}
\if{latex}{\out{\hypertarget{method-Storage-get}{}}}
\subsection{Method \code{get()}}{
Getting elements.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Storage$get(
  identifier = character(),
  ids = integer(),
  logical = "and",
  confirm = interactive() & self$confirm,
  missing_identifier = self$missing_identifier,
  id_names = FALSE
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{identifier}}{[\code{character()}]\cr
Pne or more identifiers (the identifier \code{"all"} is reserved to select
all elements).}

\item{\code{ids}}{[\code{integer()}]\cr
One or more ids.}

\item{\code{logical}}{[\code{character(1)}]\cr
In the case that multiple identifiers are selected, how should they be
combined? Options are:
\itemize{
\item \code{"and"} (the default): the identifiers are combined with logical and
(all identifiers must be \code{TRUE})
\item \code{"or"}: the identifiers are combined with logical or (at least one
identifier must be \code{TRUE})
}}

\item{\code{confirm}}{[\code{logical(1)}]\cr
Prompted for confirmation?}

\item{\code{missing_identifier}}{[\code{logical(1)} | NA]\cr
The value for not specified identifiers.}

\item{\code{id_names}}{[\code{logical(1)}]\cr
Name the elements according to their ids?}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Storage-remove"></a>}}
\if{latex}{\out{\hypertarget{method-Storage-remove}{}}}
\subsection{Method \code{remove()}}{
removing elements
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Storage$remove(
  identifier = character(),
  ids = integer(),
  logical = "and",
  confirm = interactive() & self$confirm,
  missing_identifier = self$missing_identifier,
  shift_ids = TRUE
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{identifier}}{[\code{character()}]\cr
Pne or more identifiers (the identifier \code{"all"} is reserved to select
all elements).}

\item{\code{ids}}{[\code{integer()}]\cr
One or more ids.}

\item{\code{logical}}{[\code{character(1)}]\cr
In the case that multiple identifiers are selected, how should they be
combined? Options are:
\itemize{
\item \code{"and"} (the default): the identifiers are combined with logical and
(all identifiers must be \code{TRUE})
\item \code{"or"}: the identifiers are combined with logical or (at least one
identifier must be \code{TRUE})
}}

\item{\code{confirm}}{[\code{logical(1)}]\cr
Prompted for confirmation?}

\item{\code{missing_identifier}}{[\code{logical(1)} | NA]\cr
The value for not specified identifiers.}

\item{\code{shift_ids}}{[\code{logical(1)}]\cr
Shift ids when in-between elements are removed?}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Storage-number"></a>}}
\if{latex}{\out{\hypertarget{method-Storage-number}{}}}
\subsection{Method \code{number()}}{
Computing the number of identified elements.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Storage$number(
  identifier = "all",
  missing_identifier = self$missing_identifier,
  logical = "and",
  confirm = FALSE
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{identifier}}{[\code{character()}]\cr
Pne or more identifiers (the identifier \code{"all"} is reserved to select
all elements).}

\item{\code{missing_identifier}}{[\code{logical(1)} | NA]\cr
The value for not specified identifiers.}

\item{\code{logical}}{[\code{character(1)}]\cr
In the case that multiple identifiers are selected, how should they be
combined? Options are:
\itemize{
\item \code{"and"} (the default): the identifiers are combined with logical and
(all identifiers must be \code{TRUE})
\item \code{"or"}: the identifiers are combined with logical or (at least one
identifier must be \code{TRUE})
}}

\item{\code{confirm}}{[\code{logical(1)}]\cr
Prompted for confirmation?}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Storage-indices"></a>}}
\if{latex}{\out{\hypertarget{method-Storage-indices}{}}}
\subsection{Method \code{indices()}}{
Returning indices based on defined identifiers.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Storage$indices(
  identifier = "all",
  logical = "and",
  confirm = interactive() & self$confirm
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{identifier}}{[\code{character()}]\cr
Pne or more identifiers (the identifier \code{"all"} is reserved to select
all elements).}

\item{\code{logical}}{[\code{character(1)}]\cr
In the case that multiple identifiers are selected, how should they be
combined? Options are:
\itemize{
\item \code{"and"} (the default): the identifiers are combined with logical and
(all identifiers must be \code{TRUE})
\item \code{"or"}: the identifiers are combined with logical or (at least one
identifier must be \code{TRUE})
}}

\item{\code{confirm}}{[\code{logical(1)}]\cr
Prompted for confirmation?}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Storage-print"></a>}}
\if{latex}{\out{\hypertarget{method-Storage-print}{}}}
\subsection{Method \code{print()}}{
Printing details of the saved elements.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Storage$print(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Currently not used.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R, R/mixnorm.R
\name{dmixnorm_cpp}
\alias{dmixnorm_cpp}
\alias{pmixnorm_cpp}
\alias{rmixnorm_cpp}
\alias{dmixnorm}
\alias{pmixnorm}
\alias{rmixnorm}
\title{Mixture of normal distributions}
\usage{
dmixnorm_cpp(x, mean, Sigma, proportions)

pmixnorm_cpp(x, mean, Sigma, proportions, abseps = 0.001)

rmixnorm_cpp(mean, Sigma, proportions)

dmixnorm(x, mean, Sigma, proportions)

pmixnorm(x, mean, Sigma, proportions, abseps = 0.001)

rmixnorm(n = 1, mean, Sigma, proportions)
}
\arguments{
\item{x}{[\code{numeric(p)}]\cr
A quantile vector of length \code{p}, where \code{p} is the dimension.}

\item{mean}{[\code{matrix(nrow = p, ncol = c)}]\cr
The mean vectors for each component in columns.}

\item{Sigma}{[\code{matrix(nrow = p^2, ncol = c)}]\cr
The vectorized covariance matrices for each component in columns.}

\item{proportions}{[\code{numeric(c)}]\cr
The non-negative mixing proportions for each components.

If proportions do not sum to unity, they are rescaled to do so.}

\item{abseps}{[\code{numeric(1)}]\cr
The absolute error tolerance.}

\item{n}{[\code{integer(1)}]\cr
The number of requested samples.}
}
\value{
For \code{dmixnorm()}: The density value.

For \code{pmixnorm()}: The value of the distribution function.

For \code{rmixnorm()}: If \code{n = 1} a \code{vector} of length \code{p} (note
that it is a column vector for \code{rmixnorm_cpp()}), else
a \code{matrix} of dimension \code{n} times \code{p} with samples as rows.
}
\description{
The function \code{dmixnorm()} computes the density of a mixture of multivariate
normal distribution.

The function \code{pmixnorm()} computes the cumulative distribution function of a
mixture of multivariate normal distribution.

The function \code{rmixnorm()} samples from a mixture of multivariate normal
distribution.

The functions with suffix \verb{_cpp} perform no input checks, hence are faster.

The univariate normal mixture is available as the special case \code{p = 1}.
}
\details{
\code{pmixnorm()} is based on \code{mvtnorm::pmvnorm} with the randomized
Quasi-Monte-Carlo procedure by Genz and Bretz. The argument \code{abseps} controls
the accuracy of the Gaussian integral approximation.
}
\examples{
x <- c(0, 0)
mean <- matrix(c(-1, -1, 0, 0), ncol = 2)
Sigma <- matrix(c(diag(2), diag(2)), ncol = 2)
proportions <- c(0.7, 0.3)

# compute density
dmixnorm(x = x, mean = mean, Sigma = Sigma, proportions = proportions)

# compute CDF
pmixnorm(x = x, mean = mean, Sigma = Sigma, proportions = proportions)

# sample
rmixnorm(n = 3, mean = mean, Sigma = Sigma, proportions = proportions)
}
\seealso{
Other simulation helpers: 
\code{\link{Simulator}},
\code{\link{correlated_regressors}()},
\code{\link{ddirichlet_cpp}()},
\code{\link{dmvnorm_cpp}()},
\code{\link{dtnorm_cpp}()},
\code{\link{dwishart_cpp}()},
\code{\link{gaussian_tv}()},
\code{\link{simulate_markov_chain}()}
}
\concept{simulation helpers}
\keyword{distribution}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print_matrix.R
\name{print_matrix}
\alias{print_matrix}
\title{Print (abbreviated) \code{matrix}}
\usage{
print_matrix(
  x,
  rowdots = 4,
  coldots = 4,
  digits = 2,
  label = NULL,
  simplify = FALSE,
  details = !simplify
)
}
\arguments{
\item{x}{[\code{atomic()} | \code{matrix}]\cr
The object to be printed.}

\item{rowdots}{[\code{integer(1)}]\cr
The row number which is replaced by \code{...}.}

\item{coldots}{[\code{integer(1)}]\cr
The column number which is replaced by \code{...}.}

\item{digits}{[\code{integer(1)}]\cr
The number of printed decimal places if input \code{x} is \code{numeric}.}

\item{label}{[\code{character(1)}]\cr
A label for \code{x}. Only printed if \code{simplify = FALSE}.}

\item{simplify}{[\code{logical(1)}]\cr
Simplify the output?}

\item{details}{[\code{logical(1)}]\cr
Print the type and dimension of \code{x}?}
}
\value{
Invisibly returns \code{x}.
}
\description{
This function prints a (possibly abbreviated) \code{matrix}.
}
\examples{
print_matrix(x = 1, label = "single numeric")
print_matrix(x = LETTERS[1:26], label = "letters")
print_matrix(x = 1:3, coldots = 2)
print_matrix(x = matrix(rnorm(99), ncol = 1), label = "single column matrix")
print_matrix(x = matrix(1:100, nrow = 1), label = "single row matrix")
print_matrix(x = matrix(LETTERS[1:24], ncol = 6), label = "big matrix")
print_matrix(x = diag(5), coldots = 2, rowdots = 2, simplify = TRUE)
}
\references{
This function is a modified version of \code{pprint()} from the \code{{ramify}}
package.
}
\seealso{
Other package helpers: 
\code{\link{Dictionary}},
\code{\link{Storage}},
\code{\link{check_missing}()},
\code{\link{find_namespace_calls}()},
\code{\link{identical_structure}()},
\code{\link{input_check_response}()},
\code{\link{match_arg}()},
\code{\link{package_logo}()},
\code{\link{print_data.frame}()},
\code{\link{system_information}()},
\code{\link{unexpected_error}()},
\code{\link{user_confirm}()}
}
\concept{package helpers}
\keyword{packaging}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/matrix_indices.R
\name{matrix_indices}
\alias{matrix_indices}
\title{Get \code{matrix} indices}
\usage{
matrix_indices(x, prefix = "", exclude_diagonal = FALSE)
}
\arguments{
\item{x}{[\code{matrix}]\cr
A \code{matrix}.}

\item{prefix}{[\code{character(1)}]\cr
A prefix for the indices.}

\item{exclude_diagonal}{[\code{logical(1)}]\cr
Exclude indices where row equals column?}
}
\value{
A \code{character} \code{vector}.
}
\description{
This function returns \code{matrix} indices as \code{character}.
}
\examples{
M <- diag(3)
matrix_indices(M)
matrix_indices(M, "M_")
matrix_indices(M, "M_", TRUE)
}
\seealso{
Other matrix helpers: 
\code{\link{check_correlation_matrix}()},
\code{\link{check_covariance_matrix}()},
\code{\link{check_one_hot_matrix}()},
\code{\link{check_transition_probability_matrix}()},
\code{\link{cov_to_chol}()},
\code{\link{diff_cov}()},
\code{\link{insert_matrix_column}()},
\code{\link{matrix_diagonal_indices}()},
\code{\link{sample_correlation_matrix}()},
\code{\link{sample_covariance_matrix}()},
\code{\link{sample_transition_probability_matrix}()},
\code{\link{stationary_distribution}()}
}
\concept{matrix helpers}
\keyword{indexing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/subsets.R
\name{subsets}
\alias{subsets}
\title{Generate vector subsets}
\usage{
subsets(v, n = seq_along(v))
}
\arguments{
\item{v}{[atomic()`]\cr
A vector of elements.}

\item{n}{[integer(1)`]\cr
The requested subset sizes.}
}
\value{
A \code{list}, each element is a subset of \code{v}.
}
\description{
This function generates subsets of a vector.
}
\examples{
v <- 1:3
subsets(v)
subsets(v, c(1, 3)) # only subsets of length 1 or 3
subsets(integer())  # trivial case works
}
\seealso{
Other vector helpers: 
\code{\link{check_numeric_vector}()},
\code{\link{check_probability_vector}()},
\code{\link{chunk_vector}()},
\code{\link{equidistant_vectors}()},
\code{\link{insert_vector_entry}()},
\code{\link{map_indices}()},
\code{\link{match_numerics}()},
\code{\link{permutations}()},
\code{\link{split_vector_at}()},
\code{\link{vector_occurrence}()}
}
\concept{vector helpers}
\keyword{indexing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/quiet.R
\name{quiet}
\alias{quiet}
\title{Silence R code}
\usage{
quiet(x, print_cat = TRUE, message = TRUE, warning = TRUE)
}
\arguments{
\item{x}{[\code{expression}]\cr
Any function or expression or value assignment expression.}

\item{print_cat}{[\code{logical(1)}]\cr
Silence \code{print()} and \code{cat()} outputs?}

\item{message}{[\code{logical(1)}]\cr
Silence messages?}

\item{warning}{[\code{logical(1)}]\cr
Silence warnings?}
}
\value{
Invisibly the expression \code{x}.
}
\description{
This function silences warnings, messages and any \code{cat()} or \code{print()}
output from R expressions or functions.
}
\examples{
f <- function() {
  warning("warning")
  message("message")
  cat("cat")
  print("print")
}
quiet(f())
}
\references{
This function is a modified version of \code{\link[spsUtil]{quiet}}.
}
\seealso{
Other function helpers: 
\code{\link{do.call_timed}()},
\code{\link{function_arguments}()},
\code{\link{function_body}()},
\code{\link{function_defaults}()},
\code{\link{timed}()},
\code{\link{try_silent}()},
\code{\link{variable_name}()}
}
\concept{function helpers}
\keyword{functional}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sample_covariance_matrix.R
\name{sample_covariance_matrix}
\alias{sample_covariance_matrix}
\title{Sample covariance matrix}
\usage{
sample_covariance_matrix(dim, df = dim, scale = diag(dim), diag = FALSE)
}
\arguments{
\item{dim}{[\code{integer(1)}]\cr
The dimension.}

\item{df}{[\code{integer(1)}]\cr
The degrees of freedom of the inverse Wishart distribution greater or equal
\code{dim}.}

\item{scale}{[\code{matrix()}]\cr
The scale covariance matrix of the inverse Wishart distribution of dimension
\code{dim}.}

\item{diag}{[\code{logical(1)}]\cr
Diagonal matrix?}
}
\value{
A covariance \code{matrix}.
}
\description{
This function samples a covariance matrix from an inverse Wishart
distribution.
}
\examples{
sample_covariance_matrix(dim = 3)
}
\seealso{
Other matrix helpers: 
\code{\link{check_correlation_matrix}()},
\code{\link{check_covariance_matrix}()},
\code{\link{check_one_hot_matrix}()},
\code{\link{check_transition_probability_matrix}()},
\code{\link{cov_to_chol}()},
\code{\link{diff_cov}()},
\code{\link{insert_matrix_column}()},
\code{\link{matrix_diagonal_indices}()},
\code{\link{matrix_indices}()},
\code{\link{sample_correlation_matrix}()},
\code{\link{sample_transition_probability_matrix}()},
\code{\link{stationary_distribution}()}
}
\concept{matrix helpers}
\keyword{simulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chunk_vector.R
\name{chunk_vector}
\alias{chunk_vector}
\title{Split a vector into chunks}
\usage{
chunk_vector(x, n, type = 1, strict = FALSE)
}
\arguments{
\item{x}{[atomic()`]\cr
A vector of elements.}

\item{n}{[\code{integer(1)}]\cr
A number smaller or equal \code{length(x)}.}

\item{type}{[\code{1} | \code{2}]\cr
Either
\itemize{
\item \code{1} (default) to split \code{x} into \code{n} chunks of equal size,
\item or \code{2} to split \code{x} into chunks of size \code{n}.
}}

\item{strict}{[\code{logical(1)}]\cr
Set to \code{TRUE} to fail if \code{length(x)} is not a multiple of \code{n},
or \code{FALSE} (default), else.}
}
\value{
A \code{list}.
}
\description{
This function either
\itemize{
\item splits a vector into \code{n} chunks of equal size (\code{type = 1}),
\item splits a vector into chunks of size \code{n} (\code{type = 2}).
}
}
\examples{
x <- 1:12
chunk_vector(x, n = 3, type = 1)
chunk_vector(x, n = 3, type = 2)
try(chunk_vector(x, n = 5, strict = TRUE))
}
\seealso{
Other vector helpers: 
\code{\link{check_numeric_vector}()},
\code{\link{check_probability_vector}()},
\code{\link{equidistant_vectors}()},
\code{\link{insert_vector_entry}()},
\code{\link{map_indices}()},
\code{\link{match_numerics}()},
\code{\link{permutations}()},
\code{\link{split_vector_at}()},
\code{\link{subsets}()},
\code{\link{vector_occurrence}()}
}
\concept{vector helpers}
\keyword{transformation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stationary_distribution.R
\name{stationary_distribution}
\alias{stationary_distribution}
\title{Stationary distribution}
\usage{
stationary_distribution(tpm, soft_fail = FALSE)
}
\arguments{
\item{tpm}{[\code{matrix()}]\cr
A transition probability matrix.}

\item{soft_fail}{[\code{logical(1)}]\cr
Return the discrete uniform distribution if the computation of the stationary
distribution fails for some reason? Else, throw an error.}
}
\value{
A \code{numeric} vector.
}
\description{
This function computes the stationary distribution corresponding to a
transition probability matrix.
}
\examples{
tpm <- matrix(0.05, nrow = 3, ncol = 3)
diag(tpm) <- 0.9
stationary_distribution(tpm)
}
\seealso{
Other matrix helpers: 
\code{\link{check_correlation_matrix}()},
\code{\link{check_covariance_matrix}()},
\code{\link{check_one_hot_matrix}()},
\code{\link{check_transition_probability_matrix}()},
\code{\link{cov_to_chol}()},
\code{\link{diff_cov}()},
\code{\link{insert_matrix_column}()},
\code{\link{matrix_diagonal_indices}()},
\code{\link{matrix_indices}()},
\code{\link{sample_correlation_matrix}()},
\code{\link{sample_covariance_matrix}()},
\code{\link{sample_transition_probability_matrix}()}
}
\concept{matrix helpers}
\keyword{transformation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/match_arg.R
\name{match_arg}
\alias{match_arg}
\title{Argument matching}
\usage{
match_arg(arg, choices, several.ok = FALSE, none.ok = FALSE)
}
\arguments{
\item{arg}{[\code{character()}]\cr
The function argument.}

\item{choices}{[\code{character()}]\cr
Allowed values for \code{arg}.}

\item{several.ok}{[\code{logical(1)}]\cr
Is \code{arg} allowed to have more than one element?}

\item{none.ok}{[\code{logical(1)}]\cr
Is \code{arg} allowed to have zero elements?}
}
\value{
The un-abbreviated version of the exact or unique partial match if there is
one. Otherwise, an error is signaled if \code{several.ok} is \code{FALSE}
or \code{none.ok} is \code{FALSE}.
When \code{several.ok} is \code{TRUE} and (at least) one element of
\code{arg} has a match, all un-abbreviated versions of matches are returned.
When \code{none.ok} is \code{TRUE} and \code{arg} has zero elements,
\code{character(0)} is returned.
}
\description{
This function matches function arguments and is a modified version of
\code{\link[base]{match.arg}}.
}
\seealso{
Other package helpers: 
\code{\link{Dictionary}},
\code{\link{Storage}},
\code{\link{check_missing}()},
\code{\link{find_namespace_calls}()},
\code{\link{identical_structure}()},
\code{\link{input_check_response}()},
\code{\link{package_logo}()},
\code{\link{print_data.frame}()},
\code{\link{print_matrix}()},
\code{\link{system_information}()},
\code{\link{unexpected_error}()},
\code{\link{user_confirm}()}
}
\concept{package helpers}
\keyword{packaging}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/variable_name.R
\name{variable_name}
\alias{variable_name}
\title{Determine variable name}
\usage{
variable_name(variable, fallback = "unnamed")
}
\arguments{
\item{variable}{[\code{any}]\cr
Any object.}

\item{fallback}{[\code{character(1)}]\cr
A fallback name if for some reason the actual variable name
(which must be a single \code{character}) cannot be determined.}
}
\value{
A \code{character}, the variable name.
}
\description{
This function tries to determine the name of a variable passed to a
\code{function}.
}
\examples{
variable_name(a)
f <- function(x) variable_name(x)
f(x = a)
}
\seealso{
Other function helpers: 
\code{\link{do.call_timed}()},
\code{\link{function_arguments}()},
\code{\link{function_body}()},
\code{\link{function_defaults}()},
\code{\link{quiet}()},
\code{\link{timed}()},
\code{\link{try_silent}()}
}
\concept{function helpers}
\keyword{functional}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_probability_vector.R
\name{check_probability_vector}
\alias{check_probability_vector}
\alias{assert_probability_vector}
\alias{test_probability_vector}
\title{Check probability vector}
\usage{
check_probability_vector(x, len = NULL, tolerance = sqrt(.Machine$double.eps))

assert_probability_vector(
  x,
  len = NULL,
  tolerance = sqrt(.Machine$double.eps),
  .var.name = checkmate::vname(x),
  add = NULL
)

test_probability_vector(x, len = NULL, tolerance = sqrt(.Machine$double.eps))
}
\arguments{
\item{x}{[\code{any}]\cr
Object to check.}

\item{len}{[\code{integer(1)}]\cr
Exact expected length of \code{x}.}

\item{tolerance}{[\code{numeric(1)}]\cr
A non-negative tolerance value.}

\item{.var.name}{[\code{character(1)}]\cr
Name of the checked object to print in assertions. Defaults to
the heuristic implemented in \code{\link[checkmate]{vname}}.}

\item{add}{[\code{AssertCollection}]\cr
Collection to store assertion messages. See \code{\link[checkmate]{AssertCollection}}.}
}
\value{
Same as documented in \code{\link[checkmate]{check_numeric}}.
}
\description{
These functions check whether the input fulfills the properties of a
probability matrix.
}
\examples{
p <- c(0.2, 0.3, 0.6)
check_probability_vector(p)
test_probability_vector(p)
\dontrun{
assert_probability_vector(p)
}
}
\seealso{
Other vector helpers: 
\code{\link{check_numeric_vector}()},
\code{\link{chunk_vector}()},
\code{\link{equidistant_vectors}()},
\code{\link{insert_vector_entry}()},
\code{\link{map_indices}()},
\code{\link{match_numerics}()},
\code{\link{permutations}()},
\code{\link{split_vector_at}()},
\code{\link{subsets}()},
\code{\link{vector_occurrence}()}
}
\concept{vector helpers}
\keyword{validation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_missing.R
\name{check_missing}
\alias{check_missing}
\alias{assert_missing}
\alias{test_missing}
\title{Check missing formal argument}
\usage{
check_missing(x)

assert_missing(x)

test_missing(x)
}
\arguments{
\item{x}{[\code{any}]\cr
A formal argument.}
}
\value{
Depending on the function prefix:
\itemize{
\item If the check is successful, \code{assert_missing()} returns \code{x} invisibly,
whereas \code{check_missing()} and \code{test_missing()} return \code{TRUE}.
\item If the check is not successful, \code{assert_missing()} throws an error message,
\code{test_missing()} returns \code{FALSE}, and \code{check_missing()} returns a string with
the error message.
}
}
\description{
These functions check whether a value was specified as an argument to a
function.
}
\examples{
f <- function(x) {
  check_missing(x)
}
f()

g <- function(x) {
  test_missing(x)
}
g()

h <- function(x) {
  assert_missing(x)
}
\dontrun{
h()
}
}
\seealso{
Other package helpers: 
\code{\link{Dictionary}},
\code{\link{Storage}},
\code{\link{find_namespace_calls}()},
\code{\link{identical_structure}()},
\code{\link{input_check_response}()},
\code{\link{match_arg}()},
\code{\link{package_logo}()},
\code{\link{print_data.frame}()},
\code{\link{print_matrix}()},
\code{\link{system_information}()},
\code{\link{unexpected_error}()},
\code{\link{user_confirm}()}
}
\concept{package helpers}
\keyword{validation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/system_information.R
\name{system_information}
\alias{system_information}
\title{General system level information}
\usage{
system_information()
}
\value{
A \code{list} with elements:
\itemize{
\item \code{maschine}, the model name of the device
\item \code{cores}, the number of cores
\item \code{ram}, the size of the RAM
\item \code{os}, the operating system
\item \code{rversion}, the R version used
}
}
\description{
This function returns a \code{list} of general system level information.
}
\examples{
system_information()
}
\seealso{
Other package helpers: 
\code{\link{Dictionary}},
\code{\link{Storage}},
\code{\link{check_missing}()},
\code{\link{find_namespace_calls}()},
\code{\link{identical_structure}()},
\code{\link{input_check_response}()},
\code{\link{match_arg}()},
\code{\link{package_logo}()},
\code{\link{print_data.frame}()},
\code{\link{print_matrix}()},
\code{\link{unexpected_error}()},
\code{\link{user_confirm}()}
}
\concept{package helpers}
\keyword{validation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/map_indices.R
\name{map_indices}
\alias{map_indices}
\title{Map indices}
\usage{
map_indices(indices, n)
}
\arguments{
\item{indices}{[\code{integer()}]\cr
An index vector, where each element specifies a group to be mapped from the
sequence.}

\item{n}{[\code{integer}]\cr
The size of each group of consecutive indices.}
}
\value{
An \code{integer} \code{vector}, containing the mapped indices according to the
specified group size.
}
\description{
This function maps indices from an input vector to corresponding sequences of
grouped indices. Each element from the input specifies a group to be mapped
from the sequence, determined by the grouping size \code{n}.
}
\details{
This function is useful when working with indices arranged in fixed-size
groups, where each group can be referenced by a single index. For example, if
indices are structured in chunks of 3, calling this function with \code{n = 3}
will map the corresponding groups of 3 consecutive indices for the given
input indices, see the examples.
}
\examples{
# Example: Map indices based on groups of 3
map_indices(c(1, 3, 5), 3)
}
\seealso{
Other vector helpers: 
\code{\link{check_numeric_vector}()},
\code{\link{check_probability_vector}()},
\code{\link{chunk_vector}()},
\code{\link{equidistant_vectors}()},
\code{\link{insert_vector_entry}()},
\code{\link{match_numerics}()},
\code{\link{permutations}()},
\code{\link{split_vector_at}()},
\code{\link{subsets}()},
\code{\link{vector_occurrence}()}
}
\concept{vector helpers}
\keyword{indexing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/group_data.frame.R
\name{group_data.frame}
\alias{group_data.frame}
\title{Grouping of a \code{data.frame}}
\usage{
group_data.frame(df, by, keep_by = TRUE)
}
\arguments{
\item{df}{[\code{data.frame}]\cr
A \code{data.frame}.}

\item{by}{[\code{character(1)}]\cr
The name of a column of \code{df} to group by.}

\item{keep_by}{[\code{logical(1)}]\cr
Keep the grouping column \code{by}?}
}
\value{
A \code{list} of \code{data.frame}s, subsets of \code{df}.
}
\description{
This function groups a \code{data.frame} according to values of one column.
}
\examples{
df <- data.frame("label" = c("A", "B"), "number" = 1:10)
group_data.frame(df = df, by = "label")
group_data.frame(df = df, by = "label", keep_by = FALSE)
}
\seealso{
Other data.frame helpers: 
\code{\link{delete_columns_data.frame}()},
\code{\link{occurrence_info}()},
\code{\link{round_data.frame}()}
}
\concept{data.frame helpers}
\keyword{transformation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R, R/tnorm.R
\name{dtnorm_cpp}
\alias{dtnorm_cpp}
\alias{dttnorm_cpp}
\alias{rtnorm_cpp}
\alias{rttnorm_cpp}
\alias{dtnorm}
\alias{dttnorm}
\alias{rtnorm}
\alias{rttnorm}
\title{Truncated normal distribution}
\usage{
dtnorm_cpp(x, mean, sd, point, above, log = FALSE)

dttnorm_cpp(x, mean, sd, lower, upper, log = FALSE)

rtnorm_cpp(mean, sd, point, above, log = FALSE)

rttnorm_cpp(mean, sd, lower, upper, log = FALSE)

dtnorm(x, mean, sd, point, above, log = FALSE)

dttnorm(x, mean, sd, lower, upper, log = FALSE)

rtnorm(mean, sd, point, above, log = FALSE)

rttnorm(mean, sd, lower, upper, log = FALSE)
}
\arguments{
\item{x}{[\code{numeric(1)}]\cr
A quantile.}

\item{mean}{[\code{numeric(1)}]\cr
The mean.}

\item{sd}{[\code{numeric(1)}]\cr
The non-negative standard deviation.}

\item{point, lower, upper}{[\code{numeric(1)}]\cr
The truncation point.}

\item{above}{[\code{logical(1)}]\cr
Truncate from above? Else, from below.}

\item{log}{[\code{logical(1)}]\cr
Return the logarithm of the density value?}
}
\value{
For \code{dtnorm()} and \code{dttnorm()}: The density value.

For \code{rtnorm()} and \code{rttnorm()}: The random draw
}
\description{
The function \code{dtnorm()} computes the density of a truncated normal
distribution.

The function \code{rtnorm()} samples from a truncated normal distribution.

The function \code{dttnorm()} and \code{rttnorm()} compute the density and sample from
a two-sided truncated normal distribution, respectively.

The functions with suffix \verb{_cpp} perform no input checks, hence are faster.
}
\examples{
x <- c(0, 0)
mean <- c(0, 0)
Sigma <- diag(2)

# compute density
dmvnorm(x = x, mean = mean, Sigma = Sigma)
dmvnorm(x = x, mean = mean, Sigma = Sigma, log = TRUE)

# sample
rmvnorm(n = 3, mean = mean, Sigma = Sigma)
rmvnorm(mean = mean, Sigma = Sigma, log = TRUE)
}
\seealso{
Other simulation helpers: 
\code{\link{Simulator}},
\code{\link{correlated_regressors}()},
\code{\link{ddirichlet_cpp}()},
\code{\link{dmixnorm_cpp}()},
\code{\link{dmvnorm_cpp}()},
\code{\link{dwishart_cpp}()},
\code{\link{gaussian_tv}()},
\code{\link{simulate_markov_chain}()}
}
\concept{simulation helpers}
\keyword{distribution}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/insert_vector_entry.R
\name{insert_vector_entry}
\alias{insert_vector_entry}
\title{Insert entry in vector}
\usage{
insert_vector_entry(v, x, p)
}
\arguments{
\item{v}{[\code{atomic()}]\cr
A \code{vector}.}

\item{x}{[\code{atomic(1)}]\cr
The entry to be added.}

\item{p}{[\verb{integer())}]\cr
The position(s) where to add the value, one or more of:
\itemize{
\item \code{p = 0} appends the value left
\item \code{p = length(v)} appends the value right
\item \code{p = n} inserts the value between the \code{n}-th and
\code{(n + 1)}-th entry of \code{v}.
}}
}
\value{
A \code{vector}.
}
\description{
This function inserts a value into a vector.
}
\examples{
v <- 1:3
x <- 0
insert_vector_entry(v, x, 0)
insert_vector_entry(v, x, 1)
insert_vector_entry(v, x, 2)
insert_vector_entry(v, x, 3)

### also multiple positions
insert_vector_entry(v, x, 0:3)

### also trivial case
insert_vector_entry(integer(), integer(), integer())
}
\seealso{
Other vector helpers: 
\code{\link{check_numeric_vector}()},
\code{\link{check_probability_vector}()},
\code{\link{chunk_vector}()},
\code{\link{equidistant_vectors}()},
\code{\link{map_indices}()},
\code{\link{match_numerics}()},
\code{\link{permutations}()},
\code{\link{split_vector_at}()},
\code{\link{subsets}()},
\code{\link{vector_occurrence}()}
}
\concept{vector helpers}
\keyword{transformation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/unexpected_error.R
\name{unexpected_error}
\alias{unexpected_error}
\title{Handling of an unexpected error}
\usage{
unexpected_error(
  msg = "Ups, an unexpected error occured.",
  issue_link = "https://github.com/loelschlaeger/oeli/issues"
)
}
\arguments{
\item{msg}{[\code{character(1)}]\cr
An error message.}

\item{issue_link}{[\code{character(1)}]\cr
The URL to an issues site.}
}
\value{
No return value, but it throws an error.
}
\description{
This function reacts to an unexpected error by throwing an error and linking
to an issue site with the request to submit an issue.
}
\seealso{
Other package helpers: 
\code{\link{Dictionary}},
\code{\link{Storage}},
\code{\link{check_missing}()},
\code{\link{find_namespace_calls}()},
\code{\link{identical_structure}()},
\code{\link{input_check_response}()},
\code{\link{match_arg}()},
\code{\link{package_logo}()},
\code{\link{print_data.frame}()},
\code{\link{print_matrix}()},
\code{\link{system_information}()},
\code{\link{user_confirm}()}
}
\concept{package helpers}
\keyword{packaging}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/split_vector_at.R
\name{split_vector_at}
\alias{split_vector_at}
\title{Split a vector at positions}
\usage{
split_vector_at(x, at)
}
\arguments{
\item{x}{[atomic()`]\cr
A vector of elements.}

\item{at}{[\code{integer()}]\cr
Index position(s) just before to split.

For example, \code{at = n} splits before the \code{n}th element of \code{x}.}
}
\value{
A \code{list}.
}
\description{
This function splits a vector at specific positions.
}
\examples{
x <- 1:10
split_vector_at(x, c(2, 3, 5, 7))
}
\references{
Based on https://stackoverflow.com/a/19274414.
}
\seealso{
Other vector helpers: 
\code{\link{check_numeric_vector}()},
\code{\link{check_probability_vector}()},
\code{\link{chunk_vector}()},
\code{\link{equidistant_vectors}()},
\code{\link{insert_vector_entry}()},
\code{\link{map_indices}()},
\code{\link{match_numerics}()},
\code{\link{permutations}()},
\code{\link{subsets}()},
\code{\link{vector_occurrence}()}
}
\concept{vector helpers}
\keyword{transformation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/correlated_regressors.R
\name{correlated_regressors}
\alias{correlated_regressors}
\title{Simulate correlated regressor values}
\usage{
correlated_regressors(
  labels,
  n = 100,
  marginals = list(),
  correlation = diag(length(labels)),
  verbose = FALSE
)
}
\arguments{
\item{labels}{[\code{character()}]\cr
Unique labels for the regressors.}

\item{n}{[\code{integer(1)}]\cr
The number of values per regressor.}

\item{marginals}{[\code{list()}]\cr
Optionally marginal distributions for regressors. If not specified,
standard normal marginal distributions are used.

Each list entry must be named according to a regressor label, and the
following distributions are currently supported:
\describe{
\item{discrete distributions}{
\itemize{
\item Poisson: \code{list(type = "poisson", lambda = ...)}
\item categorical: \code{list(type = "categorical", p = c(...))}
}
}
\item{continuous distributions}{
\itemize{
\item normal: \code{list(type = "normal", mean = ..., sd = ...)}
\item uniform: \code{list(type = "uniform", min = ..., max = ...)}
}
}
}}

\item{correlation}{[\code{matrix()}]\cr
A correlation matrix of dimension \code{length(labels)}, where the
\code{(p, q)}-th entry defines the correlation between regressor
\code{labels[p]} and \code{labels[q]}.}

\item{verbose}{[\code{logical(1)}]\cr
Print information about the simulated regressors?}
}
\value{
A \code{data.frame} with \code{n} rows and \code{length(labels)} columns.
}
\description{
This function simulates regressor values from various marginal distributions
with custom correlations.
}
\examples{
labels <- c("P", "C", "N1", "N2", "U")
n <- 100
marginals <- list(
  "P" = list(type = "poisson", lambda = 2),
  "C" = list(type = "categorical", p = c(0.3, 0.2, 0.5)),
  "N1" = list(type = "normal", mean = -1, sd = 2),
  "U" = list(type = "uniform", min = -2, max = -1)
)
correlation <- matrix(
  c(1, -0.3, -0.1, 0, 0.5,
    -0.3, 1, 0.3, -0.5, -0.7,
    -0.1, 0.3, 1, -0.3, -0.3,
    0, -0.5, -0.3, 1, 0.1,
    0.5, -0.7, -0.3, 0.1, 1),
  nrow = 5, ncol = 5
)
data <- correlated_regressors(
  labels = labels, n = n, marginals = marginals, correlation = correlation
)
head(data)
}
\references{
This function heavily depends on the \code{{SimMultiCorrData}} package.
}
\seealso{
Other simulation helpers: 
\code{\link{Simulator}},
\code{\link{ddirichlet_cpp}()},
\code{\link{dmixnorm_cpp}()},
\code{\link{dmvnorm_cpp}()},
\code{\link{dtnorm_cpp}()},
\code{\link{dwishart_cpp}()},
\code{\link{gaussian_tv}()},
\code{\link{simulate_markov_chain}()}
}
\concept{simulation helpers}
\keyword{simulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/input_check_response.R
\name{input_check_response}
\alias{input_check_response}
\title{Standardized response to input check}
\usage{
input_check_response(
  check,
  var_name = NULL,
  error = TRUE,
  prefix = "Input {.var {var_name}} is bad:"
)
}
\arguments{
\item{check}{[\code{TRUE} | \code{character(1)} | \code{list()}]\cr
Matches the return value of the \verb{check*} functions from the \code{{checkmate}}
package, i.e., either \code{TRUE} if the check was successful, or a \code{character}
(the error message) else.

Can also be a \code{list} of multiple such values for alternative criteria, where
at least one must be \code{TRUE} for a successful check.}

\item{var_name}{[\code{NULL} | \code{character(1)}]\cr
Optionally specifies the name of the input being checked. This name will be
used for the default value of the \code{prefix} argument.}

\item{error}{[\code{logical(1)}]\cr
If \code{check} is not \code{TRUE} (or no element in \code{check} is \code{TRUE}, if \code{check} is
a \code{list}), throw an error?}

\item{prefix}{[\code{character(1)}]\cr
A prefix for the thrown error message, only relevant if \code{error} is \code{TRUE}.}
}
\value{
\code{TRUE} if \code{check} is \code{TRUE} (or any element in \code{check} is \code{TRUE}, if \code{check}
is a \code{list}) . Else, depending on \code{error}:
\itemize{
\item If \code{error} is \code{TRUE}, throws an error.
\item If \code{error} is \code{FALSE}, returns \code{FALSE}.
}
}
\description{
This function provides a standardized response to input checks, ensuring
consistency.
}
\examples{
x <- "1"
y <- 1

### check is successful
input_check_response(
  check = checkmate::check_character(x),
  var_name = "x",
  error = TRUE
)

### alternative checks
input_check_response(
  check = list(
    checkmate::check_character(x),
    checkmate::check_character(y)
  ),
  var_name = "x",
  error = TRUE
)

### standardized check response
\dontrun{
input_check_response(
  check = checkmate::check_character(y),
  var_name = "y",
  error = TRUE
)

input_check_response(
  check = list(
    checkmate::check_flag(x),
    checkmate::check_character(y)
  ),
  var_name = "y",
  error = TRUE
)
}
}
\seealso{
Other package helpers: 
\code{\link{Dictionary}},
\code{\link{Storage}},
\code{\link{check_missing}()},
\code{\link{find_namespace_calls}()},
\code{\link{identical_structure}()},
\code{\link{match_arg}()},
\code{\link{package_logo}()},
\code{\link{print_data.frame}()},
\code{\link{print_matrix}()},
\code{\link{system_information}()},
\code{\link{unexpected_error}()},
\code{\link{user_confirm}()}
}
\concept{package helpers}
\keyword{packaging}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dictionary.R
\name{Dictionary}
\alias{Dictionary}
\title{Dictionary R6 Object}
\description{
Provides a simple key-value interface based on R6.
}
\examples{
# Managing variable metadata for a dataset

meta_dict <- Dictionary$new(
  key_name = "var_name",
  alias_name = "category",
  value_names = c("label", "type"),
  value_assert = alist(
    label = checkmate::assert_string(),
    type = checkmate::assert_choice(choices = c("numeric", "factor", "character"))
  ),
  allow_overwrite = FALSE,
  keys_reserved = c("id"),
  alias_choices = c("demographics", "outcome", "other"),
  dictionary_name = "Variable Metadata"
)

# Add entries to the dictionary
meta_dict$add(
  var_name = "age",
  label = "Age of respondent",
  type = "numeric",
  category = "demographics"
)

meta_dict$add(
  var_name = "gender",
  label = "Gender identity",
  type = "factor",
  category = "demographics"
)

meta_dict$add(
  var_name = "income",
  label = "Annual income in USD",
  type = "numeric",
  category = c("demographics", "outcome")
)

# Print dictionary
meta_dict$print()

# Retrieve full metadata for a variable
meta_dict$get("income")

# Retrieve a specific piece of metadata
meta_dict$get("income", value = "label")

# Show variables by category
meta_dict$alias
}
\seealso{
Other package helpers: 
\code{\link{Storage}},
\code{\link{check_missing}()},
\code{\link{find_namespace_calls}()},
\code{\link{identical_structure}()},
\code{\link{input_check_response}()},
\code{\link{match_arg}()},
\code{\link{package_logo}()},
\code{\link{print_data.frame}()},
\code{\link{print_matrix}()},
\code{\link{system_information}()},
\code{\link{unexpected_error}()},
\code{\link{user_confirm}()}
}
\concept{package helpers}
\keyword{indexing}
\section{Active bindings}{
\if{html}{\out{<div class="r6-active-bindings">}}
\describe{
\item{\code{keys}}{[\code{character()}]\cr
Available keys.}

\item{\code{alias}}{[\code{list()}]\cr
Available keys per alias value.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Dictionary-new}{\code{Dictionary$new()}}
\item \href{#method-Dictionary-add}{\code{Dictionary$add()}}
\item \href{#method-Dictionary-get}{\code{Dictionary$get()}}
\item \href{#method-Dictionary-remove}{\code{Dictionary$remove()}}
\item \href{#method-Dictionary-print}{\code{Dictionary$print()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dictionary-new"></a>}}
\if{latex}{\out{\hypertarget{method-Dictionary-new}{}}}
\subsection{Method \code{new()}}{
Initializing a new \code{Dictionary} object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dictionary$new(
  key_name,
  alias_name = NULL,
  value_names = character(),
  value_assert = alist(),
  allow_overwrite = TRUE,
  keys_reserved = character(),
  alias_choices = NULL,
  dictionary_name = NULL
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{key_name}}{[\code{character(1)}]\cr
The name for the key variable.}

\item{\code{alias_name}}{[\code{NULL} | \code{character(1)}]\cr
Optionally the name for the alias variable.}

\item{\code{value_names}}{[\code{character(0)}]\cr
The names of the values connected to a key.}

\item{\code{value_assert}}{[\code{alist(1)}]\cr
For each element in \code{value_names}, \code{values_assert} \emph{can} have an
identically named element of the form \code{checkmate::assert*(...)}, where
\code{...} can be any argument for the assertion function except for the
\code{x} argument.}

\item{\code{allow_overwrite}}{[\code{logical(1)}]\cr
Allow overwriting existing keys with new values?
Duplicate keys are never allowed.}

\item{\code{keys_reserved}}{[\code{character()}]\cr
Names that must not be used as keys.}

\item{\code{alias_choices}}{[\code{NULL} or \code{character()}]\cr
Optionally possible values for the alias. Can also be \code{NULL}, then all
alias values are allowed.}

\item{\code{dictionary_name}}{[\code{NULL} or \code{character()}]\cr
Optionally the name for the dictionary.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dictionary-add"></a>}}
\if{latex}{\out{\hypertarget{method-Dictionary-add}{}}}
\subsection{Method \code{add()}}{
Adding an element to the dictionary.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dictionary$add(...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{...}}{Values for
\itemize{
\item the key variable \code{key_name} (must be a single \code{character}),
\item the alias variable \code{alias_name} (optionally, must then be a
\code{character} \code{vector}),
\item all the variables specified for \code{value_names} (if any, they must
comply to the \code{value_assert} checks).
}}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dictionary-get"></a>}}
\if{latex}{\out{\hypertarget{method-Dictionary-get}{}}}
\subsection{Method \code{get()}}{
Getting elements from the dictionary.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dictionary$get(key, value = NULL)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{key}}{[\code{character(1)}]\cr
A value for the key variable \code{key_name}. Use the \code{$keys} method for
available keys.}

\item{\code{value}}{[\code{NULL} | \code{character(1)}]\cr
One of the elements in \code{value_names}, selecting the required value.
Can also be \code{NULL} (default) for all values connected to the
\code{key}, returned as a \code{list}.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dictionary-remove"></a>}}
\if{latex}{\out{\hypertarget{method-Dictionary-remove}{}}}
\subsection{Method \code{remove()}}{
Removing elements from the dictionary (and associated alias, if any).
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dictionary$remove(key)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{key}}{[\code{character(1)}]\cr
A value for the key variable \code{key_name}. Use the \code{$keys} method for
available keys.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Dictionary-print"></a>}}
\if{latex}{\out{\hypertarget{method-Dictionary-print}{}}}
\subsection{Method \code{print()}}{
Printing details of the dictionary.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Dictionary$print()}\if{html}{\out{</div>}}
}

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hermannslauf.R
\docType{data}
\name{hermann}
\alias{hermann}
\title{Hermannslauf}
\format{
A \code{tibble} with 54 rows and 8 columns:
\describe{
\item{edition}{the edition of the Hermannslauf}
\item{year}{the year}
\item{date}{the date}
\item{temp}{the temperature on that day at 12:00 noon; see details}
\item{winner_men}{the men's winner}
\item{seconds_men}{the men's winner's total time in seconds}
\item{winner_women}{the women's winner}
\item{seconds_women}{the women's winner's total time in seconds}
}
}
\source{
\url{https://de.wikipedia.org/wiki/Hermannslauf}
}
\usage{
hermann
}
\description{
The Hermannslauf is an annual long-distance road and trail race in Germany
that runs from Detmold to Bielefeld through the Teutoburg Forest. It is one
of the best-known running events in the region and is especially noted for
its demanding, hilly course of about 31 kilometers. This dataset contains
historical information on editions of the race, including the date,
temperature, and winning times for men and women.

From 1972 to 1976, the course followed the actual Hermannsweg for 30.4
kilometers, and in 1977 the first 13 kilometers were moved to a parallel
route. Since 2005, the total course length has been 31.1 kilometers; until
2004, it was 30.6 kilometers. In 2020, the Hermannslauf was canceled due to
the pandemic, and for the same reason, in 2021 it was exceptionally moved
from April to October.
}
\details{
Temperature in degrees Celsius, measured at 12:00 noon, obtained from
different sources:
\itemize{
\item from 1973 to 1992 and from 1998 to 2013 from
\url{https://de.weatherspark.com/}, Gütersloh Airport (approx. 30 km from start
and finish)
\item from 1993 to 1997 from \url{https://de.weatherspark.com/}, Wunstorf Air Base
(approx. 100 km from start and finish)
\item from 2014 to 2017 from \url{https://de.weatherspark.com/}, Hannover-Langenhagen
Airport (approx. 90 km from start and finish)
\item from 2018 onward from \url{https://meteostat.net/de/station/D7106},
Airfield Bielefeld-Windelsbleiche (approx. 20 km from start and finish)
}
}
\concept{data}
\keyword{data}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/function_arguments.R
\name{function_arguments}
\alias{function_arguments}
\title{Get function arguments}
\usage{
function_arguments(f, with_default = TRUE, with_ellipsis = TRUE)
}
\arguments{
\item{f}{[\code{function}]\cr
A \code{function}.}

\item{with_default}{[\code{logical(1)}]\cr
Include function arguments that have default values?}

\item{with_ellipsis}{[\code{logical(1)}]\cr
Include the \code{"..."} argument if present?}
}
\value{
A \code{character} vector.
}
\description{
This function returns the names of function arguments.
}
\examples{
f <- function(a, b = 1, c = "", ...) { }
function_arguments(f)
function_arguments(f, with_default = FALSE)
function_arguments(f, with_ellipsis = FALSE)
}
\seealso{
Other function helpers: 
\code{\link{do.call_timed}()},
\code{\link{function_body}()},
\code{\link{function_defaults}()},
\code{\link{quiet}()},
\code{\link{timed}()},
\code{\link{try_silent}()},
\code{\link{variable_name}()}
}
\concept{function helpers}
\keyword{functional}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_one_hot_matrix.R
\name{check_one_hot_matrix}
\alias{check_one_hot_matrix}
\alias{assert_one_hot_matrix}
\alias{test_one_hot_matrix}
\title{Check one-hot matrix}
\usage{
check_one_hot_matrix(
  x,
  nrows = NULL,
  ncols = NULL,
  tolerance = sqrt(.Machine$double.eps)
)

assert_one_hot_matrix(
  x,
  nrows = NULL,
  ncols = NULL,
  tolerance = sqrt(.Machine$double.eps),
  .var.name = checkmate::vname(x),
  add = NULL
)

test_one_hot_matrix(
  x,
  nrows = NULL,
  ncols = NULL,
  tolerance = sqrt(.Machine$double.eps)
)
}
\arguments{
\item{x}{[\code{any}]\cr
Object to check.}

\item{nrows}{[\code{integer(1)}]\cr
Exact number of rows.}

\item{ncols}{[\code{integer(1)}]\cr
Exact number of columns.}

\item{tolerance}{[\code{numeric(1)}]\cr
A non-negative tolerance value.}

\item{.var.name}{[\code{character(1)}]\cr
Name of the checked object to print in assertions. Defaults to
the heuristic implemented in \code{\link[checkmate]{vname}}.}

\item{add}{[\code{AssertCollection}]\cr
Collection to store assertion messages. See \code{\link[checkmate]{AssertCollection}}.}
}
\value{
Same as documented in \code{\link[checkmate]{check_matrix}}.
}
\description{
These functions check whether the input is a one-hot matrix, i.e., a
\code{numeric} matrix where each row contains exactly one entry equal to \code{1} and
all other entries equal to \code{0}.
}
\examples{
x <- matrix(c(
  1, 0, 0, 0,
  0, 1, 0, 0,
  0, 0, 0, 0
), nrow = 3, byrow = TRUE)

check_one_hot_matrix(x)
test_one_hot_matrix(x)
\dontrun{
assert_one_hot_matrix(x)
}
}
\seealso{
Other matrix helpers: 
\code{\link{check_correlation_matrix}()},
\code{\link{check_covariance_matrix}()},
\code{\link{check_transition_probability_matrix}()},
\code{\link{cov_to_chol}()},
\code{\link{diff_cov}()},
\code{\link{insert_matrix_column}()},
\code{\link{matrix_diagonal_indices}()},
\code{\link{matrix_indices}()},
\code{\link{sample_correlation_matrix}()},
\code{\link{sample_covariance_matrix}()},
\code{\link{sample_transition_probability_matrix}()},
\code{\link{stationary_distribution}()}
}
\concept{matrix helpers}
\keyword{validation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/permutations.R
\name{permutations}
\alias{permutations}
\title{Build permutations}
\usage{
permutations(x)
}
\arguments{
\item{x}{[\code{atomic()}]\cr
Any \code{vector}.}
}
\value{
A \code{list} of all permutations of \code{x}.
}
\description{
This function creates all permutations of a given \code{vector}.
}
\examples{
permutations(1:3)
permutations(LETTERS[1:3])
}
\references{
Modified version of \url{https://stackoverflow.com/a/20199902/15157768}.
}
\seealso{
Other vector helpers: 
\code{\link{check_numeric_vector}()},
\code{\link{check_probability_vector}()},
\code{\link{chunk_vector}()},
\code{\link{equidistant_vectors}()},
\code{\link{insert_vector_entry}()},
\code{\link{map_indices}()},
\code{\link{match_numerics}()},
\code{\link{split_vector_at}()},
\code{\link{subsets}()},
\code{\link{vector_occurrence}()}
}
\concept{vector helpers}
\keyword{indexing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R, R/mvnorm.R
\name{dmvnorm_cpp}
\alias{dmvnorm_cpp}
\alias{pmvnorm_cpp}
\alias{rmvnorm_cpp}
\alias{dmvnorm}
\alias{pmvnorm}
\alias{rmvnorm}
\title{Multivariate normal distribution}
\usage{
dmvnorm_cpp(x, mean, Sigma, log = FALSE)

pmvnorm_cpp(x, mean, Sigma, abseps = 0.001)

rmvnorm_cpp(mean, Sigma, log = FALSE)

dmvnorm(x, mean, Sigma, log = FALSE)

pmvnorm(x, mean, Sigma, abseps = 0.001)

rmvnorm(n = 1, mean, Sigma, log = FALSE)
}
\arguments{
\item{x}{[\code{numeric()}]\cr
A quantile vector of length \code{p}.}

\item{mean}{[\code{numeric()}]\cr
The mean vector of length \code{p}.

For the functions without suffix \verb{_cpp}, it can also be of length \code{1} for
convenience, then \code{rep(mean, p)} is considered.}

\item{Sigma}{[\code{matrix()}]\cr
The covariance matrix of dimension \code{p}.

For \code{rmvnorm()}, arbitrary dimensions (i.e., full rows and corresponding
columns) of \code{Sigma} can be \code{0}.

For the functions without suffix \verb{_cpp} and if \code{p = 1}, it can also be a
single \code{numeric} for convenience. Note that \code{Sigma} is this case is a
variance, which is a different format than in \code{stats::dnorm()} or
\code{stats::rnorm}, which require a standard deviation.}

\item{log}{[\code{logical(1)}]\cr
Consider the log-normal distribution?}

\item{abseps}{[\code{numeric(1)}]\cr
The absolute error tolerance.}

\item{n}{[\code{integer(1)}]\cr
The number of requested samples.}
}
\value{
For \code{dmvnorm()}: The density value.

For \code{pmvnorm()}: The value of the distribution function.

For \code{rmvnorm()}: If \code{n = 1} a \code{vector} of length \code{p} (note
that it is a column vector for \code{rmvnorm_cpp()}), else
a \code{matrix} of dimension \code{n} times \code{p} with samples as rows.
}
\description{
The function \code{dmvnorm()} computes the density of a multivariate normal
distribution.

The function \code{pmvnorm()} computes the cumulative distribution function of a
multivariate normal distribution.

The function \code{rmvnorm()} samples from a multivariate normal distribution.

The functions with suffix \verb{_cpp} perform no input checks, hence are faster.

The univariate normal distribution is available as the special case \code{p = 1}.
}
\details{
\code{pmvnorm()} just calls \code{mvtnorm::pmvnorm} with the randomized
Quasi-Monte-Carlo procedure by Genz and Bretz. The argument \code{abseps} controls
the accuracy of the Gaussian integral approximation.
}
\examples{
x <- c(0, 0)
mean <- c(0, 0)
Sigma <- diag(2)

# compute density
dmvnorm(x = x, mean = mean, Sigma = Sigma)
dmvnorm(x = x, mean = mean, Sigma = Sigma, log = TRUE)

# compute CDF
pmvnorm(x = x, mean = mean, Sigma = Sigma)

# sample
rmvnorm(n = 3, mean = mean, Sigma = Sigma)
rmvnorm(mean = mean, Sigma = Sigma, log = TRUE)
}
\seealso{
Other simulation helpers: 
\code{\link{Simulator}},
\code{\link{correlated_regressors}()},
\code{\link{ddirichlet_cpp}()},
\code{\link{dmixnorm_cpp}()},
\code{\link{dtnorm_cpp}()},
\code{\link{dwishart_cpp}()},
\code{\link{gaussian_tv}()},
\code{\link{simulate_markov_chain}()}
}
\concept{simulation helpers}
\keyword{distribution}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/cholesky_root.R
\name{cov_to_chol}
\alias{cov_to_chol}
\alias{chol_to_cov}
\alias{unique_chol}
\title{Cholesky root of covariance matrix}
\usage{
cov_to_chol(cov, unique = TRUE)

chol_to_cov(chol)

unique_chol(chol)
}
\arguments{
\item{cov}{[\code{matrix()}]\cr
A covariance matrix.

It can also be the zero matrix, in which case the Cholesky root is defined as
the zero matrix.}

\item{unique}{[\code{logical(1)}]\cr
Ensure that the Cholesky decomposition is unique by restricting the diagonal
elements to be positive?}

\item{chol}{[\code{numeric()}]\cr
Cholesky root elements.}
}
\value{
For \code{\link{cov_to_chol}} a \code{numeric} \code{vector} of Cholesky root
elements.

For \code{\link{chol_to_cov}} a covariance \code{matrix}.
}
\description{
These functions compute the Cholesky root elements of a covariance matrix
and, conversely, build a covariance matrix from its Cholesky root elements.
}
\examples{
cov <- sample_covariance_matrix(4)
chol <- cov_to_chol(cov)
all.equal(cov, chol_to_cov(chol))
}
\seealso{
Other matrix helpers: 
\code{\link{check_correlation_matrix}()},
\code{\link{check_covariance_matrix}()},
\code{\link{check_one_hot_matrix}()},
\code{\link{check_transition_probability_matrix}()},
\code{\link{diff_cov}()},
\code{\link{insert_matrix_column}()},
\code{\link{matrix_diagonal_indices}()},
\code{\link{matrix_indices}()},
\code{\link{sample_correlation_matrix}()},
\code{\link{sample_covariance_matrix}()},
\code{\link{sample_transition_probability_matrix}()},
\code{\link{stationary_distribution}()}
}
\concept{matrix helpers}
\keyword{transformation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/timed.R
\name{timed}
\alias{timed}
\title{Interrupt long evaluations}
\usage{
timed(expression, seconds = Inf, on_time_out = "silent")
}
\arguments{
\item{expression}{[\code{expression}]\cr
An R expression to be evaluated.}

\item{seconds}{[\code{numeric(1)}]\cr
The number of seconds.}

\item{on_time_out}{[\code{character(1)}]\cr
Defines what action to take if the evaluation time exceeded, either:
\itemize{
\item \code{"error"} to throw an error exception
\item \code{"warning"} to return \code{NULL} along with a warning
\item \code{"silent"} (the default) to just return \code{NULL}
}}
}
\value{
The value of \code{expression} or, if the evaluation time exceeded, whatever
is specified for \code{on_time_out}.
}
\description{
This function interrupts an evaluation after a certain number of seconds.
Note the limitations documented in \code{\link[base]{setTimeLimit}}.
}
\examples{
foo <- function(x) {
  for (i in 1:10) Sys.sleep(x / 10)
  return(x)
}
timed(foo(0.5), 1)
timed(foo(1.5), 1)
}
\seealso{
Other function helpers: 
\code{\link{do.call_timed}()},
\code{\link{function_arguments}()},
\code{\link{function_body}()},
\code{\link{function_defaults}()},
\code{\link{quiet}()},
\code{\link{try_silent}()},
\code{\link{variable_name}()}
}
\concept{function helpers}
\keyword{simulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_numeric_vector.R
\name{check_numeric_vector}
\alias{check_numeric_vector}
\alias{assert_numeric_vector}
\alias{test_numeric_vector}
\title{Check numeric vector}
\usage{
check_numeric_vector(
  x,
  lower = -Inf,
  upper = Inf,
  finite = FALSE,
  any.missing = TRUE,
  all.missing = TRUE,
  len = NULL,
  min.len = NULL,
  max.len = NULL,
  unique = FALSE,
  sorted = FALSE,
  names = NULL,
  typed.missing = FALSE,
  null.ok = FALSE
)

assert_numeric_vector(
  x,
  lower = -Inf,
  upper = Inf,
  finite = FALSE,
  any.missing = TRUE,
  all.missing = TRUE,
  len = NULL,
  min.len = NULL,
  max.len = NULL,
  unique = FALSE,
  sorted = FALSE,
  names = NULL,
  typed.missing = FALSE,
  null.ok = FALSE,
  .var.name = checkmate::vname(x),
  add = NULL
)

test_numeric_vector(
  x,
  lower = -Inf,
  upper = Inf,
  finite = FALSE,
  any.missing = TRUE,
  all.missing = TRUE,
  len = NULL,
  min.len = NULL,
  max.len = NULL,
  unique = FALSE,
  sorted = FALSE,
  names = NULL,
  typed.missing = FALSE,
  null.ok = FALSE
)
}
\arguments{
\item{x}{[\code{any}]\cr
Object to check.}

\item{lower}{[\code{numeric(1)}]\cr
Lower value all elements of \code{x} must be greater than or equal to.}

\item{upper}{[\code{numeric(1)}]\cr
Upper value all elements of \code{x} must be lower than or equal to.}

\item{finite}{[\code{logical(1)}]\cr
Check for only finite values? Default is \code{FALSE}.}

\item{any.missing}{[\code{logical(1)}]\cr
Are vectors with missing values allowed? Default is \code{TRUE}.}

\item{all.missing}{[\code{logical(1)}]\cr
Are vectors with no non-missing values allowed? Default is \code{TRUE}.
Note that empty vectors do not have non-missing values.}

\item{len}{[\code{integer(1)}]\cr
Exact expected length of \code{x}.}

\item{min.len}{[\code{integer(1)}]\cr
Minimal length of \code{x}.}

\item{max.len}{[\code{integer(1)}]\cr
Maximal length of \code{x}.}

\item{unique}{[\code{logical(1)}]\cr
Must all values be unique? Default is \code{FALSE}.}

\item{sorted}{[\code{logical(1)}]\cr
Elements must be sorted in ascending order. Missing values are ignored.}

\item{names}{[\code{character(1)}]\cr
Check for names. See \code{\link[checkmate]{checkNamed}} for possible values.
Default is \dQuote{any} which performs no check at all.
Note that you can use \code{\link[checkmate]{checkSubset}} to check for a specific set of names.}

\item{typed.missing}{[\code{logical(1)}]\cr
If set to \code{FALSE} (default), all types of missing values (\code{NA}, \code{NA_integer_},
\code{NA_real_}, \code{NA_character_} or \code{NA_character_}) as well as empty vectors are allowed
while type-checking atomic input.
Set to \code{TRUE} to enable strict type checking.}

\item{null.ok}{[\code{logical(1)}]\cr
If set to \code{TRUE}, \code{x} may also be \code{NULL}.
In this case only a type check of \code{x} is performed, all additional checks are disabled.}

\item{.var.name}{[\code{character(1)}]\cr
Name of the checked object to print in assertions. Defaults to
the heuristic implemented in \code{\link[checkmate]{vname}}.}

\item{add}{[\code{AssertCollection}]\cr
Collection to store assertion messages. See \code{\link[checkmate]{AssertCollection}}.}
}
\value{
Same as documented in \code{\link[checkmate]{check_numeric}}.
}
\description{
These functions check whether the input is a numeric vector.
}
\examples{
x <- c(1, 2, "3")
check_numeric_vector(x)
test_numeric_vector(x)
\dontrun{
assert_numeric_vector(x)
}
}
\seealso{
Other vector helpers: 
\code{\link{check_probability_vector}()},
\code{\link{chunk_vector}()},
\code{\link{equidistant_vectors}()},
\code{\link{insert_vector_entry}()},
\code{\link{map_indices}()},
\code{\link{match_numerics}()},
\code{\link{permutations}()},
\code{\link{split_vector_at}()},
\code{\link{subsets}()},
\code{\link{vector_occurrence}()}
}
\concept{vector helpers}
\keyword{validation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/identical_structure.R
\name{identical_structure}
\alias{identical_structure}
\title{Check if two objects have identical structure}
\usage{
identical_structure(x, y)
}
\arguments{
\item{x, y}{[\code{any}]\cr
Two objects.}
}
\value{
Either \code{TRUE} if \code{x} and \code{y} have the same structure, and
\code{FALSE}, else.
}
\description{
This function determines whether two objects have the same structure,
\itemize{
\item which includes the \code{\link[base]{mode}}, \code{\link[base]{class}} and
dimension
\item but does \emph{not} include concrete values or attributes.
}
}
\examples{
identical_structure(integer(1), 1L)
identical_structure(diag(2), matrix(rnorm(4), 2, 2))
identical_structure(diag(2), data.frame(diag(2)))
}
\references{
Inspired by \url{https://stackoverflow.com/a/45548885/15157768}.
}
\seealso{
Other package helpers: 
\code{\link{Dictionary}},
\code{\link{Storage}},
\code{\link{check_missing}()},
\code{\link{find_namespace_calls}()},
\code{\link{input_check_response}()},
\code{\link{match_arg}()},
\code{\link{package_logo}()},
\code{\link{print_data.frame}()},
\code{\link{print_matrix}()},
\code{\link{system_information}()},
\code{\link{unexpected_error}()},
\code{\link{user_confirm}()}
}
\concept{package helpers}
\keyword{validation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/user_confirm.R
\name{user_confirm}
\alias{user_confirm}
\title{User confirmation}
\usage{
user_confirm(question = "Question?", default = FALSE)
}
\arguments{
\item{question}{[\code{character(1)}]\cr
The binary question to ask. It should end with a question mark.}

\item{default}{[\code{logical(1)}]\cr
The default decision.}
}
\value{
Either \code{TRUE} or \code{FALSE}.
}
\description{
This function asks in an interactive question a binary question.
}
\seealso{
Other package helpers: 
\code{\link{Dictionary}},
\code{\link{Storage}},
\code{\link{check_missing}()},
\code{\link{find_namespace_calls}()},
\code{\link{identical_structure}()},
\code{\link{input_check_response}()},
\code{\link{match_arg}()},
\code{\link{package_logo}()},
\code{\link{print_data.frame}()},
\code{\link{print_matrix}()},
\code{\link{system_information}()},
\code{\link{unexpected_error}()}
}
\concept{package helpers}
\keyword{validation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/find_namespace_calls.R
\name{find_namespace_calls}
\alias{find_namespace_calls}
\title{Namespace calls}
\usage{
find_namespace_calls(path = "R", triple_colon = FALSE, as_list = FALSE)
}
\arguments{
\item{path}{[\code{character(1)}]\cr
The path name to a folder. All \code{.R} files in this folder and sub-directories
will be searched.}

\item{triple_colon}{[\code{logical(1)}]\cr
Also search for \code{:::}?}

\item{as_list}{[\code{logical(1)}]\cr
Simplify the output into a \code{list} of unique function names per package?}
}
\value{
A \code{data.frame}. If \code{as_list = TRUE}, a \code{list}.
}
\description{
This function searches for namespace calls in \code{.R} files, i.e., code lines of
the format \verb{<package name>::<function name>}.
}
\examples{
\dontrun{
find_namespace_calls()
find_namespace_calls(as_list = TRUE)
}
}
\seealso{
Other package helpers: 
\code{\link{Dictionary}},
\code{\link{Storage}},
\code{\link{check_missing}()},
\code{\link{identical_structure}()},
\code{\link{input_check_response}()},
\code{\link{match_arg}()},
\code{\link{package_logo}()},
\code{\link{print_data.frame}()},
\code{\link{print_matrix}()},
\code{\link{system_information}()},
\code{\link{unexpected_error}()},
\code{\link{user_confirm}()}
}
\concept{package helpers}
\keyword{packaging}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/oeli-package.R
\docType{package}
\name{oeli-package}
\alias{oeli}
\alias{oeli-package}
\title{oeli: Some Utilities for Developing Data Science Software}
\description{
\if{html}{\figure{logo.png}{options: style='float: right' alt='logo' width='120'}}

A collection of general-purpose helper functions that I (and maybe others) find useful when developing data science software. Includes tools for simulation, data transformation, input validation, and more.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://github.com/loelschlaeger/oeli}
  \item \url{http://loelschlaeger.de/oeli/}
  \item Report bugs at \url{https://github.com/loelschlaeger/oeli/issues}
}

}
\author{
\strong{Maintainer}: Lennart Oelschläger \email{oelschlaeger.lennart@gmail.com}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print_data.frame.R
\name{print_data.frame}
\alias{print_data.frame}
\title{Print (abbreviated) \code{data.frame}}
\usage{
print_data.frame(
  x,
  rows = NULL,
  cols = NULL,
  digits = NULL,
  row.names = TRUE,
  col.names = TRUE
)
}
\arguments{
\item{x}{[\code{data.frame}]\cr
A \code{data.frame}.}

\item{rows, cols}{[\code{integer(1)} | \code{NULL} ]\cr
The number of rows or columns to be printed, greater or equal \code{2}.

Printing is abbreviated in the middle.

Can be \code{NULL} to print everything.}

\item{digits}{[\code{integer(1)} | \code{NULL} ]\cr
The number of decimal places to be used.

Negative values are allowed, resulting in rounding to a power of ten.

Can be \code{NULL} to not round.}

\item{row.names, col.names}{[\code{logical(1)}]\cr
Print row names or column names?}
}
\value{
Invisibly returns \code{x}.
}
\description{
This function prints a (possibly abbreviated) \code{data.frame}.
}
\examples{
x <- data.frame(1:10, LETTERS[1:10], stats::rnorm(10))
print_data.frame(x, rows = 7)
print_data.frame(x, rows = 7, cols = 2)
print_data.frame(x, rows = 7, cols = 2, digits = 1)
print_data.frame(x, rows = 7, cols = 2, digits = 1, row.names = FALSE)
print_data.frame(x, rows = 7, cols = 2, digits = 1, col.names = FALSE)
}
\seealso{
Other package helpers: 
\code{\link{Dictionary}},
\code{\link{Storage}},
\code{\link{check_missing}()},
\code{\link{find_namespace_calls}()},
\code{\link{identical_structure}()},
\code{\link{input_check_response}()},
\code{\link{match_arg}()},
\code{\link{package_logo}()},
\code{\link{print_matrix}()},
\code{\link{system_information}()},
\code{\link{unexpected_error}()},
\code{\link{user_confirm}()}
}
\concept{package helpers}
\keyword{packaging}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/covariance_differencing.R
\name{diff_cov}
\alias{diff_cov}
\alias{undiff_cov}
\alias{delta}
\alias{M}
\title{Difference and un-difference covariance matrix}
\usage{
diff_cov(cov, ref = 1)

undiff_cov(cov_diff, ref = 1)

delta(ref = 1, dim)

M(ranking = seq_len(dim), dim)
}
\arguments{
\item{cov, cov_diff}{[\code{matrix()}]\cr
A (differenced) covariance matrix of dimension \code{dim}
(or \code{dim} - 1, respectively).}

\item{ref}{[\code{integer(1)}]\cr
The reference row between \code{1} and \code{dim} for
differencing that maps \code{cov} to \code{cov_diff}, see details.}

\item{dim}{[\code{integer(1)}]\cr
The matrix dimension.}

\item{ranking}{[\code{integer()}]\cr
The integers \code{1} to \code{dim} in arbitrary order.}
}
\value{
A (differenced or un-differenced) covariance \code{matrix}.
}
\description{
These functions difference and un-difference random vectors and covariance
matrices.
}
\details{
Assume \eqn{x \sim N(0, \Sigma)} is a multivariate normally distributed
random vector of dimension \eqn{n}. We may want to consider the differenced
vector \deqn{\tilde x = (x_1 - x_k, x_2 - x_k, \dots, x_n - x_k)',} excluding
the \eqn{k}th element (hence, \eqn{\tilde x} is of dimension
\eqn{(n - 1) \times 1}). Formally, \eqn{\tilde x = \Delta_k x}, where
\eqn{\Delta_k} is a difference operator that depends on the reference
row \eqn{k}. More precise, \eqn{\Delta_k} is the identity matrix of dimension
\eqn{n} without row \eqn{k} and with \eqn{-1}s in column \eqn{k}.
The difference operator \eqn{\Delta_k} can be computed via
\code{delta(ref = k, dim = n)}.

Then, \eqn{\tilde x \sim N(0, \tilde \Sigma)}, where
\deqn{\tilde{\Sigma} = \Delta_k \Sigma \Delta_k'}
is the differenced covariance matrix with respect to row \eqn{k = 1,\dots,n}.
The differenced covariance matrix \eqn{\tilde \Sigma} can be computed via
\code{diff_delta(Sigma, ref = k)}.

Since \eqn{\Delta_k} is a non-bijective mapping, \eqn{\Sigma} cannot be
uniquely restored from \eqn{\tilde \Sigma}. However, it is possible to
compute a non-unique solution \eqn{\Sigma_0}, such that
\eqn{\Delta_k \Sigma_0 \Delta_k = \tilde \Sigma}. For such a non-unique
solution, we add a column and a row of zeros
at column and row number \eqn{k} to \eqn{\tilde{\Sigma}}, respectively.
An "un-differenced" covariance matrix \eqn{\Sigma_0} can be computed via
\code{undiff_delta(Sigma_diff, ref = k)}.

As a alternative to \eqn{\Delta_k}, the function \code{M()} returns a matrix for
taking differences such that the resulting vector is negative.
}
\examples{
n <- 4
Sigma <- sample_covariance_matrix(dim = n)
k <- 2
x <- c(1, 3, 2, 4)

# build difference operator
delta_k <- delta(ref = k, dim = n)

# difference vector
delta_k \%*\% x

# difference Sigma
(Sigma_diff <- diff_cov(Sigma, ref = k))

# un-difference Sigma
(Sigma_0 <- undiff_cov(Sigma_diff, ref = k))

# difference again
Sigma_diff_2 <- diff_cov(Sigma_0, ref = k)
all.equal(Sigma_diff, Sigma_diff_2)

# difference such that the resulting vector is negative
M(ranking = order(x, decreasing = TRUE), dim = n) \%*\% x
}
\seealso{
Other matrix helpers: 
\code{\link{check_correlation_matrix}()},
\code{\link{check_covariance_matrix}()},
\code{\link{check_one_hot_matrix}()},
\code{\link{check_transition_probability_matrix}()},
\code{\link{cov_to_chol}()},
\code{\link{insert_matrix_column}()},
\code{\link{matrix_diagonal_indices}()},
\code{\link{matrix_indices}()},
\code{\link{sample_correlation_matrix}()},
\code{\link{sample_covariance_matrix}()},
\code{\link{sample_transition_probability_matrix}()},
\code{\link{stationary_distribution}()}
}
\concept{matrix helpers}
\keyword{transformation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/delete_columns_data.frame.R
\name{delete_columns_data.frame}
\alias{delete_columns_data.frame}
\title{Deleting \code{data.frame} columns}
\usage{
delete_columns_data.frame(df, column_names)
}
\arguments{
\item{df}{[\code{data.frame}]\cr
A \code{data.frame}.}

\item{column_names}{[\code{character()}]\cr
The name(s) of column(s) of \code{df} to delete.}
}
\value{
The input \code{df} without the columns defined by \code{column_names}.
}
\description{
This function deletes columns of a \code{data.frame} by name.
}
\examples{
df <- data.frame("label" = c("A", "B"), "number" = 1:10)
delete_columns_data.frame(df = df, column_names = "label")
delete_columns_data.frame(df = df, column_names = "number")
delete_columns_data.frame(df = df, column_names = c("label", "number"))
}
\seealso{
Other data.frame helpers: 
\code{\link{group_data.frame}()},
\code{\link{occurrence_info}()},
\code{\link{round_data.frame}()}
}
\concept{data.frame helpers}
\keyword{transformation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vector_occurrence.R
\name{vector_occurrence}
\alias{vector_occurrence}
\title{Find the positions of first or last occurrence of unique vector elements}
\usage{
vector_occurrence(x, type = "first")
}
\arguments{
\item{x}{[\code{atomic()}]\cr
A \code{vector}.}

\item{type}{[\code{character(1)}]\cr
Either \code{"first"} for the first or \code{"last"} for the last occurrence.}
}
\value{
An \code{integer} \code{vector}, the positions of the unique vector elements.
The ordering corresponds to \code{unique(x)}, i.e., the \eqn{i}-th element in
the output is the (first or last) occurrence of the \eqn{i}-th element from
\code{unique(x)}.
}
\description{
This function finds the positions of first or last occurrence of unique
vector elements.
}
\examples{
x <- c(1, 1, 1, 2, 2, 2, 3, 3, 3)
unique(x)
vector_occurrence(x, "first")
vector_occurrence(x, "last")
}
\seealso{
Other vector helpers: 
\code{\link{check_numeric_vector}()},
\code{\link{check_probability_vector}()},
\code{\link{chunk_vector}()},
\code{\link{equidistant_vectors}()},
\code{\link{insert_vector_entry}()},
\code{\link{map_indices}()},
\code{\link{match_numerics}()},
\code{\link{permutations}()},
\code{\link{split_vector_at}()},
\code{\link{subsets}()}
}
\concept{vector helpers}
\keyword{indexing}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/equidistant_vectors.R
\name{equidistant_vectors}
\alias{equidistant_vectors}
\title{Generate equidistant vectors in Euclidean space}
\usage{
equidistant_vectors(dim, n = dim + 1, dist = 1, center = rep(0, dim))
}
\arguments{
\item{dim}{[\code{integer(1)}]\cr
The dimension.}

\item{n}{[\code{integer(1)}]\cr
The number of vertices to return. Cannot be larger than \code{dim + 1}.}

\item{dist}{[\code{numeric(1)}]\cr
Desired pairwise Euclidean distance between any two vertices.}

\item{center}{[\code{numeric(dim)}]\cr
Desired center.}
}
\value{
A matrix, where each column is a vertex of the simplex.
}
\description{
This function constructs the coordinates of vertices of a regular simplex
in \eqn{\mathbb{R}^{\code{dim}}} and returns the first \code{n} of them,
\itemize{
\item scaled so that the pairwise Euclidean distance between any two vertices
equals \code{dist},
\item and centered so their centroid is at \code{center}.
}
}
\examples{
dim <- n <- 3
(dist <- runif(1))
(center <- rnorm(dim))
(V <- equidistant_vectors(dim = dim, n = n, dist = dist, center = center))
rowMeans(V)
dist(t(V))
}
\seealso{
Other vector helpers: 
\code{\link{check_numeric_vector}()},
\code{\link{check_probability_vector}()},
\code{\link{chunk_vector}()},
\code{\link{insert_vector_entry}()},
\code{\link{map_indices}()},
\code{\link{match_numerics}()},
\code{\link{permutations}()},
\code{\link{split_vector_at}()},
\code{\link{subsets}()},
\code{\link{vector_occurrence}()}
}
\concept{vector helpers}
\keyword{simulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/simulator.R
\name{Simulator}
\alias{Simulator}
\title{Simulator R6 Object}
\description{
Creates a simulation setup, where a function \code{f} is evaluated \code{runs} times,
optionally at each combination of input values.

Provides some convenience (see below for more details):
\itemize{
\item Simulation results can be restored from a backup if the R session crashes.
\item More simulation runs can be conducted after the initial simulation, failed
simulation cases can be re-run.
\item Parallel computation and progress updates are supported.
}
}
\details{
\subsection{Backup}{

Simulation results can be saved to disk, allowing you to restore the results
if the R session is interrupted or crashes before the simulation completes.
To enable backup, set \code{backup = TRUE} in the \verb{$go()} method, which will
create a backup directory at the location specified by \code{path}.
To restore, use \code{Simulator$initialize(use_backup = path)}.
}

\subsection{More runs and re-run}{

If additional simulation runs are needed, simply call the \verb{$go()} method
again. Any cases that were not successfully completed in previous runs will
be attempted again.
}

\subsection{Parallel computation}{

By default, simulations run sequentially. But since they are independent,
they can be parallelized to decrease computation time. To enable parallel
computation, use the \href{https://future.futureverse.org/}{\code{{future}} framework}.
For example, run
\preformatted{future::plan(future::multisession, workers = 4)}
in advance for computation in 4 parallel R sessions.
}

\subsection{Progress updates}{

Use the \href{https://progressr.futureverse.org/}{\code{{progressr}} framework} to
get progress updates. For example, run the following in advance:
\preformatted{progressr::handlers(global = TRUE)
progressr::handlers(
  progressr::handler_progress(format = ">> :percent, :eta to go :message")
)
}
}
}
\examples{
# 1. Initialize a new simulation setup:
object <- Simulator$new(verbose = TRUE)

# 2. Define function `f` and arguments (if any):
f <- function(x, y = 1) {
  Sys.sleep(runif(1)) # to see progress updates
  x + y
}
x_args <- list(1, 2)
object$define(f = f, x = x_args)
print(object)

# 3. Define 'future' and 'progress' (optional):
\dontrun{
future::plan(future::sequential)
progressr::handlers(global = TRUE)}

# 4. Evaluate `f` `runs` times at each parameter combination (backup is optional):
path <- file.path(tempdir(), paste0("backup_", format(Sys.time(), "\%Y-\%m-\%d-\%H-\%M-\%S")))
object$go(runs = 2, backup = TRUE, path = path)

# 5. Access the results:
object$results

# 6. Check if cases are pending or if an error occurred:
object$cases

# 7. Restore simulation results from backup:
object_restored <- Simulator$new(use_backup = path)
print(object_restored)
\dontrun{all.equal(object, object_restored)}

# 8. Run more simulations and pending simulations (if any):
object_restored$go(runs = 2)
}
\seealso{
Other simulation helpers: 
\code{\link{correlated_regressors}()},
\code{\link{ddirichlet_cpp}()},
\code{\link{dmixnorm_cpp}()},
\code{\link{dmvnorm_cpp}()},
\code{\link{dtnorm_cpp}()},
\code{\link{dwishart_cpp}()},
\code{\link{gaussian_tv}()},
\code{\link{simulate_markov_chain}()}
}
\concept{simulation helpers}
\keyword{simulation}
\section{Active bindings}{
\if{html}{\out{<div class="r6-active-bindings">}}
\describe{
\item{\code{results}}{[\code{tibble}, read-only]\cr
The simulation results.}

\item{\code{cases}}{[\code{tibble}, read-only]\cr
The simulation cases.}
}
\if{html}{\out{</div>}}
}
\section{Methods}{
\subsection{Public methods}{
\itemize{
\item \href{#method-Simulator-new}{\code{Simulator$new()}}
\item \href{#method-Simulator-print}{\code{Simulator$print()}}
\item \href{#method-Simulator-define}{\code{Simulator$define()}}
\item \href{#method-Simulator-go}{\code{Simulator$go()}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Simulator-new"></a>}}
\if{latex}{\out{\hypertarget{method-Simulator-new}{}}}
\subsection{Method \code{new()}}{
Initialize a \code{Simulator} object, either a new one or from backup.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Simulator$new(
  use_backup = NULL,
  verbose = getOption("verbose", default = FALSE)
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{use_backup}}{[\code{NULL} | \code{character(1)}]\cr
Optionally a path to a backup folder previously used in \verb{$go()}.}

\item{\code{verbose}}{[\code{logical(1)}]\cr
Provide info? Does not include progress updates. For that, see details.}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Simulator-print"></a>}}
\if{latex}{\out{\hypertarget{method-Simulator-print}{}}}
\subsection{Method \code{print()}}{
Print method.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Simulator$print()}\if{html}{\out{</div>}}
}

}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Simulator-define"></a>}}
\if{latex}{\out{\hypertarget{method-Simulator-define}{}}}
\subsection{Method \code{define()}}{
Define function and arguments for a new \code{Simulator} object.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Simulator$define(f, ...)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{f}}{[\code{function}]\cr
A \code{function} to evaluate.}

\item{\code{...}}{Arguments for \code{f}. Each value must be
\enumerate{
\item named after an argument of \code{f}, and
\item a \code{list}, where each element is a variant of that argument for \code{f}.
}}
}
\if{html}{\out{</div>}}
}
}
\if{html}{\out{<hr>}}
\if{html}{\out{<a id="method-Simulator-go"></a>}}
\if{latex}{\out{\hypertarget{method-Simulator-go}{}}}
\subsection{Method \code{go()}}{
Run simulations.
\subsection{Usage}{
\if{html}{\out{<div class="r">}}\preformatted{Simulator$go(
  runs = 0,
  backup = FALSE,
  path = paste0("backup_", format(Sys.time(), "\%Y-\%m-\%d-\%H-\%M-\%S"))
)}\if{html}{\out{</div>}}
}

\subsection{Arguments}{
\if{html}{\out{<div class="arguments">}}
\describe{
\item{\code{runs}}{[\code{integer(1)}]\cr
The number of (additional) simulation runs.

If \code{runs = 0}, only pending cases (if any) are solved.}

\item{\code{backup}}{[\code{logical(1)}]\cr
Create a backup under \code{path}?}

\item{\code{path}}{[\code{character(1)}]\cr
Only relevant, if \code{backup = TRUE}.

In this case, a path for a new folder, which does not yet exist and
allows reading and writing.}
}
\if{html}{\out{</div>}}
}
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/round_data.frame.R
\name{round_data.frame}
\alias{round_data.frame}
\title{Round \code{numeric} columns of a \code{data.frame}}
\usage{
round_data.frame(df, digits = 0)
}
\arguments{
\item{df}{[\code{data.frame}]\cr
A \code{data.frame}.}

\item{digits}{[\code{integer(1)} | \code{NULL} ]\cr
The number of decimal places to be used.

Negative values are allowed, resulting in rounding to a power of ten.

Can be \code{NULL} to not round.}
}
\value{
A \code{data.frame}.
}
\description{
This function rounds (only) the \code{numeric} columns of a
\code{data.frame}.
}
\examples{
df <- data.frame("label" = c("A", "B"), "number" = rnorm(10))
round_data.frame(df, digits = 1)
}
\seealso{
Other data.frame helpers: 
\code{\link{delete_columns_data.frame}()},
\code{\link{group_data.frame}()},
\code{\link{occurrence_info}()}
}
\concept{data.frame helpers}
\keyword{transformation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/function_defaults.R
\name{function_defaults}
\alias{function_defaults}
\title{Get default function arguments}
\usage{
function_defaults(f, exclude = NULL)
}
\arguments{
\item{f}{[\code{function}]\cr
A \code{function}.}

\item{exclude}{[\code{NULL} | \code{character()}]\cr
Argument names to exclude.

Can be \code{NULL} (default) to not exclude any argument names.}
}
\value{
A named \code{list}.
}
\description{
This function returns the default function arguments (if any).
}
\examples{
f <- function(a, b = 1, c = "", ...) { }
function_defaults(f)
function_defaults(f, exclude = "b")
}
\seealso{
Other function helpers: 
\code{\link{do.call_timed}()},
\code{\link{function_arguments}()},
\code{\link{function_body}()},
\code{\link{quiet}()},
\code{\link{timed}()},
\code{\link{try_silent}()},
\code{\link{variable_name}()}
}
\concept{function helpers}
\keyword{functional}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merge_lists.R
\name{merge_lists}
\alias{merge_lists}
\title{Merge named lists}
\usage{
merge_lists(...)
}
\arguments{
\item{...}{One or more named \code{list}(s).}
}
\value{
A \code{list}.
}
\description{
This function merges \code{list}s based on their element names. Elements are
only included in the final output \code{list}, if no former \code{list} has
contributed an element with the same name.
}
\examples{
merge_lists(list("a" = 1, "b" = 2), list("b" = 3, "c" = 4, "d" = NULL))
}
\seealso{
Other list helpers: 
\code{\link{check_list_of_lists}()}
}
\concept{list helpers}
\keyword{transformation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_correlation_matrix.R
\name{check_correlation_matrix}
\alias{check_correlation_matrix}
\alias{assert_correlation_matrix}
\alias{test_correlation_matrix}
\title{Check correlation matrix}
\usage{
check_correlation_matrix(x, dim = NULL, tolerance = sqrt(.Machine$double.eps))

assert_correlation_matrix(
  x,
  dim = NULL,
  tolerance = sqrt(.Machine$double.eps),
  .var.name = checkmate::vname(x),
  add = NULL
)

test_correlation_matrix(x, dim = NULL, tolerance = sqrt(.Machine$double.eps))
}
\arguments{
\item{x}{[\code{any}]\cr
Object to check.}

\item{dim}{[\code{integer(1)}]\cr
The matrix dimension.}

\item{tolerance}{[\code{numeric(1)}]\cr
A non-negative tolerance value.}

\item{.var.name}{[\code{character(1)}]\cr
Name of the checked object to print in assertions. Defaults to
the heuristic implemented in \code{\link[checkmate]{vname}}.}

\item{add}{[\code{AssertCollection}]\cr
Collection to store assertion messages. See \code{\link[checkmate]{AssertCollection}}.}
}
\value{
Same as documented in \code{\link[checkmate]{check_matrix}}.
}
\description{
These functions check whether the input fulfills the properties of a
correlation matrix.
}
\examples{
M <- matrix(c(1,  0.9,  0.9, 0.9,  1,  -0.9, 0.9,  -0.9,  1), nrow = 3)
check_correlation_matrix(M)
test_correlation_matrix(M)
\dontrun{
assert_correlation_matrix(M)
}
}
\seealso{
Other matrix helpers: 
\code{\link{check_covariance_matrix}()},
\code{\link{check_one_hot_matrix}()},
\code{\link{check_transition_probability_matrix}()},
\code{\link{cov_to_chol}()},
\code{\link{diff_cov}()},
\code{\link{insert_matrix_column}()},
\code{\link{matrix_diagonal_indices}()},
\code{\link{matrix_indices}()},
\code{\link{sample_correlation_matrix}()},
\code{\link{sample_covariance_matrix}()},
\code{\link{sample_transition_probability_matrix}()},
\code{\link{stationary_distribution}()}
}
\concept{matrix helpers}
\keyword{validation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/try_silent.R
\name{try_silent}
\alias{try_silent}
\title{Try an expression silently}
\usage{
try_silent(expr)
}
\arguments{
\item{expr}{[\code{expression}]\cr
An R expression to be evaluated.}
}
\value{
Either the value of \code{expr} or in case of a failure an object of class
\code{fail}, which contains the error message.
}
\description{
This function tries to execute \code{expr} and returns a string with the
error message if the execution failed.
}
\details{
This function is a wrapper for \code{\link[base]{try}}.
}
\examples{
\dontrun{
try_silent(1 + 1)
try_silent(1 + "1")
}
}
\seealso{
Other function helpers: 
\code{\link{do.call_timed}()},
\code{\link{function_arguments}()},
\code{\link{function_body}()},
\code{\link{function_defaults}()},
\code{\link{quiet}()},
\code{\link{timed}()},
\code{\link{variable_name}()}
}
\concept{function helpers}
\keyword{functional}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_list_of_lists.R
\name{check_list_of_lists}
\alias{check_list_of_lists}
\alias{assert_list_of_lists}
\alias{test_list_of_lists}
\title{Check list of lists}
\usage{
check_list_of_lists(x, len = NULL)

assert_list_of_lists(
  x,
  len = NULL,
  .var.name = checkmate::vname(x),
  add = NULL
)

test_list_of_lists(x, len = NULL)
}
\arguments{
\item{x}{[\code{any}]\cr
Object to check.}

\item{len}{[\code{integer(1)}]\cr
Exact expected length of \code{x}.}

\item{.var.name}{[\code{character(1)}]\cr
Name of the checked object to print in assertions. Defaults to
the heuristic implemented in \code{\link[checkmate]{vname}}.}

\item{add}{[\code{AssertCollection}]\cr
Collection to store assertion messages. See \code{\link[checkmate]{AssertCollection}}.}
}
\value{
Same as documented in \code{\link[checkmate]{check_list}}.
}
\description{
These functions check whether the input is a list that contains list
elements.
}
\examples{
L <- list(list(1), list(2), 3)
check_list_of_lists(L)
test_list_of_lists(L)
\dontrun{
assert_list_of_lists(L)
}
}
\seealso{
Other list helpers: 
\code{\link{merge_lists}()}
}
\concept{list helpers}
\keyword{validation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/do.call_timed.R
\name{do.call_timed}
\alias{do.call_timed}
\title{Measure computation time}
\usage{
do.call_timed(what, args, units = "secs")
}
\arguments{
\item{what, args}{Passed to \code{\link[base]{do.call}}.}

\item{units}{Passed to \code{\link[base]{difftime}}.}
}
\value{
A list of the two elements \code{"result"} (the results of the \code{do.call}
call) and \code{"time"} (the computation time).
}
\description{
This function measures the computation time of a call.
}
\details{
This function is a wrapper for \code{\link[base]{do.call}}.
}
\examples{
\dontrun{
what <- function(s) {
  Sys.sleep(s)
  return(s)
}
args <- list(s = 1)
do.call_timed(what = what, args = args)
}
}
\seealso{
Other function helpers: 
\code{\link{function_arguments}()},
\code{\link{function_body}()},
\code{\link{function_defaults}()},
\code{\link{quiet}()},
\code{\link{timed}()},
\code{\link{try_silent}()},
\code{\link{variable_name}()}
}
\concept{function helpers}
\keyword{simulation}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gaussian_tv.R
\name{gaussian_tv}
\alias{gaussian_tv}
\title{Gaussian total variation}
\usage{
gaussian_tv(
  mean1,
  mean2,
  Sigma1,
  Sigma2,
  method = c("auto", "mc", "cubature"),
  n = 10000,
  tol_equal = 1e-06,
  eps = 1e-06
)
}
\arguments{
\item{mean1, mean2}{[\code{numeric(p)}]\cr
The mean vectors.}

\item{Sigma1, Sigma2}{[\code{matrix(nrow = p, ncol = p)}]\cr
The covariance matrices.}

\item{method}{[\code{character(1)}]\cr
Computation method. One of:
\itemize{
\item \code{"auto"}: use closed-form formula when covariances are equal, otherwise
use \code{"cubature"} for \eqn{p \le 2} and \code{"mc"} for higher dimensions.
\item \code{"mc"}: estimate via Monte Carlo importance sampling from the mixture
\eqn{0.5 (f_1 + f_2)}.
\item \code{"cubature"}: compute overlap via adaptive cubature integration over a
bounding box, then convert to TV. Exact but slow for \eqn{p \ge 2}.
}}

\item{n}{[\code{integer(1)}]\cr
Number of Monte Carlo samples to draw.}

\item{tol_equal}{[\code{numeric(1)}]\cr
Numerical tolerance used to decide whether \code{Sigma1} and \code{Sigma2} are
considered equal (enabling the closed-form formula in \code{"auto"} mode).}

\item{eps}{[\code{numeric(1)}]\cr
Only used when \code{method = "cubature"}. Specifies the total probability mass
allowed to lie outside the integration hyper-rectangle across all dimensions.
This determines the numerical integration bounds: the function chooses limits
so that the probability of a point from either Gaussian falling outside the
box is at most \code{eps}. The bound is split evenly across dimensions via a union
bound, so the per-dimension tail probability is approximately \code{eps / p}.
Smaller values produce wider bounds (slower but more accurate integration),
while larger values yield narrower bounds (faster but potentially less
accurate).}
}
\value{
The total variation in [0, 1].
}
\description{
Computes the total variation (TV) between two multivariate Gaussian
distributions \eqn{f_1,f_2}: \deqn{\mathrm{TV}(f_1, f_2) = \tfrac{1}{2}
\int_{\mathbb{R}^p} \lvert f_1(x) - f_2(x) \rvert \, dx.}
The value ranges from 0 (identical distributions) to 1 (no overlap).
}
\examples{
### univariate case
mean1 <- 0
mean2 <- 1
Sigma1 <- Sigma2 <- matrix(1)
gaussian_tv(mean1, mean2, Sigma1, Sigma2)

### bivariate case
mean1 <- c(0, 0)
mean2 <- c(1, 1)
Sigma1 <- matrix(c(1, 0.2, 0.2, 1), ncol = 2)
Sigma2 <- matrix(c(1.5, -0.3, -0.3, 1), ncol = 2)
gaussian_tv(mean1, mean2, Sigma1, Sigma2, method = "mc", n = 1e3)
}
\seealso{
Other simulation helpers: 
\code{\link{Simulator}},
\code{\link{correlated_regressors}()},
\code{\link{ddirichlet_cpp}()},
\code{\link{dmixnorm_cpp}()},
\code{\link{dmvnorm_cpp}()},
\code{\link{dtnorm_cpp}()},
\code{\link{dwishart_cpp}()},
\code{\link{simulate_markov_chain}()}
}
\concept{simulation helpers}
\keyword{distribution}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/RcppExports.R, R/wishart.R
\name{dwishart_cpp}
\alias{dwishart_cpp}
\alias{rwishart_cpp}
\alias{dwishart}
\alias{rwishart}
\title{Wishart distribution}
\usage{
dwishart_cpp(x, df, scale, log = FALSE, inv = FALSE)

rwishart_cpp(df, scale, inv = FALSE)

dwishart(x, df, scale, log = FALSE, inv = FALSE)

rwishart(df, scale, inv = FALSE)
}
\arguments{
\item{x}{[\code{matrix()}]\cr
A covariance matrix of dimension \code{p}.}

\item{df}{[\code{integer()}]\cr
The degrees of freedom greater of equal \code{p}.}

\item{scale}{[\code{matrix()}]\cr
The scale covariance matrix of dimension \code{p}.}

\item{log}{[\code{logical(1)}]\cr
Return the logarithm of the density value?}

\item{inv}{[\code{logical(1)}]\cr
Use this inverse Wishart distribution?}
}
\value{
For \code{dwishart()}: The density value.

For \code{rwishart()}: A \code{matrix}, the random draw.
}
\description{
The function \code{dwishart()} computes the density of a Wishart distribution.

The function \code{rwishart()} samples from a Wishart distribution.

The functions with suffix \verb{_cpp} perform no input checks, hence are faster.
}
\examples{
x <- diag(2)
df <- 6
scale <- matrix(c(1, -0.3, -0.3, 0.8), ncol = 2)

# compute density
dwishart(x = x, df = df, scale = scale)
dwishart(x = x, df = df, scale = scale, log = TRUE)
dwishart(x = x, df = df, scale = scale, inv = TRUE)

# sample
rwishart(df = df, scale = scale)
rwishart(df = df, scale = scale, inv = TRUE)

# expectation of Wishart is df * scale
n <- 100
replicate(n, rwishart(df = df, scale = scale), simplify = FALSE) |>
  Reduce(f = "+") / n
df * scale

# expectation of inverse Wishart is scale / (df - p - 1)
n <- 100
replicate(n, rwishart(df = df, scale = scale, TRUE), simplify = FALSE) |>
  Reduce(f = "+") / n
scale / (df - 2 - 1)
}
\seealso{
Other simulation helpers: 
\code{\link{Simulator}},
\code{\link{correlated_regressors}()},
\code{\link{ddirichlet_cpp}()},
\code{\link{dmixnorm_cpp}()},
\code{\link{dmvnorm_cpp}()},
\code{\link{dtnorm_cpp}()},
\code{\link{gaussian_tv}()},
\code{\link{simulate_markov_chain}()}
}
\concept{simulation helpers}
\keyword{distribution}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/match_numerics.R
\name{match_numerics}
\alias{match_numerics}
\title{Best-possible match of two numeric vectors}
\usage{
match_numerics(x, y)
}
\arguments{
\item{x, y}{[\code{numeric()}]\cr
Two vectors of the same length.}
}
\value{
An \code{integer} vector of length \code{length(x)} with the positions of
\code{y} in \code{x}.
}
\description{
This function matches the indices of two numeric vectors as good as possible
(that means with the smallest possible sum of deviations).
}
\examples{
x <- c(-1, 0, 1)
y <- c(0.1, 1.5, -1.2)
match_numerics(x, y)
}
\seealso{
Other vector helpers: 
\code{\link{check_numeric_vector}()},
\code{\link{check_probability_vector}()},
\code{\link{chunk_vector}()},
\code{\link{equidistant_vectors}()},
\code{\link{insert_vector_entry}()},
\code{\link{map_indices}()},
\code{\link{permutations}()},
\code{\link{split_vector_at}()},
\code{\link{subsets}()},
\code{\link{vector_occurrence}()}
}
\concept{vector helpers}
\keyword{indexing}
